(declaim (optimize (debug 3) (speed 0) (size 0)))

(in-package :tutor)

(defclass region-view (view)
  () (:documentation "This is created  "))

(defclass alignment-view (region-view)
  ())

;; value can be a box instead of a number in which case we draw a ?
;; and a line from the ? to the op of the box pointed to by the
;; value.
(defclass argument ()
  ;; value is the current value of the argument, but when it's being
  ;; edited, the actual answer is stored in answer.
  ((x :type number :initarg :x :initform 0 :accessor arg-x)
   (y :type number :initarg :y :initform 0 :accessor arg-y)
   (align :initform nil :accessor arg-align)
   (id :initarg :id :initform (gentemp "ARG" "KEYWORD") :accessor arg-id)
   (value :initarg :value :initform nil :accessor arg-value)
   (answer :initform nil :initarg :answer :accessor arg-answer)
   (editable :initform nil :initarg :editable :accessor arg-editable)
   (user-answer :initform nil :accessor arg-user-answer)
   (widget :initform nil :accessor arg-widget)))

(defmethod print-object ((obj argument) stream)
  (print-unreadable-object (obj stream :type t :identity t)
    (format stream "~a" (if (arg-editable obj)
                            (format nil "e=~a" (arg-user-answer obj))
                          (if (or (symbolp (arg-value obj))
                                  (numberp (arg-value obj)))
                              (format nil "~a" (arg-value obj))
                            "?")))))

(defgeneric arg-value-or-editable (arg))
(defmethod arg-value-or-editable ((arg argument))
  (if (arg-editable arg)
      (arg-user-answer arg)
    (arg-value arg)))
      

(defclass operator-arg (argument)
  ())

;; child-width is the number of columns the branch takes up.
(defclass box ()
  ((x :type number :initarg :x :initform 0 :accessor box-x)
   (y :type number :initarg :y :initform 0 :accessor box-y)
   (parents :type (or list box) :initform nil :initarg :parents :accessor box-parents)
   (pointed-to :type boolean :initarg :pointed-to :initform nil :accessor box-pointed-to)
   (nleaves :initarg :nleaves :initform nil :accessor box-nleaves)
   (depth :initarg :depth :initform nil :accessor box-depth)
   (id :initarg :id :initform (gentemp "BOX" "KEYWORD") :accessor box-id)
   (user-input :initform nil :accessor box-user-input)
   (answer-view :accessor box-answer-view :initform nil :documentation "The view associated with answer input.")
   (selected :type boolean :initarg :selected :initform nil :accessor box-selected)))

(defclass arg-box (box)
   ((op :type argument :initarg :op :accessor box-op)
    (arg1 :type argument :initarg :arg1 :accessor box-arg1)
    (arg2 :type argument :initarg :arg2 :accessor box-arg2)
    (lchild-width :initarg :lchild-width :initform nil :accessor box-lchild-width)))
   
(defclass equal-box (arg-box)
  ((op :type argument :initform (make-instance 'operator-arg :value '=) :accessor box-op)))

(defclass x-box (box)
  ((reposition :initarg :reposition :initform 0 :accessor x-box-reposition :documentation
	       "When t, the next call to draw-tree will reposition the box.")))

(defclass answer-box (box)
  ((answer :type argument :initarg :answer :accessor answer)))

(defclass number-box (box)
  ((number :type argument :initarg :number :accessor number)))

(defmethod print-object ((obj arg-box) stream)
  (print-unreadable-object (obj stream :type t :identity t)
    (format stream "~s ~s ~s" (box-arg1 obj) (box-op obj) (box-arg2 obj))))

(defmethod print-object ((obj answer-box) stream)
  (print-unreadable-object (obj stream :type t :identity t)
    (format stream "~s" (answer obj))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; (defun box-bottom (box)
;;   (+ (box-y box) +box-height+))

;; (defun box-right (box)
;;   (+ (box-x box) +box-width+))

(defun box-bc (box)
  "The bottom center of the box"
  (make-point (+ (box-x box) (truncate +box-width+ 2))
	      (+ (box-y box) +box-height+ +pad+)))

(defun unknown-tc (p)
  "the top center of an unknown box."
  p
;;   (make-point (+ (point-h p) (truncate +unknown-width+ 2))
;; 	      (point-v p))
  )

(defun text-center (p str)
  "Return a point that will center str at p"
  (let ((w (string-width str)))
    (make-point (- (point-h p) (truncate w 2))
		(point-v p))))

(defgeneric redraw-arg-answer (arg))

(defmethod redraw-arg-answer ((arg argument))
  (unless (null (arg-user-answer arg))
    (let* ((view (tutor-eqn-view *tutor*));(view-container (arg-widget arg)))
	   (txt (format nil "~a" (arg-user-answer arg)))
	   (x (- (arg-x arg) (calc-arg-offset arg txt)))
	   (y (- (arg-y arg) (truncate (line-height *box-font*) 2))))
      (when (arg-widget arg)
        (remove-subviews view (arg-widget arg)))
      (setf (arg-widget arg) ;; (draw-text-with-font view x y txt *box-font* `(:text ,*input-color* :body nil) t)
            (draw-transparent-text view x y txt *box-font* *input-color*)
            ))))

(defgeneric editables (box))

(defmethod editables ((box arg-box))
  (with-slots (arg1 arg2 op) box
    (listnn (when (arg-editable arg1) arg1)
            (when (arg-editable op)   op)
            (when (arg-editable arg2) arg2))))

(defmethod editables ((box answer-box))
  (listnn (when (arg-editable (answer box)) (answer box))))

(defmethod editables ((box number-box))
  nil)

(defmethod editables ((box x-box))
  nil)


(defmacro dolist-editables ((arg box) &body body)
  "Run BODY on the editable args in BOX. ARG is bound to each
  editable arg in turn."
  (let ((arg1 (gensym))
        (arg2 (gensym))
        (op (gensym)))
    `(with-slots (,arg1 ,arg2 ,op) ,box
       (dolist (,arg (listnn (when (arg-editable ,arg1) ,arg1)
                             (when (arg-editable ,op)   ,op)
                             (when (arg-editable ,arg2) ,arg2)))
         ,@body))))

(defun redraw-box-answer (box)
  "Used for arg boxes when filling in input."
  ;; The initial green box is removed once answers have been filled in.
  (let ((old-view (box-answer-view box)))
    (when (box-answer-view box)
      (remove-subviews (view-container (box-answer-view box)) (box-answer-view box))
      (setf (box-answer-view box) nil))
    (if (box-user-input box)
        (progn
          (setf (box-answer-view box) (draw-transparent-text (tutor-eqn-view *tutor*) (box-x box) (box-y box) (format nil "~a" (box-user-input box)) *box-font* *input-color*))
          (set-view-position (box-answer-view box)
                             (+ (box-x box) (truncate (- +box-width+ (view-width (box-answer-view box))) 2))
                             (+ (box-y box) (truncate (- +box-height+ (view-height (box-answer-view box))) 2))))
      (progn
        ;; Each arg draws its own argument, but we need a single view to
        ;; stick in box-answer-view so that the input window in
        ;; fmri-mode knows where to appear.
        (let ((editables (editables box)))
          (mapc #'redraw-arg-answer editables)
          (setf (box-answer-view box) (make-instance 'simple-view
                                                     :view-position (view-position old-view)
                                                     :view-container (tutor-eqn-view *tutor*)
                                                     :view-size (view-size old-view))))))))

;; (defgeneric respace-arg (arg x y))

;; (defmethod respace-arg ((arg argument) x y)
;;   (setf (arg-x arg) x
;; 	(arg-y arg) y))

(defun respace-box-args (box)
  ;;(format t "respace args ~s ~a ~a~%" box (box-x box) (box-y box))
  (etypecase box
    (x-box)
    (number-box
     (setf (arg-x (number box)) (+ (box-x box) (truncate +box-width+ 2))
           (arg-y (number box)) (+ (box-y box) (truncate +box-height+ 2))))
    (answer-box
     (setf (arg-x (answer box)) (+ (box-x box) (truncate +box-width+ 2))
           (arg-y (answer box)) (+ (box-y box) (truncate +box-height+ 2))))
    (equal-box
     (let ((x (+ (box-x box) (truncate +box-width+ 2)))
           (y (+ (box-y box) (truncate +box-height+ 2))))
       (setf (arg-x (box-arg1 box)) x
             (arg-y (box-arg1 box)) y
             (arg-x (box-arg2 box)) x
             (arg-y (box-arg2 box)) y)))
    (arg-box
     (setf (arg-x (box-op box)) (+ (box-x box) (truncate +box-width+ 2))
           (arg-y (box-op box)) (+ (box-y box) (truncate +box-height+ 2))

           (arg-x (box-arg1 box)) (+ (box-x box) (- (truncate (- +box-width+ (string-width "+" *box-font*)) 4) +arg-pad+))
           (arg-y (box-arg1 box)) (+ (box-y box) (truncate +box-height+ 2))
           (arg-align (box-arg1 box)) :right

           (arg-x (box-arg2 box)) (+ (box-x box) (- +box-width+ (- (truncate (- +box-width+ (string-width "+" *box-font*)) 4) +arg-pad+)))
           (arg-y (box-arg2 box)) (+ (box-y box) (truncate +box-height+ 2))
           (arg-align (box-arg2 box)) :left))))

(defun the-width-iwant ()
  (-
   ;; the start of the op
   (truncate (- +box-width+ (string-width "+" *box-font*)) 2)
   ;; middle of the arg
   (- (truncate (- +box-width+ (string-width "+" *box-font*)) 4) +arg-pad+)))


(defun calc-arg-offset (arg txt)
  (let* ((len (string-width txt *box-font*))
         (len2 (truncate len 2))
         (dist (the-width-iwant))
         (ofs (+ (case (arg-align arg)
                   (:left
                    (if (> len2 dist) (- dist len2) 0))
                   (:right 
                    (if (> len2 dist) (- len2 dist) 0))
                   (t 0))
                 (truncate len 2))))
    ofs))

(defun tutor1-draw-arg (arg view color)
;;   ;; If the arg is an unknown then we need to draw a line to the box
;;   (when (typep (arg-value arg) 'box)
;;     (let ((p1 (box-bc (arg-value arg)))
;; 	  (p2 (unknown-tc (make-point x (- y (truncate +unknown-height+ 2))))))
;;       (draw-line view (point-h p1) (point-v p1) (point-h p2) (point-v p2) *black-color*)))
  ;; Then draw the argument
  (let ((x (arg-x arg))
	(y (arg-y arg)))
    (cond 
     ((arg-editable arg)
      ;;     (setf (arg-widget arg) (make-instance 'editable-text-dialog-item
      ;; 					  :allow-returns t
      ;; 					  :view-container view
      ;; 					  :view-position (make-point (+ x 3) (+ (- y (truncate +unknown-height+ 2)) 3))
      ;; 					  :view-size (make-point (- +unknown-width+ 5) (- +unknown-height+ 5)))))
      (let* ((x1 (- x (truncate +unknown-width+ 2)))
	     (y1 (- y (truncate +unknown-height+ 2)))
	     (x2 (+ x1 +unknown-width+))
	     (y2 (+ y1 +unknown-height+)))
	;;       (draw-rect view x1 y1 x2 y2 *green-color*)
        (draw-rect view x1 y1 x2 y2 *input-box-color* t)
	(setf (arg-widget arg) (make-instance 'simple-view :view-container view :view-position (make-point x1 y1) :view-size (make-point (- x2 x1) (- y2 y1))))
	(register-click-region x1 y1 x2 y2 (arg-id arg) (lambda () 
                                                          (handle-fill-arg arg)))))
     ;; for boxes or nil we draw box
     ((or (typep (arg-value arg) 'box)
	  (null (arg-value arg)))
      (let ((x (- x (truncate +unknown-width+ 2)))
            (y (- y (truncate +unknown-height+ 2))))
	(draw-rect view x y (+ x +unknown-width+) (+ y +unknown-height+) color)))
     ;; otherwise draw the value
     (t
      (let* ((txt (format nil "~a" (arg-value arg)))
             (ofs (calc-arg-offset arg txt)))
	(draw-text-with-font view
		   (- x ofs)
		   (- y (truncate (line-height *box-font*) 2))
		   txt *box-font* color))))
    (event-dispatch)))

(defmacro make-clickable (container reg form register-form)
  (with-gensyms (view)
    `(let* ((,reg (make-instance 'region-view
                                 :view-position #@(0 0)
                                 :view-container ,container))
            (,view ,form))
       (set-view-size ,reg (view-size ,view))
       ,register-form
       reg)))

;;(defgeneric draw-arg (arg container parens))

;; (defun tutor2-draw-arg (arg container parens &optional selected)
;;   (if (arg-editable arg)
;;       (make-clickable container reg
;;                       (if (arg-user-answer arg)
;;                           (if (numberp (arg-user-answer arg))
;;                               (draw-number (arg-user-answer arg) reg *green-color*)
;;                             (draw-text-with-font reg 0 0 (format nil "~a" (arg-user-answer arg))
;;                                                  +eqn-font+ *green-color*))
;;                         (setf (arg-widget arg) (draw-rect reg 0 0 20 20 *green-color*)))
;;                       (register-click-view reg (arg-id arg) (lambda ()
;;                                                               (handle-fill-arg arg))))
;;     (draw-tree (arg-value arg) container parens selected)))

(defgeneric draw-box (box view))

(defmethod draw-box ((box arg-box) view)
  (macrolet ((greenbox (view box start-arg end-arg fn)
               (with-gensyms (x1 y1 x2 y2)
                 `(let ((,x1 (- (arg-x ,start-arg) (truncate +unknown-width+ 2)))
                        (,y1 (- (arg-y ,start-arg) (truncate +unknown-height+ 2)))
                        (,x2 (+ (arg-x ,end-arg) (truncate +unknown-width+ 2)))
                        (,y2 (+ (arg-y ,end-arg) (truncate +unknown-height+ 2))))
                    (draw-rect ,view ,x1 ,y1 ,x2 ,y2 *input-box-color* t)
                    (setf (box-answer-view ,box) (make-instance 'simple-view :view-container ,view :view-position (make-point ,x1 ,y1) :view-size (make-point (- ,x2 ,x1) (- ,y2 ,y1))))
                    (register-click-region ,x1 ,y1 ,x2 ,y2 (box-id box) (lambda () ,fn))))))
    ;; draw a big green box for the parts of the box that are editable.
    (with-slots (op arg1 arg2 x y selected) box
      (let ((color (if selected *selected-color* *foreground-color*)))
        (cond ((and (arg-editable arg1)
                    (arg-editable op)
                    (arg-editable arg2))
               (greenbox view box arg1 arg2 (handle-fill-box box)))
              ((and (arg-editable arg1)
                    (arg-editable op))
               (greenbox view box arg1 op (handle-fill-arg-op box)))
              ((and (arg-editable op)
                    (arg-editable arg2))
               (greenbox view box op arg2 (handle-fill-op-arg box)))
              (t
               (dolist (a (list arg1 op arg2))
                 (when (arg-editable a)
                   (tutor1-draw-arg a view color)))))
        ;; draw the uneditable args
        (dolist (a (list arg1 op arg2))
          (unless (arg-editable a)
            (tutor1-draw-arg a view color)))))))

;;             ;; Set x,y
;;             (dolist (a (list arg2 op arg1))
;;               (when (arg-editable a)
;;                 (setf x (arg-x a)
;;                       y (arg-y a))))
;;             ;; set mx, my
;;             (dolist (a (list arg1 op arg2))
;;               (when (arg-editable a)
;;                 (setf mx (+ (arg-x a) +unknown-width+)
;;                       my (+ (arg-y a) +unknown-height+))))
;;             (when x
;;               (decf y (truncate +unknown-height+ 2))
;;               (decf my (truncate +unknown-height+ 2))
;;               ;; Record the view so we can delete it later
;;               (setf (box-answer-views box) (list (draw-rect view x y mx my *green-color*)))
;;               (register-click-region x y mx my (box-id box) (lambda () 
;;                                                      (handle-fill-box box)))))

(defmethod draw-box ((box equal-box) view)
  ;; This needs to be a macro so we can create the proper lambda function.
  (with-slots (op arg1 arg2 x y selected) box
    ;; Place the arguments and operator
    (let ((color (if selected *selected-color* *black-color*)))
      (if (or (arg-editable arg1)
	      (and (not (arg-editable arg2))
		   (numberp (arg-value arg1))))
	  (tutor1-draw-arg arg1 view color)
	(tutor1-draw-arg arg2 view color)))))

(defun draw-right-angle (box view)
    ;; Draw a line, connecting the box to its parent.
    (let* ((p1 (box-bc box))
	   (args (box-containers box)))
      (dolist (a args)
	(let* ((p2 (unknown-tc (make-point (arg-x a) (- (arg-y a) (truncate +unknown-height+ 2)))))
	       (mid (+ (point-v p1) 31)))
          (if (= (point-v p2) mid)
              (draw-pointer view (point-h p1) (point-v p1)
                         (point-h p2) mid)
            (progn
              (draw-line view (point-h p1) (point-v p1)
                         (point-h p2) mid)
              (draw-pointer view (point-h p2) mid
                            (point-h p2) (point-v p2))))))))

(defun draw-direct (box view)
    ;; Draw a line, connecting the box to its parent.
    (let* ((p1 (box-bc box))
	   (args (box-containers box)))
      (dolist (a args)
	(let* ((p2 (unknown-tc (make-point (arg-x a) (- (arg-y a) (truncate +unknown-height+ 2))))))
	  (draw-pointer view (point-h p1) (point-v p1) 
                        (point-h p2) (point-v p2))))))

(defmethod draw-box :before ((box box) view)
  "draw the rectangle around the box."
  ;; Draw a box around them
  (with-slots (x y selected) box
    (let ((color (if selected *selected-color* *background-color*)))
      ;; this draws a thicker box when selected.
      (draw-rect view
                 (1- (+ (- x +pad+ 2))) (1- (+ (- y +pad+)))
                 (1+ (+ x +box-width+ +pad+ 2)) (1+ (+ y +box-height+ +pad+))
                 (if selected color *foreground-color*))
      (draw-rect view
		 (+ (- x +pad+ 2)) (+ (- y +pad+))
		 (+ x +box-width+ +pad+ 2) (+ y +box-height+ +pad+)
		 color))
    ;; draw the arrows
    (when (box-pointed-to box)
      (multiple-value-bind (x y) (box-left-side box)
        (draw-arrow (tutor-eqn-area-view *tutor*) x y 20 20 :right *selected-color*)))
    (if (typep box 'x-box)
        (draw-right-angle box view)
      (draw-direct box view))))

(defmethod draw-box ((box x-box) view)
  "draw the X box."
  ;; Its just an empty box
  (declare (ignore view)))
	   
(defmethod draw-box ((box answer-box) view)
  "An answer box has an unknown sitting in the middle."
  (with-slots (answer x y) box
    (tutor1-draw-arg answer view *foreground-color*)))

(defmethod draw-box ((box number-box) view)
  (with-slots (number x y selected) box
    (let ((color (if selected *selected-color* *foreground-color*)))
      (tutor1-draw-arg number view color))))

(defgeneric children (box))

(defmethod children ((box arg-box))
  (list (arg-value (box-arg1 box))
	(arg-value (box-arg2 box))))

(defmethod children ((box answer-box))
  (list (arg-value (answer box))))

(defmethod children ((box number-box))
  (list (arg-value (number box))))

(defmethod children ((box x-box))
  nil)

(defmethod children ((box number))
  nil)

(defun box-args (box)
  (etypecase box
    (arg-box
     (list (box-arg1 box) (box-arg2 box)))
    (x-box nil)
    (number-box (number box))
    (answer-box (answer box))))

;;;; tree-nleaves
  
(defgeneric tree-nleaves (tree)
  (:documentation "return the number of leaf boxes in the tree."))

(defmethod tree-nleaves ((tree x-box))
  1)

(defmethod tree-nleaves ((tree answer-box))
  (max 1 (tree-nleaves (arg-value (answer tree)))))

(defmethod tree-nleaves ((tree arg-box))
  (max 1 (+ (tree-nleaves (arg-value (box-arg1 tree)))
	    (tree-nleaves (arg-value (box-arg2 tree))))))

(defmethod tree-nleaves ((tree number-box))
  (max 1 (tree-nleaves (arg-value (number tree)))))

(defmethod tree-nleaves ((tree number))
  0)

(defmethod tree-nleaves ((a argument))
  (tree-nleaves (arg-value a)))

;;;; tree-depth

(defgeneric tree-depth (tree))

(defmethod tree-depth ((tree x-box))
  1)

(defmethod tree-depth ((tree answer-box))
  (1+ (tree-depth (arg-value (answer tree)))))

(defmethod tree-depth ((tree number-box))
  (1+ (tree-depth (arg-value (number tree)))))

(defmethod tree-depth ((tree arg-box))
  (1+ (max (tree-depth (arg-value (box-arg1 tree)))
	   (tree-depth (arg-value (box-arg2 tree))))))

(defmethod tree-depth ((tree number))
  0)

(defun other-which (which)
  (if (eq which 'box-arg1)
      'box-arg2
    'box-arg1))

(defmacro def-ms (name fn arg amount)
  (with-gensyms (subtree amt b)
    `(defun ,name (,subtree &optional (,amt ,amount))
       (move-subtree-subr (lambda (,b)
                            (,fn (,arg ,b) ,amt))
                          ,subtree))))

(def-ms move-subtree-up    decf box-y +level-height+)
(def-ms move-subtree-down  incf box-y +level-height+)
(def-ms move-subtree-left  decf box-x +node-width+)
(def-ms move-subtree-right incf box-x +node-width+)

(defun move-subtree-subr (subr subtree)
  (when (typep subtree 'box)
    (funcall subr subtree)
    (respace-box-args subtree))
  (mapc (lambda (tree) (move-subtree-subr subr tree)) (only-boxes (children subtree))))

;; How these two functions Do The Right Thing is beyond me.
(defun right-align-subtree (subtree)
  (move-subtree-left subtree (- (abs (eqn-right subtree t)) (truncate +node-width+ 2))))

(defun left-align-subtree (subtree)
  (move-subtree-right subtree (+ (abs (eqn-left subtree t)) (truncate +node-width+ 2))))

(defun all-box (box)
  (etypecase box
    (number-box t)
    (x-box t)
    (arg-box (and (typep (arg-value (box-arg1 box)) 'box)
                  (typep (arg-value (box-arg2 box)) 'box)))
    (answer-box (typep (arg-value (answer box)) 'box))))

(defun all-not-box (box)
  (etypecase box
    (number-box nil)
    (x-box nil)
    (arg-box (not (or (typep (arg-value (box-arg1 box)) 'box)
                      (typep (arg-value (box-arg2 box)) 'box))))
    (answer-box (not (typep (arg-value (answer box)) 'box)))))

(defun one-num-one-box (tree child1 child2)
  (and (numberp (funcall child1 tree))
       (and (typep (funcall child2 tree) 'box)
            ;; Either its not an x-box or it is but it
            ;; has only 1 parent (so it doesn't float)
            (or (not (typep (funcall child2 tree) 'x-box))
                (and (typep (funcall child2 tree) 'x-box)
                     (= (length (box-parents (funcall child2 tree))) 1))))))

(defun box-not-floating (box)
  (and (typep box 'box)
       (or (not (typep box 'x-box))
           (and (typep box 'x-box)
                (= (length (box-parents box)) 1)))))

(defun uber-respace-tree (subtree)
  (new-respace-tree subtree)
  ;; handle the x box placement
  (let ((tree (tree-x-box subtree)))
    ;; Check for the x-box's parents. because if it doesn't have any
    ;; its about to be axed by a 0 * X evaluation.
    (when (and tree 
               (box-parents tree))
      (let* ((minx (apply #'min (mapcar #'box-x (box-parents tree))))
             (maxx (apply #'max (mapcar (lambda (b) (+ (box-x b) +box-width+)) (box-parents tree))))
             (miny (apply #'min (mapcar #'box-y (box-parents tree))))
             (x (+ minx (truncate (- maxx minx) 2) (- (truncate +box-width+ 2)))))
        (setf (box-y tree) (- miny +level-height+)
              (box-x tree) x))
      ;;(setf (box-y tree) (eqn-top subtree))
      ;; this is a hack to make sure the x-box doesn't draw ontop of its
      ;; sibling and their input arrows don't cross.
      (when (loop for b in (remove tree (only-boxes (apply #'nconc (mapcar 'children (box-parents tree)))))
                  thereis (or 
                           ;; do they overlap?
                           (and (not (or (> (box-x tree) (+ (box-x b) +box-width+))
                                         (< (+ (box-x tree) +box-width+) (box-x b))))
                                (= (box-y tree) (box-y b)))
                           ;; do their arrows cross?
                           (and (< (box-x tree) (box-x b))
                                (box-left-child-p b)
                                (> (length (box-parents tree)) 1)
                                (= (box-y tree) (box-y b)))))
        (decf (box-y tree) +level-height+)))))

(defun new-respace-tree (subtree)
  ;; FIXME: this maybe doesn't need to be done.
  (when (typep subtree 'box)
    (let* ((nleaves (tree-nleaves subtree))
           (depth (tree-depth subtree)))
      (setf (box-nleaves subtree) nleaves)
      (setf (box-depth subtree) depth)
      (setf (box-x subtree) 0
            (box-y subtree) 0)))
  (etypecase subtree
    (number )
    (number-box
     (respace-box-args subtree))
    (x-box )
    (arg-box
     (new-respace-tree (box-left subtree))
     (new-respace-tree (box-right subtree))
     (cond ((all-not-box subtree)       ; 1+1
;;             (dformat "here0~%")
            )
           ((one-num-one-box subtree 'box-left 'box-right)
;;             (dformat "here1~%")
            (move-subtree-up (box-right subtree)))
           ((one-num-one-box subtree 'box-right 'box-left)
;;             (dformat "here2~%")
            (move-subtree-up (box-left subtree)))
           ;; both boxes, one is all boxes the other is all numbers
           ((and (box-not-floating (box-left subtree))
                 (box-not-floating (box-right subtree))
                 (or (and (all-not-box (box-left subtree))
                          (not (all-not-box (box-right subtree))))
                     (and (not (all-not-box (box-left subtree)))
                          (all-not-box (box-right subtree)))))
;;             (dformat "here3~%")
            (move-subtree-up (box-left subtree))
            (move-subtree-up (box-right subtree))
            (if (all-not-box (box-left subtree))
                (move-subtree-left (box-left subtree))
              (move-subtree-right (box-right subtree)))
            )
           ;; both boxes
           ((and (box-not-floating (box-left subtree))
                 (box-not-floating (box-right subtree)))
;;             (dformat "here4~%")
            (move-subtree-up (box-left subtree))
            (move-subtree-up (box-right subtree))
            (right-align-subtree (box-left subtree))
            (left-align-subtree (box-right subtree)))
           ;; one box is the X box, and the X box has >1 connection
           ((or (typep (box-left subtree) 'x-box)
                (typep (box-right subtree) 'x-box))
            (if (typep (box-left subtree) 'x-box)
                (when (typep (box-right subtree) 'box)
                  (move-subtree-up (box-right subtree))
                  (left-align-subtree (box-right subtree)))
              (when (typep (box-left subtree) 'box)
                (move-subtree-up (box-left subtree))
                (right-align-subtree (box-left subtree)))))
           (t (error "oops!")))
     (respace-box-args subtree)
     )
    (answer-box
;;      (dformat "here5~%")
     (new-respace-tree (arg-value (answer subtree)))
     (respace-box-args subtree)
     (move-subtree-up (arg-value (answer subtree))))))
       
;; (defun tree-x-child-depth (tree which)
;;   "the number of boxes down the left/right side with 2
;;   children (depending on WHICH) whose right/left child isn't an all-not-box box."
;;   (labels ((one-num-one-box (tree child1 child2)
;;                             (and (numberp (arg-value (funcall child1 tree)))
;;                                  (and (typep (arg-value (funcall child2 tree)) 'box)
;;                                       ;; Either its not an x-box or it is but it
;;                                       ;; has only 1 parent (so it doesn't float)
;;                                       (or (not (typep (arg-value (funcall child2 tree)) 'x-box))
;;                                           (and (typep (arg-value (funcall child2 tree)) 'x-box)
;;                                                (= (length (box-parents (arg-value (funcall child2 tree)))) 1))))))
;;            (box-not-floating (box)
;;                              (and (typep box 'box)
;;                                   (or (not (typep box 'x-box))
;;                                       (and (typep box 'x-box)
;;                                            (= (length (box-parents box)) 1))))))
;;     (cond ((numberp tree)
;;          0)
;;         ((typep tree 'arg-box)
;; ;;          (cond ((all-not-box tree)      ; 1+1
;; ;;                 1)
;; ;;                ((one-num-one-box tree 'box-arg1 'box-arg2)
;; ;;                 ;; it will be at least 1 and the same width as it's box child
;; ;;                 (max 1 (tree-x-child-depth (arg-value (box-arg2 tree)))))
;; ;;                ((one-num-one-box tree 'box-arg2 'box-arg1)
;; ;;                 ;; it will be at least 1 and the same width as it's box child
;; ;;                 (max 1 (tree-x-child-depth (arg-value (box-arg1 tree)))))
;; ;;                ;; both boxes, one is all boxes
;; ;;                ((and (box-not-floating (box-left tree))
;; ;;                      (box-not-floating (box-right tree))
;; ;;                      (or (and (all-not-box (box-left tree))
;; ;;                               (not (all-not-box (box-right tree))))
;; ;;                          (and (not (all-not-box (box-left tree)))
;; ;;                               (all-not-box (box-right tree)))))
;; ;;                 (1+ (max 1 (tree-x-child-depth (if (all-not-box (box-left tree))
;; ;;                                                    (box-right tree)
;; ;;                                                  (box-left tree))))))
;; ;;                ;; both boxes
;; ;;                ((and (box-not-floating (box-left tree))
;; ;;                      (box-not-floating (box-right tree)))
;; ;;                 (max 2 (+ (tree-x-child-depth (box-left tree))
;; ;;                           (tree-x-child-depth (box-right tree)))))
               
                     

;;          (let ((c (only-boxes (children tree))))
;;            (ecase (length c)
;;              (0 1)
;;              (1 (max 1 (tree-x-child-depth (first c) which)))
;;              (2 (cond ((all-not-box (arg-value (funcall which tree)))
;;                        3)
;;                       ((and (all-box (arg-value (box-arg1 tree)))
;;                             (all-box (arg-value (box-arg2 tree))))
;;                        (1+ (tree-x-child-depth (arg-value (funcall which tree)) which)))
;;                       ;; this is pretty gross.
;;                       (t (if (eq which 'box-arg1)
;;                              (max 1 (tree-x-child-depth (arg-value (funcall which tree)) which))
;;                            (+ 1 (tree-x-child-depth (arg-value (funcall which tree)) which))))))))
;;          )
;; ;;         ((typep tree 'equal-box)
;; ;;          (tree-left-child-depth (arg-value (box-arg1 tree))))
;; ;;         ((typep tree 'answer
;;         ((typep tree 'number-box)
;;          1)
;;         ((typep tree 'x-box)
;;          (if (= (length (box-parents tree)) 1)
;;              1
;;            0))
;;         (t (error "you missed a box type ~a." (type-of tree))))))

;; (defun tree-left-child-depth (tree)
;;   "the number of boxes down the left side with 2 children"
;;   (tree-x-child-depth tree 'box-arg1))

;; (defun tree-right-child-depth (tree)
;;   "the number of boxes down the left side with 2 children"
;;   (tree-x-child-depth tree 'box-arg2))

(defun make-box (op arg1 arg2 parent &optional op-answer arg1-answer arg2-answer)
  (if (eq op '=)
      (make-instance 'equal-box
		     :parents (when parent (list parent))
		     :arg1 (make-instance 'argument :value arg1 :editable (true-p arg1-answer) :answer arg1-answer)
		     :arg2 (make-instance 'argument :value arg2 :editable (true-p arg2-answer) :answer arg2-answer))		   
    (make-instance 'arg-box
		   :parents (when parent (list parent))
		   :op (make-instance 'operator-arg :value op :editable (true-p op-answer) :answer op-answer)
		   :arg1 (make-instance 'argument :value arg1 :editable (true-p arg1-answer) :answer arg1-answer)
		   :arg2 (make-instance 'argument :value arg2 :editable (true-p arg2-answer) :answer arg2-answer))))

(defun make-tree (tree)
  "Create a tree made of boxes."
  (let ((x-box (make-instance 'x-box)))
    (labels ((make (tree parent)
	       (cond ((null tree) nil)
		     ((numberp tree) tree)
		     ((symbolp tree)
		      (push parent (box-parents x-box))
		      x-box)
		      ((and (listp tree) (= (length tree) 1))
		       ;; This is a special case. its a number box.
		       (let ((nb (make-instance 'number-box
						:parents (list parent)
						:number (make-instance 'argument))))
			 ;; Sorta hacky. How else can we do it though?
			 (setf (arg-value (number nb)) (make (first tree) nb))
			 nb))
		      (t
		       (let ((b (make-box (first tree) nil nil parent)))
			 (setf (arg-value (box-arg1 b)) (make (second tree) b)
			       (arg-value (box-arg2 b)) (make (third tree) b))
			 b)))))
	     (if (and (listp tree)
		      (eq '= (first tree)))
		 (make tree nil)
	       ;; If it isn't an = box then make an answer box
	       (let ((a (make-instance 'answer-box
				       :parents nil
				       :answer (make-instance 'argument))))
		 (setf (arg-value (answer a)) (make tree a))
		 a)))))

;; This is sorta sucky duplicate code from tree-nleaves
(defun tree-selected-leaves (root)
  (cond ((not (typep root 'box)) 0)
	((and (or (not (typep (arg-value (box-arg1 root)) 'box))
		  (not (box-selected (arg-value (box-arg1 root)))))
	      (or (not (typep (arg-value (box-arg2 root)) 'box))
		  (not (box-selected (arg-value (box-arg2 root))))))
	 1)
	(t 
	 (+ (tree-nleaves (arg-value (box-arg1 root)))
	    (tree-nleaves (arg-value (box-arg2 root)))))))

;;;; drawing

;; (defgeneric respace-tree (tree redge ledge respace-p depth &optional ofs-x ofs-y)
;;   (:documentation "Create room for new boxes, if needed."))

;; (defmethod respace-tree ((tree number) redge ledge respace-p depth &optional (ofs-x 0) (ofs-y 0))
;;   (declare (ignore tree redge ledge respace-p depth ofs-x ofs-y))
;;   t)

;; (defmethod respace-tree ((tree box) redge ledge respace-p depth &optional (ofs-x 0) (ofs-y 0))
;;   "This is the common respacing that all boxes do."
;;   (declare (ignore redge ledge respace-p depth ofs-x ofs-y))
;;   (let* ((nleaves (tree-nleaves tree))
;; 	 (depth (tree-depth tree)))
;;     (setf (box-nleaves tree) nleaves)
;;     (setf (box-depth tree) depth)))

;; (defmethod respace-tree :around ((tree number-box) rightp)
;;   (call-next-method)
;;   ;; Do the rest
;;   (respace-tree (arg-value (number tree)) rightp))
  
;; (defmethod respace-tree :after ((tree x-box) redge ledge respace-p depth &optional (ofs-x 0) (ofs-y 0))
;; ;;   (if (x-box-reposition tree)
;;   (declare (ignore depth))
;;   (if (= (length (box-parents tree)) 1)
;;       (let ((x (if respace-p
;; 		   (if redge
;; 		       (- redge (truncate +node-width+ 2))
;; 		     (+ ledge (truncate +node-width+ 2)))
;; 		 (or redge ledge))))
;; 	(setf (box-x tree) (+ ofs-x x)))
;;     (let* ((min (apply #'min (mapcar #'box-x (box-parents tree))))
;; 	   (max (apply #'max (mapcar (lambda (b) (+ (box-x b) +box-width+)) (box-parents tree))))
;; 	   (x (+ min (truncate (- max min) 2) (- (truncate +box-width+ 2)))))
;;       (setf (box-x tree) x)))		;(+ ofs-x x)
;;   (setf (box-y tree) (+ ofs-y (* (box-depth tree) +level-height+)))
;;   ;; this is a hack to make sure the x-box doesn't draw ontop of its
;;   ;; sibling and their input arrows don't cross.
;;   (when (loop for b in (remove tree (only-boxes (apply #'nconc (mapcar 'children (box-parents tree)))))
;;               thereis (or 
;;                        ;; do they overlap?
;;                        (and (not (or (> (box-x tree) (+ (box-x b) +box-width+))
;;                                      (< (+ (box-x tree) +box-width+) (box-x b))))
;;                             (= (box-y tree) (box-y b)))
;;                        ;; do their arrows cross?
;;                        (and (< (box-x tree) (box-x b))
;;                             (box-left-child-p b)
;;                             (> (length (box-parents tree)) 1)
;;                             (= (box-y tree) (box-y b)))))
;;               (decf (box-y tree) +level-height+)))
;; 	  (x-box-reposition tree) nil))))

;; (defmethod respace-tree :around ((tree answer-box) redge ledge respace-p depth &optional (ofs-x 0) (ofs-y 0))
;;   (declare (ignore respace-p))
;;   (call-next-method)
;;   (let ((x (or redge ledge)))
;;     (setf (box-x tree) (+ ofs-x x)
;; 	  (box-y tree) (+ ofs-y (* depth +level-height+)))
;;     (respace-arg (answer tree) 
;; 		 (+ (box-x tree) (- (truncate +unknown-width+ 2)) (truncate +box-width+ 2))
;; 		 (+ (box-y tree) (truncate +box-height+ 2)))
;;     (respace-tree (arg-value (answer tree)) redge ledge nil (1- depth) ofs-x ofs-y)))

;; (defmethod respace-tree :after ((tree number-box) redge ledge respace-p depth &optional (ofs-x 0) (ofs-y 0))
;;   (declare (ignore respace-p))
;;   (let ((x (or redge ledge)))
;;   (setf (box-x tree) (+ ofs-x x)
;; 	(box-y tree) (+ ofs-y (* depth +level-height+)))
;;   (respace-arg (number tree)
;; 	       (+ (box-x tree) (- (truncate +unknown-width+ 2)) (truncate +box-width+ 2))
;; 	       (+ (box-y tree) (truncate +box-height+ 2)))))

;; (defun can-shrink-a-bit (box)
;;   "return T if BOX's sibling has no childen. If so, we can safely
;;   draw boxes above it and save space."
;;   ;; its always 1 parent
;;   (let* ((parents (box-parents box))
;;          (other (first (remove box (children (first parents))))))
;;     (null (only-boxes (children other)))))


(defun contains-one-all-number-box (parent)
  "Return true if the children of parent are boxes and only one has all numbers."
  (when (and (typep parent 'arg-box)
             (typep (arg-value (box-arg1 parent)) 'box)
             (typep (arg-value (box-arg2 parent)) 'box))
    (cond ((and (not (all-not-box (arg-value (box-arg1 parent))))
                (all-not-box (arg-value (box-arg2 parent))))
           (arg-value (box-arg2 parent)))
          ((and (not (all-not-box (arg-value (box-arg2 parent))))
                (all-not-box (arg-value (box-arg1 parent))))
           (arg-value (box-arg1 parent))))))


;; (defmethod respace-tree :around ((box arg-box) redge ledge respace-p depth &optional (ofs-x 0) (ofs-y 0))
;;   (call-next-method)
;;     (let* ((nchildren (tree-nleaves (arg-value (box-arg1 box))))
;; 	   (x (if respace-p
;; 		  (if redge
;; 		      (- redge (* (truncate (* (tree-right-child-depth box) +node-width+) 2)))
;;                     (+ ledge 
;;                        (truncate +node-width+ 2)
;;                        (truncate (* (1- (tree-left-child-depth box)) +node-width+) 2)))
;; 		(or ledge redge))))
;;       (setf (box-lchild-width box) nchildren
;; 	    (box-x box) (+ ofs-x x)
;; 	    (box-y box) (+ ofs-y (* depth +level-height+)))
;;       (with-slots (op arg1 arg2 x y) box
;; 		  ;; This is gross.
;; 		  (if (typep box 'equal-box)
;; 		      (progn
;; 			(let ((x (+ x (- (truncate +unknown-width+ 2)) (truncate +box-width+ 2)))
;; 			      (y (+ y (truncate +box-height+ 2))))
;; 			  (respace-arg arg1 x y)
;; 			  (respace-arg arg2 x y)))
;; 		    (progn
;; 		      (respace-arg arg1 x (+ y (truncate +box-height+ 2)))
;; 		      (respace-arg op (+ x (- (truncate +unknown-width+ 2)) (truncate +box-width+ 2)) (+ y (truncate +box-height+ 2)))
;; 		      (respace-arg arg2 (+ x (- +box-width+ +unknown-width+)) (+ y (truncate +box-height+ 2))))))
;;       ;; Do the rest
;;       (let ((b (contains-one-all-number-box box)))
;;         (if b
;;             (if (eq b (arg-value (box-arg1 box)))
;;                 (progn
;;                   (respace-tree (arg-value (box-arg1 box)) (- x +node-width+) nil nil (1- depth) ofs-x ofs-y)
;;                   (respace-tree (arg-value (box-arg2 box)) nil x nil (1- depth) ofs-x ofs-y))
;;               (progn
;;                 (respace-tree (arg-value (box-arg1 box)) nil x nil (1- depth) ofs-x ofs-y)
;;                 (respace-tree (arg-value (box-arg2 box)) (+ x +node-width+) nil nil (1- depth) ofs-x ofs-y)))
;;           (progn
;;             (respace-tree (arg-value (box-arg1 box)) x nil (> (length (only-boxes (children box))) 1) (1- depth) ofs-x ofs-y)
;;             (respace-tree (arg-value (box-arg2 box)) nil x (> (length (only-boxes (children box))) 1) (1- depth) ofs-x ofs-y))))))

(defun tutor1-redraw-tree (tree view)
  (draw-box tree view)
  (dolist (b (only-boxes (children tree)))
    (tutor1-redraw-tree b view)))

(defvar *views-with-arrows* nil)
(defvar *hilite-components* nil
  "Set this to true to draw hilited boxes.")

(defclass tutor2-eqn-view (view)
  ((root :initform nil :accessor eqn-view-root))
  (:documentation "A special view that points to the equation that it contains. Used by the actr-pm stuff."))

(defun tutor2-redraw-tree (tree view arrows-p hilite-p)
  ;;(draw-box tree view)
  (setf *views-with-arrows* nil
        *hilite-components* hilite-p)
  (let ((v ;; (draw-tree tree view nil)
         (render-compiled-eqn view (compile-eqn tree nil nil nil nil))
           ))
    (when (and *views-with-arrows*
               arrows-p)
      (set-view-size v (view-width v) (+ (view-height v) 22))
      (draw-arrows v))
    (setf (eqn-view-root v) tree)
    v))

;;;;;; tutor2 helper functions
;;;;;;

(defun do-text-boxes (reg &rest txt)
  (let* ((v (make-instance 'view :view-position #@(0 0) :view-container reg))
         (blah (loop for i in txt
                   collect (if (numberp i)
                               (draw-number i reg 0 0 nil nil nil `(:text ,*input-color* :body ,*input-box-color*) t)
                             (draw-text-with-font v 0 0 (format nil "~a" i) +eqn-font+ 
                                                  `(:text ,*input-color* :body ,*input-box-color*) t)))))
    (order-views blah)
    (set-view-size v (apply '+ (mapcar 'view-width blah))
                   (apply 'max (mapcar 'view-height blah)))
    v))

;; (defparameter +match-thingy+ '(:matches (:arg :op (:matches (+ -) :name op1)
;;                                               :arg1 (:matches (:arg :op (:matches (+ -) :name op3)
;;                                                                     :arg1 (:matches (:arg :op (:matches (*) :name op2)
;;                                                                                           :arg1 (:matches (:x) :name x)
;;                                                                                           :arg2 (:matches (:or :number :box) :name n1)
;;                                                                                           :permute t)
;;                                                                                     :name b3)
;;                                                                     :arg2 (:matches :number :name nx)
;;                                                                     :permute t)
;;                                                               :name bx)
;;                                               :arg2 (:matches (:arg :op (:matches (*) :name op2)
;;                                                                     :arg1 (:matches (:x) :name x)
;;                                                                     :arg2 (:matches :number :name n2)
;;                                                                     :permute t) :name b2)
;;                                               :permute nil)
;;                                         :name b1)
;;   "(nX +/- 4) +/- mX")

(defun special-x-box (box)
  (with-slots (op arg1 arg2) box
    (and *special-x-box*
         (typep (arg-value arg2) 'x-box)
         (eq (arg-value op) '*)
         (numberp (arg-value arg1)))))

;; (defun draw-box-args (box view parens)
;;   (with-slots (op arg1 arg2 x y selected) box
;;     (cond ((and (arg-editable arg1)
;;                 (arg-editable op)
;;                 (arg-editable arg2))
;;            ;;(dformat "11111~%")
;;            (list 
;;             (make-clickable view reg
;;                             (cond ((box-user-input box)
;;                                    (do-text-boxes reg (box-user-input box)))
;;                                    ((and (arg-user-answer arg1)
;;                                          (arg-user-answer op)
;;                                          (arg-user-answer arg2))
;;                                     (do-text-boxes reg (arg-user-answer arg1)
;;                                                    (arg-user-answer op)
;;                                                    (arg-user-answer arg2)))
;;                                    (t
;;                                     (setf (box-answer-view box) (draw-rect reg 0 0 60 20 *green-color*))))
;;                             (register-click-view reg (box-id box) (lambda () (handle-fill-box box))))))
;;           ((and (arg-editable arg1)
;;                 (arg-editable op))
;;            ;;(dformat "aaaa~%")
;;            (list
;;             (make-clickable view reg
;;                             (if (and (arg-user-answer arg1)
;;                                      (arg-user-answer op))
;;                                 (do-text-boxes reg (arg-user-answer arg1)
;;                                                (arg-user-answer op))
;;                               (setf (box-answer-view box) (draw-rect reg 0 0 40 20 *green-color*)))
;;                             (register-click-view reg (box-id box) (lambda () (handle-fill-arg-op box))))
;;             (tutor2-draw-arg arg2 view parens)))
;;           ((and (arg-editable op)
;;                 (arg-editable arg2))
;;            ;;(dformat "bbbb~%")
;;            (list
;;             (tutor2-draw-arg arg1 view parens)
;;             (make-clickable view reg
;;                             (if (and (arg-user-answer op)
;;                                      (arg-user-answer arg2))
;;                                 (do-text-boxes reg (arg-user-answer op)
;;                                                (arg-user-answer arg2))
;;                               (setf (box-answer-view box) (draw-rect reg 0 0 40 20 *green-color*)))
;;                             (register-click-view reg (box-id box) (lambda () (handle-fill-op-arg box))))))
;;           (t
;;            ;;(dformat "HEE~%")
;;            ;;(dformat "so? ~a" (looking-at box +match-thingy+))
;;            (cond ((simple-match box '(+ (* :n :x) (:a :n (* :n :x))))
;;                   (list (tutor2-draw-arg arg1 view parens selected)
;;                         (tutor2-draw-arg op view parens selected)
;;                         (tutor2-draw-arg arg2 view nil selected)))
;;                  ;; ((or (simple-match box '(:a (:a (* :n :x) :n) (:a (* :n :x) :n)))
;; ;;                       (simple-match box '(:a (:a (* :n :x) :n) (:a :n (* :n :x))))
;; ;;                       (simple-match box '(:a (:a :n (* :n :x)) (:a :n (* :n :x)))))
;; ;;                   (list (tutor2-draw-arg arg1 view nil selected)
;; ;;                         (tutor2-draw-arg op view parens selected)
;; ;;                         (tutor2-draw-arg arg2 view nil selected)))
;;                  ((simple-match box '(:a (:a (* :n :x) :n) (* :n :x)))
;;                   (list (tutor2-draw-arg arg1 view nil selected)
;;                         (tutor2-draw-arg op view parens selected)
;;                         (tutor2-draw-arg arg2 view parens selected)))
;;                  ((special-x-box box)
;;                   (let* ((r (make-instance 'region-view
;;                                            :view-container view))
;;                          (v (tutor2-draw-arg arg1 r nil selected))
;;                          (x (tutor2-draw-arg arg2 r nil selected)))
;;                     (order-views (list v x))
;;                     (resize-view-for r (list v x))
;;                     (list r)))
;;                  (t
;;                   (loop for a in (list arg1 op arg2)
;;                         collect (tutor2-draw-arg a view parens selected))))))))
  
(defun flip-paren (paren)
  (when paren
    (if (eq paren :bracket)
        :paren
      :bracket)))

(defun paren-string (paren open-p)
  (if (eq paren :bracket)
      (if open-p "[" "]")
    (if open-p "(" ")")))

(defun draw-arrows (root-view)
  (dolist (i *views-with-arrows*)
    (let ((p (convert-coordinates (make-point (truncate (view-width i) 2)
                                     (view-height i))
                         i root-view)))
      (draw-vert-arrow root-view (- (point-h p) 12) (point-v p) 20 20 *selected-color*))))

;; (defun draw-arrows (view root-view)
;;   "Return a list of arrows"
;;   (do-subviews (v view)
;;     (if (or (typep view 'region-view)
;;             (typep view 'alignment-view))
;;         (draw-vert-arrow root-view (convert-coordinates (make-point (- (view-x v) 5)
;;                                                                     (view-height v))
;;                                                         view root-view))
;;       (draw-arrows (v root-view)))))

;; (defgeneric draw-tree (box container parens &optional selected)
;;   (:documentation "draw it."))

;; (defmethod draw-tree ((box arg-box) container parens &optional selected)
;;   (declare (ignore selected))
;;   (let* ((color *black-color*)
;;          (selected (box-selected box))
;;          (view (make-instance 'view
;;                               :view-container container
;;                               :view-size #@(0 0)
;;                               :view-position #@(0 0)))
;;          (lb (when (and (not (typep box 'equal-box))
;;                         (not (special-x-box box))
;;                         parens)
;;                         ;;(not (eq (arg-value (box-op box)) '*)))
;;                (if selected
;;                    (draw-text-with-font view 0 0 
;;                                         (paren-string parens t) 
;;                                         +eqn-font+
;;                                         +selected-part-color-list+ t)
;;                  (draw-hilite-text view 0 0 (paren-string parens t) +eqn-font+))
;; ;;                (make-instance 'static-text-dialog-item
;; ;;                               :view-font +eqn-font+
;; ;;                               :view-position #@(0 0)
;; ;;                               :view-container view
;; ;;                               :part-color-list (if selected
;; ;;                                                   +selected-part-color-list+
;; ;;                                                 `(:text ,*black-color*))
;; ;;                               :dialog-item-text (paren-string parens t))
;;                ))
;;          (rb (when (and (not (typep box 'equal-box))
;;                         (not (special-x-box box))
;;                         parens)
;;                         ;;(not (eq (arg-value (box-op box)) '*)))
;;                (if selected
;;                    (draw-text-with-font view 0 0 
;;                                         (paren-string parens nil) 
;;                                         +eqn-font+
;;                                         +selected-part-color-list+ t)
;;                  (draw-hilite-text view 0 0 (paren-string parens nil) +eqn-font+))
;; ;;                (make-instance 'static-text-dialog-item
;; ;;                               :view-font +eqn-font+
;; ;;                               :view-position #@(0 0)
;; ;;                               :view-container view
;; ;;                               :part-color-list (if selected
;; ;;                                                    +selected-part-color-list+
;; ;;                                                  `(:text ,*black-color*))
;; ;;                               :dialog-item-text (paren-string parens nil))
;;                ))
;;                ;;(draw-text-with-font view 0 0 (paren-string parens nil) +eqn-font+ color)))
;;          (views (nconc1 lb (draw-box-args box view (not (typep box 'equal-box))) rb)))
;;     ;; now align the args in VIEW
;;     (resize-view-for view views)
;;     (order-views views)
;; ;;     (when (box-selected box)
;; ;;       (draw-line view 0 (1- (view-height view)) (view-width view) (1- (view-height view)) *selected-color*))
;;     ;; register the op so we can select the box.
;;     (when (not (tutor-submit-enabled *tutor*))
;;       (let ((reg (find-if (lambda (i)
;;                             (typep i 'region-view))
;;                           views)))
;;         (when reg
;;           (register-click-view reg (box-id box) (lambda () (click-box box)))
;;           ;; draw an arrow
;;           (when (box-pointed-to box)
;;             (push reg *views-with-arrows*)))
;; ;;             (set-view-size view (view-width view) (max (view-height view) 
;; ;;                                                        (+ (view-height view) 20)))
;; ;;             (draw-vert-arrow view (- (view-x reg) 5) (view-height reg) 20 20))
;;           ))
;;     view))

;; (defmethod draw-tree ((box number) container parens &optional selected)
;;   (declare (ignore parens))
;;   (if selected 
;;       (draw-number box container +selected-part-color-list+ t)
;;     (draw-number box container *black-color*)))

;; (defmethod draw-tree ((box symbol) container parens &optional selected)
;;   (declare (ignore parens))
;;   (let* ((region (make-instance (if (eq box '=)
;; 				    'alignment-view
;; 				  'region-view)
;;                                 :view-container container))
;;          (txt (make-instance 'static-text-dialog-item
;;                              :view-font +eqn-font+
;;                              :view-position #@(0 0)
;;                              :view-container region
;;                              :part-color-list (if selected
;;                                                   +selected-part-color-list+
;;                                                 `(:text ,*black-color*))
;;                              :dialog-item-text (prin1-to-string box))))
;;     ;;(dformat "GOOO~%")
;;     (set-view-size region (view-width txt) (view-height txt))
;;     region))

;; (defmethod draw-tree ((box number-box) container parens &optional selected)
;;   (declare (ignore parens selected))
;;   (let ((v (make-instance 'view
;;                           :view-container container)))
;;     (resize-view-for v (list (make-clickable v reg
;;                                              (make-instance 'static-text-dialog-item
;;                                                             :view-font +eqn-font+
;;                                                             :view-position #@(0 0)
;;                                                             :view-container reg
;;                                                             :part-color-list (if (box-selected box)
;;                                                                                  +selected-part-color-list+
;;                                                                                `(:text ,*black-color*))
;;                                                             :dialog-item-text "x")
;;                                              ;;(draw-text-with-font reg 0 0 "x" +eqn-font+ (if (box-selected box) *selected-color* *black-color*))
;;                                              (register-click-view reg (box-id box) (lambda () (click-box box))))))
;;     (when (box-pointed-to box)
;;       (push v *views-with-arrows*))
;;     v))

;; (defmethod draw-tree ((box answer-box) container parens &optional selected)
;;   (declare (ignore parens selected))
;;   (let ((b (find-tree box (lambda (b)
;;                             (and (typep b 'number-box)
;;                                  (not (arg-editable (first (box-containers b)))))))))
;;     (if b
;;         (let* ((v (make-instance 'view
;;                                  :view-container container))
;;                (views (list (progn 
;;                               (dformat "bo: ~a~%" b)
;;                               (draw-text-with-font v 0 0  (format nil "For x=~a, " (arg-value (number b))) +eqn-font+))
;;                             (tutor2-draw-arg (answer box) v nil))))
;;           (order-views views)
;;           (resize-view-for v views)
;;           v)
;;       (tutor2-draw-arg (answer box) container nil))))

;; (defmethod draw-tree ((box x-box) container parens &optional selected)
;;   (declare (ignore parens))
;;   (make-instance 'static-text-dialog-item
;;                  :view-font +eqn-font+
;;                  :view-container container
;;                  :dialog-item-text "x"
;;                  :part-color-list (if selected
;;                                       +selected-part-color-list+
;;                                     `(:text ,*black-color*))))


(defstruct eqn-elt
  type contents arrow selected click-fn id)

(defun compile-box-args (box parens)
  (with-slots (op arg1 arg2 x y selected) box
    (cond ((and (arg-editable arg1)
                (arg-editable op)
                (arg-editable arg2))
           (list
            (make-eqn-elt :type 'editable-box
                          :contents (cond ((box-user-input box) 
                                           (box-user-input box))
                                          ((and (arg-user-answer arg1)
                                                (arg-user-answer op)
                                                (arg-user-answer arg2))
                                           (list (arg-user-answer arg1)
                                                 (arg-user-answer op)
                                                 (arg-user-answer arg2)))
                                          (t nil))
                          :id box
                          :click-fn (lambda () (handle-fill-box box)))))
          ;;(draw-rect reg 0 0 60 20 *green-color*)
          ((and (arg-editable arg1)
                (arg-editable op))
           ;;(dformat "aaaa~%")

           ;;(draw-rect reg 0 0 40 20 *green-color*)
           (cons
            (make-eqn-elt :type 'editable-arg-op
                          :contents (when (and (arg-user-answer arg1)
                                               (arg-user-answer op))
                                      (list (arg-user-answer arg1)
                                            (arg-user-answer op)))
                          :id box
                          :click-fn (lambda () (handle-fill-arg-op box)))
            (compile-eqn (arg-value arg2) parens selected nil box)))
          ((and (arg-editable op)
                (arg-editable arg2))
           (nconc
            (compile-eqn (arg-value arg1) parens selected nil box)
            (list
             (make-eqn-elt :type 'editable-op-arg
                           :contents (when (and (arg-user-answer op)
                                                (arg-user-answer arg2))
                                       (list (arg-user-answer op)
                                             (arg-user-answer arg2)))
                          :id box
                           :click-fn (lambda () (handle-fill-op-arg box))))))
          ;; handle individually editable args. FIXME: duplicate code.
          ((arg-editable arg1)
           (nconc
            (list (make-eqn-elt :type 'editable-arg
                                :contents (arg-user-answer arg1)
                                :id arg1
                                :click-fn (lambda () (handle-fill-arg arg1))))
            (compile-eqn (arg-value op) parens selected (box-pointed-to box) box)
            (compile-eqn (arg-value arg2) parens selected nil box)))
          ((arg-editable op)
           (nconc
            (compile-eqn (arg-value arg1) parens selected nil box)
            (list (make-eqn-elt :type 'editable-arg
                                :contents (arg-user-answer op)
                                :id op
                                :click-fn (lambda () (handle-fill-arg op))))
            (compile-eqn (arg-value arg2) parens selected nil box)))
          ((arg-editable arg2)
           (nconc
            (compile-eqn (arg-value arg1) parens selected nil box)
            (compile-eqn (arg-value op) parens selected (box-pointed-to box) box)
            (list (make-eqn-elt :type 'editable-arg
                                :contents (arg-user-answer arg2)
                                :id arg2
                                :click-fn (lambda () (handle-fill-arg arg2))))))
          (t
           (cond ((simple-match box '(+ (* :n :x) (:a :n (* :n :x))))
                  (nconc (compile-eqn (arg-value arg1) parens selected nil box)
                         (compile-eqn (arg-value op) parens selected (box-pointed-to box) box)
                         (compile-eqn (arg-value arg2) nil selected nil box)))
                 ((simple-match box '(:a (:a (* :n :x) :n) (* :n :x)))
                  (nconc (compile-eqn (arg-value arg1) nil selected nil box)
                         (compile-eqn (arg-value op) parens selected (box-pointed-to box) box)
                         (compile-eqn (arg-value arg2) parens selected nil box)))
                 ((special-x-box box)
                  ;;                   (let* ((r (make-instance 'region-view
                  ;;                                            :view-container view))
                  ;;                          (v (compile-eqn arg1 r nil selected))
                  ;;                          (x (compile-eqn arg2 r nil selected)))
                  ;;                     (order-views (list v x))
                  ;;                     (resize-view-for r (list v x))
                  ;;                     (list r))
                  (nconc
                   (compile-eqn (arg-value arg1) nil selected nil box)
                   (compile-eqn (arg-value arg2) nil selected nil box)))
                 (t
                  (nconc (compile-eqn (arg-value arg1) parens selected nil box)
                         (compile-eqn (arg-value op) parens selected (box-pointed-to box) box)
                         (compile-eqn (arg-value arg2) parens selected nil box))))))))

(defun compile-eqn (root parens selected arrow id)
  "Return a list of eqn-elt's which is used in rendering the equation."
  (etypecase root
    ;; FIXME: hack alert. seems like argument should appear in the
    ;; compile-box-args function. not here.
;;     (argument
;;      (if (arg-editable root)
;;          (list (make-eqn-elt :type 'editable-arg
;;                              :contents (arg-user-answer root)
;;                              :id (arg-id root)
;;                              :click-fn (lambda () (handle-fill-arg root))))
;;        (compile-eqn (arg-value root) parens selected arrow id)))
    (equal-box
      (compile-box-args root nil))
    (arg-box
     (let ((lp (make-eqn-elt :type 'left-paren :contents "(" :selected (box-selected root)  :click-fn (lambda ())))
           (rp (make-eqn-elt :type 'right-paren :contents ")" :selected (box-selected root) :click-fn (lambda ()))))
       ;; making the parens clickable when no part of the box is editable.
       (unless (tutor-submit-enabled *tutor*)
         (setf (eqn-elt-click-fn lp) (lambda () (click-box root))
               (eqn-elt-click-fn rp) (lambda () (click-box root))))
       (setf (eqn-elt-id lp) root
             (eqn-elt-id rp) root)
       (dformat "lp: ~a~%rp: ~a~%" lp rp)
       (nconc
        (when parens (list lp))
        (compile-box-args root t)
        (when parens (list rp)))))
    (number-box
     (list (make-eqn-elt :type 'number-box :contents (arg-value (number root)) :selected (box-selected root) :arrow (box-pointed-to root) :id root :click-fn (lambda () (click-box root)))))
    (x-box
     (list (make-eqn-elt :type 'x-box :contents "x" :selected selected :id id :click-fn (lambda () (click-box id)))))
    (answer-box
     (if (arg-editable (answer root))
         (list (make-eqn-elt :type 'editable-arg
                             :contents (arg-user-answer (answer root))
                             :id (answer root)
                             :click-fn (lambda () (handle-fill-arg (answer root)))))
       (compile-eqn (arg-value (answer root)) nil (box-selected root) nil root)))
    (number
     (list (make-eqn-elt :type 'number :contents root :id id :selected selected :arrow arrow :click-fn (lambda () (click-box id)))))
    (symbol
     (list (make-eqn-elt :type 'operator :contents root :id id :selected selected :arrow arrow :click-fn (lambda () (click-box id)))))))

(defun abs-box-x (box view)
  (+ (abs-x view)
     (box-x box)))
;;         (point-h (view-scroll-position view)))))

(defun abs-box-y (box view)
  (+ (abs-y view)
     (box-y box)))
;;         (point-v (view-scroll-position view)))))

(defun hilite-enter-leave (view container enter-p)
  ;; search all the other hilite-text's for matching IDs. the whole
  ;; hilite thing isn't really all that general at all. it's a bunch
  ;; of half-baked general stuff to perform 1 specific task.
;;   (if enter-p
;;       (history-enter-expression (hilite-id view))
;;     (history-leave-expression (hilite-id view)))
  (dolist (i (all-subviews container 'hilite-text 'hilite-line))
    (when (and (hilite-id i)
               (eql (hilite-id i) (hilite-id view)))
      (if enter-p
          (if (typep i 'hilite-line)
              (progn 
                (setf (color i) *hilite-color*)
                (invalidate-view i))
            (set-part-color i :text *hilite-color*))
          (if (typep i 'hilite-line)
              (progn
                (setf (color i) (hilite-color i))
                (invalidate-view i))
            (let ((c (getf (hilite-color i) :text)))
              (set-part-color i :text c)))))))

(defun gen-hilite-fn (container enter-p)
  (if *hilite-components*
      (lambda (view) 
        (hilite-enter-leave view container enter-p))
    (lambda (view)
      (declare (ignore view)))))

(defun gen-hilite-view (text id reg container color-list)
  (let ((id (if *hilite-components* id nil)))
    (if (numberp text)
        (draw-number text reg 0 0 id
                     (gen-hilite-fn container t)
                     (gen-hilite-fn container nil)
                     color-list t)
      (draw-hilite-text reg 0 0 
                        text
                        +eqn-font+
                        id
                        (gen-hilite-fn container t)
                        (gen-hilite-fn container nil)
                        color-list
                        t))))

(defun thing-to-id (thing)
  (etypecase thing
    (box (box-id thing))
    (argument (arg-id thing))))

(defun render-compiled-eqn (container elts)
  (dformat "elts: ~a~%" elts)
  (let* ((view (make-instance 'tutor2-eqn-view
                              :view-container container))
         (views (loop for i in elts
                      collect (ecase (eqn-elt-type i)
                                (editable-box
                                 (make-clickable view reg
                                                 (if (eqn-elt-contents i)
                                                     (if (listp (eqn-elt-contents i))
                                                         (apply #'do-text-boxes reg (eqn-elt-contents i))
                                                       (do-text-boxes reg (eqn-elt-contents i)))
                                                   (setf (box-answer-view (eqn-elt-id i)) (draw-rect reg 0 0 60 20 *input-box-color* t)))
                                                 (register-click-view reg (thing-to-id (eqn-elt-id i)) (eqn-elt-click-fn i))))
                                ((editable-arg-op editable-op-arg)
                                 (make-clickable view reg
                                                 (if (eqn-elt-contents i)
                                                     (apply #'do-text-boxes reg (eqn-elt-contents i))
                                                   (setf (box-answer-view (eqn-elt-id i)) (draw-rect reg 0 0 40 20 *input-box-color* t)))
                                                 (register-click-view reg (thing-to-id (eqn-elt-id i)) (eqn-elt-click-fn i))))
                                (editable-arg
                                 (make-clickable view reg
                                                 (if (eqn-elt-contents i)
                                                     (if (listp (eqn-elt-contents i))
                                                         (apply #'do-text-boxes reg (eqn-elt-contents i))
                                                       (do-text-boxes reg (eqn-elt-contents i)))
                                                   (setf (arg-widget (eqn-elt-id i)) (draw-rect reg 0 0 20 20 *input-box-color* t)))
                                                 (register-click-view reg (thing-to-id (eqn-elt-id i)) (eqn-elt-click-fn i))))
                                ((x-box number-box)
                                 (let ((v (make-clickable view reg
                                                          (gen-hilite-view "x" (thing-to-id (eqn-elt-id i)) reg container
                                                                           (if (eqn-elt-selected i)
                                                                               +selected-part-color-list+
                                                                             `(:text ,*foreground-color*)))
                                                          (register-click-view reg (thing-to-id (eqn-elt-id i)) (eqn-elt-click-fn i)))))
                                   ;; keep track of which views need arrows drawn below them.
                                   (when (eqn-elt-arrow i)
                                     (push v *views-with-arrows*))
                                   v))
                                (number
                                 (make-clickable view reg
                                                 (gen-hilite-view (eqn-elt-contents i)
                                                                (thing-to-id (eqn-elt-id i)) reg container 
                                                                (if (eqn-elt-selected i)
                                                                    +selected-part-color-list+
                                                                  `(:text ,*foreground-color*)))
                                                 (register-click-view reg (thing-to-id (eqn-elt-id i)) (eqn-elt-click-fn i))))
                                ((operator left-paren right-paren)
                                 (let ((v (make-clickable view reg
                                                          (gen-hilite-view (format nil "~a" (eqn-elt-contents i)) (thing-to-id (eqn-elt-id i)) reg container 
                                                                           (if (eqn-elt-selected i)
                                                                               +selected-part-color-list+
                                                                             `(:text ,*foreground-color*)))
                                                          (register-click-view reg (thing-to-id (eqn-elt-id i)) (eqn-elt-click-fn i)))))
                                   ;; keep track of which views need arrows drawn below them.
                                   (when (eqn-elt-arrow i)
                                     (push v *views-with-arrows*))
                                   v)))))
         (nb (find 'number-box elts :key 'eqn-elt-type)))
    (when nb
      (push (draw-text-with-font view 0 0 (format nil "For x=~a, " (eqn-elt-contents nb)) +eqn-font+)
            views))
    (order-views views)
    (set-view-scroll-position view 0 (apply #'min (mapcar #'view-y views)))
    (resize-view-for view views)
    view))

;;;;;;;;
;;;;;;;;

(defun eqn-size (eqn &optional not-x-box)
  "Return the width and height as a point needed to display
  eqn. If NOT-X-BOX is T then don't include the x-box in the eqn
  size."
;;   (dformat "~a ~a~%" (box-nleaves eqn) (box-depth eqn))
  (let (minx
	miny
	maxy
	maxx)
  (labels ((remove-x-box (list)
                         (if not-x-box
                             (remove-if (lambda (b) (typep b 'x-box)) list)
                           list))
           (doit (box)
		 (setf minx (min (or minx (box-x box)) (box-x box))
		       miny (min (or miny (box-y box)) (box-y box))
		       maxx (max (or maxx (+ (box-x box) +node-width+)) (+ (box-x box) +node-width+))
		       maxy (max (or maxy (+ (box-y box) +level-height+)) (+ (box-y box) +level-height+)))
		 (mapc #'doit (remove-x-box (only-boxes (children box))))))
    (doit eqn)
    (values minx miny maxy maxx))))

(defun eqn-width (eqn)
  (multiple-value-bind (minx miny maxy maxx) (eqn-size eqn)
    (declare (ignore miny maxy))
    (- maxx minx)))

;; This is a terrible macro, but it'll work.
(defmacro gen-eqn-edge (name which)
  (with-gensyms (eqn not-x-box)
  `(defun ,name (,eqn &optional ,not-x-box)
     (multiple-value-bind (minx miny maxy maxx) (eqn-size ,eqn ,not-x-box)
       (declare (ignore ,@(remove which '(minx miny maxy maxx))))
       ,which))))

(gen-eqn-edge eqn-left minx)
(gen-eqn-edge eqn-right maxx)
(gen-eqn-edge eqn-top miny)
(gen-eqn-edge eqn-bottom maxy)

(defun eqn-height (eqn)
  (multiple-value-bind (minx miny maxy maxx) (eqn-size eqn)
    (declare (ignore minx maxx))
    (- maxy miny)))

(defun recenter-tree (root view &optional (width (point-h (view-size (tutor-eqn-area-view *tutor*)))) (height (point-v (view-size (tutor-eqn-area-view *tutor*)))))
  (funcall *recenter-tree-fn* root view width height))

(defun tutor2-recenter-tree (root view &optional (width (point-h (view-size (tutor-eqn-area-view *tutor*)))) (height (point-v (view-size (tutor-eqn-area-view *tutor*)))))
  (declare (ignore root))
  (if *simplified-ui*
      (set-view-position view 
                         (truncate (- width (view-width view)) 2)
                         (truncate (- height 
                                      (view-height view) 
                                      (view-height (tutor-buttons *tutor*))
                                      (view-height *message-view*)
                                      (if *views-with-arrows*
                                          -20 0))
                                   2))
    (set-view-position view 
                       (truncate (- width (view-width view)) 2)
                       (truncate (- height (view-height view)) 2))))


(defun tutor1-recenter-tree (root view &optional (move-view t) (width (point-h (view-size (tutor-eqn-area-view *tutor*)))) (height (point-v (view-size (tutor-eqn-area-view *tutor*)))))
  (multiple-value-bind (minx miny maxy maxx) (eqn-size root)
;;     (dformat "~a ~a ~a ~a~%" minx miny maxy maxx)
    (decf miny (- +level-height+ +box-height+))
    (decf minx (- +node-width+ +box-width+))
    (decf minx +view-pad+)
    (decf miny +view-pad+)
    (incf maxx +view-pad+)
    (incf maxy +view-pad+)
    (set-view-scroll-position view (make-point minx miny))
    ;; This is a bit of a hack. When the bottom box is an equal box,
    ;; then we want it to be top aligned because the we're finding the
    ;; value of the unknown box at the top. So everything feeds into
    ;; the top box. When we're evaluating into an answer box we want
    ;; it to be bottom aligned because everything is feeding into the
    ;; bottom. Except when we're in simplified-ui mode, because then
    ;; the buttons keep jumping up. In that case we always want the
    ;; equation to go down.
    (let ((top (if (and (typep root 'equal-box)
                        (not *simplified-ui*))
                   (point-v (view-position view))
                 (- (+ (point-v (view-position view)) (point-v (view-size view)))
                    (- maxy miny)))))
      (when move-view
	;; Reposition so that the bottom stays aligned with the previous eqn state
	(set-view-position view (make-point (- (truncate width 2) (truncate (- maxx minx) 2))
 					    (if (= (view-size view) #@(0 0))
						(- (truncate height 2) (truncate (- maxy miny) 2))
					      top))))
	;; Set the new size
	(set-view-size view (make-point (- maxx minx)
					(- maxy miny))))))


(defun remove-box-parent (box parent)
  (setf (box-parents box) (delete parent (box-parents box))))

(defun add-box-parent (box parent)
  (push parent (box-parents box)))

(defun reparent-box (box old new)
  (remove-box-parent box old)
  (add-box-parent box new))

(defun box-left-child-p (box &optional (p (first (box-parents box))))
  "Return true if box is the left child of its parent."
  (and (typep p 'arg-box)
       (eq (arg-value (box-arg1 p)) box)))

(defun make-and-space-tree (eqn)
  "Create a box tree out of EQN that is ready for drawing."
  (let ((tree (make-tree eqn)))
    (when (tree-x-box tree)
      (setf (x-box-reposition (tree-x-box tree)) t))
;;     (respace-tree tree 0 nil nil (tree-depth tree) +view-pad+ +view-pad+)
    (uber-respace-tree tree)
    tree))

(defun resize-view-for-boxes (view boxes)
  "Given a subtree, render it in VIEW. deletes any subviews in VIEW."
  (dformat "~a~%" boxes)
  (when boxes
    (let ((maxx (apply 'max (mapcar 'box-x boxes)))
          (maxy (apply 'max (mapcar 'box-y boxes)))
          (minx (apply 'min (mapcar 'box-x boxes)))
          (miny (apply 'min (mapcar 'box-y boxes))))
      (set-view-scroll-position view (- minx 10) (- miny 10))
      (set-view-size view (+ +box-width+ 20 (- maxx minx)) (+ +box-height+ 20 (- maxy miny)))
      (dformat "resize: ~a ~a ~a ~a~%" maxx minx (- maxx minx) (- maxy miny))
      ;; center the view inside its view
      (set-view-position view (truncate (- (view-width (view-container view)) (view-width view)) 2)
                         (truncate (- (view-height (view-container view)) (view-height view)) 2))

 )))

(defun draw-eqn (eqn)
  (let* ((w (make-instance 'window))
         (v (make-instance 'view
                           :view-container w))
         (tree (make-and-space-tree eqn)))
    (tutor1-recenter-tree tree v)
    (set-view-size w (view-size v))
    (set-view-position v #@(0 0))
    (tutor1-redraw-tree tree v)
    ))
