;;; Implementation of a simple calculator

(in-package :calc)

(defparameter +pad+ 2)
(defparameter +button-height+ 30)
(defparameter +button-width+ 30)
(defparameter +col-width+ 40)
(defparameter +row-height+ 40)

(defparameter +calc-window-width+ (* +col-width+ 4))
(defparameter +calc-window-height+ (* +row-height+ 6))

(defvar *calc-window* nil)
(defvar *digit-view* nil)
(defvar *digit-text* nil)

(defvar *digit-widget* nil)

(defvar *stack* nil)

(defvar *last-op* nil)

(defstruct (calc-state (:conc-name cs-))
  op n1 n2 state)

(defvar *calc-state* nil)

(defun init-digit-widget ()
  (setf *digit-widget* nil))

(defun draw-number (n)
  (let ((txt (apply 'concatenate 'string (mapcar 'prin1-to-string n))))
    (setf *digit-widget* (tutor::draw-text *digit-view* 
                                           (- (point-h (view-size *digit-view*)) (string-width txt tutor::*default-font*) +pad+)
                                           7
                                           txt))))

(defun update-digit-widget ()
  (when *digit-widget*
    (remove-subviews *digit-view* *digit-widget*))
  (ecase (cs-state *calc-state*)
    (start
     (draw-number '(0)))
    ((number-1 number-1-denominator equal)
     (draw-number (or (cs-n1 *calc-state*)
                      '(0))))
    ((number-2 number-2-denominator)
     (draw-number (or (cs-n2 *calc-state*)
                      (cs-n1 *calc-state*))))))

;;   (let ((txt (prin1-to-string (find-if #'numberp *stack*))))
;;     (setf *digit-widget* (tutor::draw-text *digit-view* (- (point-h (view-size *digit-view*)) (string-width txt '("Monaco" 14)) +pad+) 7 txt)))
;;   )

(defclass calc-window (window)
  ())

(defmethod view-key-event-handler ((w calc-window) ch)
  "activate a button when the user presses a numeric key."
  (let ((trans '((#\+ +) (#\- -) (#\* *) (#\/ safe-/) (#\= =))))
    (cond ((and (char>= ch #\0)
                (char<= ch #\9))
           (operation :number (- (char-code ch) (char-code #\0))))
          ((find ch trans :key #'first :test 'char=)
           (operation :op (second (assoc ch trans))))
          ((char= ch #\Enter)
           (operation :op '=))
          ((char= ch #\.)
           (operation :divider nil))
          ((or (char= ch #\c)
               (char= ch #\Esc))
           (operation :clear nil))
          (t (format t "~s" ch)))))

(defun init-calc ()
  (setf *calc-window* (make-instance 'calc-window
				:window-title "Calculator"
				:window-type :document
				:view-font '("Monaco" 14)
				:view-size (make-point +calc-window-width+ +calc-window-height+))
	*digit-view* (make-instance 'view
				    :view-position (make-point +pad+ +pad+)
				    :view-size (make-point (- +calc-window-width+ (* +pad+ 2)) +button-height+)
				    :view-container *calc-window*)
        *calc-state* (make-calc-state :state 'start)
;; 	*digit-text* nil
;; 	*last-op* nil
;;         *numerator* nil
;;         *op*
;;         *denominator* nil
        )
  (init-digit-widget)
  (tutor::draw-rect *calc-window* +pad+ +pad+ (- +calc-window-width+ +pad+) (+ +button-height+ +pad+)))

(defun calc ()
  (init-calc)
  (draw-buttons)
  (update-digit-widget))

;; (defun add-digit (d)
;;   "D can be 'x or 0-9."
;;   (tutor::history-add-calc-press d)
;;   (when (eql *last-op* '=)
;;     (setf *stack* nil))
;;   (when (symbolp (car *stack*))
;;     (push 0 *stack*))
;;   (vector-push-extend 

;; (if *denominator*
;;                           (+ (car *stack*) (/ d 
;;                           (+ (* 10 (car *stack*)) d))
;;   (setf *last-op* nil)
;;   (update-digit-widget))

(defun glob-number (n)
  "Given the N, list representing a number return it as a number"
  (when (eq (car (last n)) '/)
    (setf n (butlast n)))
  (when (equal n '(-))
    (setf n '(0)))
  (read-from-string (apply 'concatenate 'string (mapcar 'prin1-to-string n))))

(defun do-opn (op n1 n2)
  (let ((result (funcall op n1 n2)))
    (if (< result 0)
        (list '- (abs result))
      (list result))))

(defun log-op (op arg clickp)
  (ecase op
    ((:number :op) (tutor::history-add-calc-press arg clickp))
    (:minus (tutor::history-add-calc-press :minus clickp))
    (:divider (tutor::history-add-calc-press :divider clickp))
    (:clear (tutor::history-add-calc-press :clear clickp))))

(defun operation (op arg &optional clickp)
  (log-op op arg clickp)
  (if (eq op :clear)
      (setf *calc-state* (make-calc-state :state 'start))
    (ecase (cs-state *calc-state*)
      (start
       (ecase op
         (:op )
         (:number
          (setf (cs-n1 *calc-state*) (list arg)
                (cs-state *calc-state*) 'number-1))
         (:minus 
          (if (eq (car (cs-n1 *calc-state*)) '-)
              (pop (cs-n1 *calc-state*))
            (push '- (cs-n1 *calc-state*)))
          (setf (cs-state *calc-state*) 'number-1))
         (:divider
          (setf (cs-n1 *calc-state*) (list 1 '/)
                (cs-state *calc-state*) 'number-1-denominator))))
      (number-1
       (ecase op
         (:op
          (unless (eq arg '=)
            (setf (cs-op *calc-state*) arg
                  (cs-n2 *calc-state*) nil
                  (cs-state *calc-state*) 'number-2)))
         (:number
          (setf (cs-n1 *calc-state*) (append (cs-n1 *calc-state*) (list arg))))
         (:minus 
          (if (eq (car (cs-n1 *calc-state*)) '-)
              (pop (cs-n1 *calc-state*))
            (push '- (cs-n1 *calc-state*))))
         (:divider
          (cond ((null (cs-n1 *calc-state*))
                 (setf (cs-n1 *calc-state*) (list 1 '/)))
                ((equal (cs-n1 *calc-state*) '(-))
                 (setf (cs-n1 *calc-state*) (list '- 1 '/)))
                (t (setf (cs-n1 *calc-state*) (append (cs-n1 *calc-state*) (list '/)))))
          (setf (cs-state *calc-state*) 'number-1-denominator))))
      (number-1-denominator
       (ecase op
         (:op
          (unless (or (eq arg '=)
                      (eq (cdr (last (cs-n1 *calc-state*))) '/)
                      (and (cs-n1 *calc-state*)
                           (equal (list (nth (- (length (cs-n1 *calc-state*)) 2) (cs-n1 *calc-state*))
                                        (nth (1- (length (cs-n1 *calc-state*))) (cs-n1 *calc-state*)))
                                  '(/ 0))))
            (setf (cs-op *calc-state*) arg
                  (cs-n2 *calc-state*) nil
                  (cs-state *calc-state*) 'number-2)))
         (:number
          (setf (cs-n1 *calc-state*) (append (cs-n1 *calc-state*) (list arg))))
         (:minus 
          (if (eq (car (cs-n1 *calc-state*)) '-)
              (pop (cs-n1 *calc-state*))
            (push '- (cs-n1 *calc-state*))))
         (:divider ;; do nothing
          )))
      (number-2
       (ecase op
         (:op
          ;; just change the operator if we haven't typed a number yet.
          (if (null (cs-n2 *calc-state*))
              (unless (eq arg '=)
                (setf (cs-op *calc-state*) arg))
            (let ((result (do-opn (cs-op *calc-state*) 
                                  (glob-number (cs-n1 *calc-state*))
                                  (glob-number (cs-n2 *calc-state*)))))
              (if (eq arg '=)
                  (setf (cs-n1 *calc-state*) result
                        (cs-state *calc-state*) 'equal)
                (setf (cs-n1 *calc-state*) result
                      (cs-op *calc-state*) arg
                      (cs-n2 *calc-state*) nil)))))
          (:number
          (setf (cs-n2 *calc-state*) (append (cs-n2 *calc-state*) (list arg))))
         (:minus 
          (if (eq (car (cs-n2 *calc-state*)) '-)
              (pop (cs-n2 *calc-state*))
            (push '- (cs-n2 *calc-state*))))
         (:divider
          (cond ((null (cs-n2 *calc-state*))
                 (setf (cs-n2 *calc-state*) (list 1 '/)))
                ((equal (cs-n2 *calc-state*) '(-))
                 (setf (cs-n2 *calc-state*) (list '- 1 '/)))
                (t (setf (cs-n2 *calc-state*) (append (cs-n2 *calc-state*) (list '/)))))
          (setf (cs-state *calc-state*) 'number-2-denominator))))
      (number-2-denominator
       (ecase op
         (:op
          (unless (or (eq (cdr (last (cs-n2 *calc-state*))) '/)
                      (and (cs-n2 *calc-state*)
                           (equal (list (nth (- (length (cs-n2 *calc-state*)) 2) (cs-n2 *calc-state*))
                                        (nth (1- (length (cs-n2 *calc-state*))) (cs-n2 *calc-state*)))
                                  '(/ 0))))
          (let ((result (do-opn (cs-op *calc-state*) 
                                (glob-number (cs-n1 *calc-state*))
                                (glob-number (cs-n2 *calc-state*)))))
            (if (eq arg '=)
                (setf (cs-n1 *calc-state*) result
                      (cs-state *calc-state*) 'equal)
              (setf (cs-n1 *calc-state*) result
                    (cs-op *calc-state*) arg
                    (cs-n2 *calc-state*) nil
                    (cs-state *calc-state*) 'number-2)))))
         (:number
          (setf (cs-n2 *calc-state*) (append (cs-n2 *calc-state*) (list arg))))
         (:minus 
          (if (eq (car (cs-n2 *calc-state*)) '-)
              (pop (cs-n2 *calc-state*))
            (push '- (cs-n2 *calc-state*))))
         (:divider ;; do nothing
          )))
      (equal
       (ecase op
         (:op
          (unless (eq arg '=)
            (setf (cs-op *calc-state*) arg
                  (cs-n2 *calc-state*) nil
                  (cs-state *calc-state*) 'number-2)))
         (:minus 
          (if (eq (car (cs-n1 *calc-state*)) '-)
              (pop (cs-n1 *calc-state*))
            (push '- (cs-n1 *calc-state*))))
         (:divider
          (setf (cs-n1 *calc-state*) (list 1 '/)
                (cs-state *calc-state*) 'number-1-denominator))
         (:number
          (setf (cs-n1 *calc-state*) (list arg)
                (cs-state *calc-state*) 'number-1))))))
    (update-digit-widget))
              
(defun safe-/ (&rest args)
  (or (ignore-errors (apply #'/ args))
      0))

;; (defun do-operation (opn)
;;   (tutor::history-add-calc-press opn)
;;   (case opn
;;     (= (if (= (length *stack*) 3)
;; 	   (setf *stack* (list (funcall (second *stack*) (third *stack*) (first *stack*))))
;; 	 (setf *stack* (last *stack*))))
;;     (C (setf *stack* (list 0)))
;;     (t (when (= (length *stack*) 3)
;; 	 (setf *stack* (list (funcall (second *stack*) (third *stack*) (first *stack*)))))
;;        (if (symbolp (car *stack*))
;; 	   (setf (car *stack*) opn)
;; 	 (push opn *stack*))))
;;   (update-digit-widget)
;;   (setf *last-op* opn))

(defun add-numeric-button (&key x y digit (width +button-width+) (height +button-height+))
  (make-instance 'button-dialog-item
		 :view-container *calc-window*
		 :view-position (make-point (+ x +pad+) (+ y +pad+))
		 :view-size (make-point width height)
		 :dialog-item-text (prin1-to-string digit)
		 :dialog-item-action (lambda (button) (declare (ignore button))
				       (operation :number digit t))))
			   
(defun add-opn-button (&key x y opn (width +button-width+) (height +button-height+))
  (make-instance 'button-dialog-item
		 :view-container *calc-window*
		 :view-position (make-point (+ x +pad+) (+ y +pad+))
		 :view-size (make-point width height)
		 :dialog-item-text (symbol-name (if (eql opn 'safe-/) '/ opn))
		 :dialog-item-action (lambda (button) (declare (ignore button))
				       (operation :op opn t))))

(defun add-opn-button (&key x y opn sym (width +button-width+) (height +button-height+))
  (make-instance 'button-dialog-item
		 :view-container *calc-window*
		 :view-position (make-point (+ x +pad+) (+ y +pad+))
		 :view-size (make-point width height)
		 :dialog-item-text (if sym (symbol-name sym) (symbol-name opn))
		 :dialog-item-action (lambda (button) (declare (ignore button))
				       (operation :op opn))))

(defun add-spe-button (&key x y special sym (width +button-width+) (height +button-height+))
  (make-instance 'button-dialog-item
		 :view-container *calc-window*
		 :view-position (make-point (+ x +pad+) (+ y +pad+))
		 :view-size (make-point width height)
		 :dialog-item-text (symbol-name sym)
		 :dialog-item-action (lambda (button) (declare (ignore button))
				       (operation special nil t))))

(defun draw-buttons ()
  ;; calculator buttons
  (loop for i from 0 to 2
	do (loop for j from 0 to 2
		 do (add-numeric-button :x (* j +col-width+)
					:y (- +calc-window-height+ (* (+ i 2) +row-height+))
					:digit (+ (* i 3) j 1))))
  ;; Zero


  (add-numeric-button :x 0 :y (- +calc-window-height+ +row-height+) :digit 0)
  ;; Operations
  (add-spe-button :x (* +col-width+ 0) :y (* 1 +row-height+) :special :clear :sym 'C :width (+ +col-width+ +button-width+))
  (add-spe-button :x (* +col-width+ 2) :y (* 1 +row-height+) :special :minus :sym ')
  (add-opn-button :x (* +col-width+ 3) :y (* 1 +row-height+) :opn '*)
  (add-opn-button :x (* +col-width+ 3) :y (* 2 +row-height+) :opn 'safe-/ :sym ')
  (add-opn-button :x (* +col-width+ 3) :y (* 3 +row-height+) :opn '-)
  (add-opn-button :x (* +col-width+ 3) :y (* 4 +row-height+) :opn '+)
  (add-spe-button :x (* +col-width+ 1) :y (* 5 +row-height+) :special :divider :sym '/)
  (add-opn-button :x (* +col-width+ 2) :y (* 5 +row-height+) :opn '= :width (+ +col-width+ +button-width+)))

(defun random-clicks (window)
  (let ((buttons (tutor::all-subviews window 'button-dialog-item)))
    (loop
       (let ((b (nth (random (length buttons)) buttons)))
         (funcall (dialog-item-action-function b) b)))))

