;;; Equation History Support

(in-package :tutor)

;; keep track of the equation and whether the operation done to get
;; here was executed correctly (all the Right numbers).
(defstruct history
  eqn incorrect)

(defun eqn-history-back ()
  (when (tutor-eqn-history-before *tutor*)
    ;; If they backed out of an operation then it shouldn't be added it to the eqn-history.
    (clear-selected)
    (unless (tutor-submit-enabled *tutor*)
      (push (make-history :eqn (tutor-eqn-root *tutor*)
                          :incorrect (tutor-incorrect-submit *tutor*)) (tutor-eqn-history-after *tutor*)))
    (setf (tutor-submit-enabled *tutor*) nil)
    (clear-input-dialog)
    (let ((h (pop (tutor-eqn-history-before *tutor*))))
      (setf (tutor-eqn-root *tutor*) (history-eqn h)
            (tutor-incorrect-submit *tutor*) (history-incorrect h)
            (tutor-eqn-history-backup *tutor*) (make-history :eqn (copy-boxes (tutor-eqn-root *tutor*) nil)
                                                             :incorrect (tutor-incorrect-submit *tutor*))))))

(defun eqn-history-forward ()
  (when (tutor-eqn-history-after *tutor*)
    (clear-selected)
    (push (make-history :eqn (tutor-eqn-root *tutor*) :incorrect (tutor-incorrect-submit *tutor*)) (tutor-eqn-history-before *tutor*))
    (let ((h (pop (tutor-eqn-history-after *tutor*))))
      (setf (tutor-eqn-root *tutor*) (history-eqn h)
            (tutor-incorrect-submit *tutor*) (history-incorrect h)
            (tutor-eqn-history-backup *tutor*) (make-history :eqn (copy-boxes (tutor-eqn-root *tutor*) nil)
                                                             :incorrect (tutor-incorrect-submit *tutor*))))))

(defun eqn-history-beginning ()
  "go to the beginning of the problem."
  (loop while (tutor-eqn-history-before *tutor*)
        do (eqn-history-back)))

(defun eqn-history-push ()
  "Push the current backup onto the history. Used when the user clicks
on an operation."
  (push (tutor-eqn-history-backup *tutor*) (tutor-eqn-history-before *tutor*)))

(defun eqn-history-backup ()
  "backup the equation into the history backup. This is called when
the user clicks the submit button."
  (setf (tutor-eqn-history-backup *tutor*) (make-history :eqn (copy-boxes (tutor-eqn-root *tutor*) nil)
                                           :incorrect (tutor-incorrect-submit *tutor*)))
  ;;   ;; Dock them for backtracking
  ;;   (when *eqn-history-after*
  ;;     (inc-demerits +backup-demerits+))
  (setf (tutor-eqn-history-after *tutor*) nil))
