(in-package :tutor)

(defmacro defcustom (var initval txt type &rest opts)
  `(progn
     (defvar ,var ,initval ,txt)
     (setf (get ',var :customize) t
           (get ',var :type) ,type
           (get ',var :values) ',opts)))

(defun get-customizable-values ()
  "Return a list customizable value pairs."
  (loop for i being each symbol in :tutor
     when (get i :customize)
     collect `(,i ,(symbol-value i))))

(defparameter *lightest-pink-color* (make-color 65000 55000 55000))
(defparameter *lightest-green-color* (make-color 45000 65000 45000))
(defparameter *lightest-gray-color* (make-color 55000 55000 55000))

;; These are all the colors used in the experiment.
(defparameter *selected-color* *red-color*)
(defparameter *selected-bk-color* *lightest-pink-color*)
(defparameter *hilite-color* *orange-color*)
(defparameter *input-color* *dark-green-color*)
(defparameter *input-box-color* *lightest-green-color*)
(defparameter *background-color* *white-color*)
(defparameter *foreground-color* *black-color*)
(defparameter *last-eqn-state-color* *light-gray-color*)

(defparameter +eqn-font+ '("Monaco" 19))
(defparameter +selected-eqn-font+ '("Monaco" :bold 19 (:color #.*selected-color*)))
(defparameter +selected-part-color-list+ `(:text ,*selected-color* :body ,*selected-bk-color*))

(defparameter *default-font* '("Geneva" 18))
(defparameter *box-font* '("Monaco" 20))
(defparameter *title-font* '("Geneva" 28 :bold))
(defparameter *section-font* '("Geneva" 24 :bold))
(defparameter *text-font* '("Geneva" 18))
(defparameter *message-font* '("Geneva" 18))
(defparameter *fine-grained-hint-font* '("Monaco" 14))
(defparameter *0back-font* '("Geneva" 36))

(defparameter +fill-in-user-hint-text+ "Please fill in the hint text before moving on.")

(defvar *window* nil)
(defvar *input-window* nil)
(defvar *input-text* nil)

(defvar *fine-grained-hint-view* nil)
(defvar *fine-grained-hint-pointing-to* nil)

;; (defvar *eqn-view* nil
;;   "The view containing the current state of the equation")
;; (defvar *last-eqn-area-view* nil
;;   )

;; (defvar *eqn-area-view* nil
;;   )
;; (defvar *last-eqn-view* nil
;;   "The view that shows the state of the equation before the last operation.")

(defvar *message-view* nil
  "An area for messages.")

;; (defvar *eqn-root*)

;; ;; (defvar *eqn-x-box*)

;; (defvar *selected-tree* nil)

(defparameter +window-pad+ 30)
(defparameter +view-pad+ 10)

(defparameter +window-width+ 1220)
(defparameter +window-height+ 730)

(defparameter +pad+ 3)
(defparameter +arg-pad+ 5)
(defparameter +level-height+ 60)
(defparameter +node-width+ 120)
(defparameter +unknown-width+ 23)
(defparameter +unknown-height+ 23)
(defparameter +box-width+ 80)
(defparameter +box-height+ (+ +unknown-height+ (* 2 +pad+)))

;; (defvar *eqn-history-before* nil
;;   "The list of equation states before the current equation.")

;; (defvar *eqn-history-after* nil
;;   "The list of euation states after the current equation. Unless
;; browsing the history, this is nil.")

;; (defvar *eqn-history-backup* nil
;;   "some operation destructively alter the equation. This variable is
;; assigned a copy of the equation after a submit operation is performed. It is what is placed on the history.")

;; (defvar *submit-selections* nil)
;; (defvar *submit-tree* nil)
;; (defvar *submit-enabled* nil)
;; ;; (defvar *submit-fn* nil
;; ;;   "call this when the user presses the submit button")
;; (defvar *submit-answers* nil
;;   "The possible correct resulting trees after the operation is applied.")

;; (defparameter +arithmetic-demerits+ 10)
;; (defparameter +bad-operation-demerits+ 10)
;; (defparameter +backup-demerits+ 10)
;; (defparameter +submit-demerits+ 10)
;; (defparameter +whats-next-demerits+ 5)

(defparameter +bad-fill-in+ 7)
(defparameter +operation-points+ 7)
(defparameter +bad-operation-points+ 3)
(defparameter +hint-points+ 4)

;; (defparameter +bad-fill-in+ 4)
;; (defparameter +operation-points+ 4)
;; (defparameter +bad-operation-points+ 2)
;; (defparameter +hint-points+ 3)

;; (defparameter +session1-total-points+ (* (- (+ 42 40 50 56 24)
;;                                             (+ 2 2 2 6 4)) +operation-points+))

;; (defparameter +session2-total-points+ (* (- (+ 70 159 77)
;;                                             (+ 10 8 7)) +operation-points+))

;; (defparameter +total-points+ (+ +session1-total-points+ +session2-total-points+))


(defclass problem-set ()
  ((instructions :initform nil :initarg :instructions :accessor instructions)
   (hint-modifiers :initarg :hint-modifiers :accessor hint-modifiers
                    :documentation "A list of features that change how hints work.")
   (problems :initform nil :initarg :problems :accessor problems)))

(defclass problem-type ()
  ((instructions :initform nil :initarg :instructions :accessor instructions)
   (template :initform nil :initarg :template :accessor template)
   (constraints :initform nil :initarg :constraints :accessor constraints)
   (nright :initform nil :initarg :nright :accessor problem-type-nright)
   (answer :initform nil :initarg :template-answer :accessor template-answer)))

(defclass problem ()
  ((eqn :initform nil :initarg :eqn :accessor problem-eqn)
   (answer :initarg nil :initarg :answer :accessor problem-answer)
   (points :initarg nil :initarg :points :accessor problem-points)
   (hint-modifiers :initarg :hint-modifiers :accessor hint-modifiers
                    :documentation "A list of features that change how hints work.")
   (auto-hint :initarg nil :initarg :auto-hint :accessor problem-auto-hint)))

(defvar *history-id* nil
  "The current subject ID.")

(defparameter *operations* nil
  "the Defined operations")

(defparameter *transformations* (make-hash-table))

(defstruct submit
  "This struct carries what is needed by the hint system when in submit-mode."
  operation hint-index)

(defstruct tutor
  "The big bad-ass tutor state (BBATS)."
  ;; command tracking. So far this is only used by the whats next
  ;; system to give a useful hint after First Mistake. TODO: actually
  ;; implement this.
  (last-command nil)
  ;; we need to know this for the tutor1 fmri mode
  (last-selected-boxes nil)
  ;; gui state
  (eqn-view nil)
  ;; used by the simplified-ui code to store the last-eqn-area plus
  ;; the gray surrounding box for tutor1.
  (last-eqn-top-view nil)
  (last-eqn-area-view nil)
  (last-eqn-view nil)
  (eqn-area-view nil)
  (buttons nil)
  ;;(message-view nil)
  (eqn-root nil)
  (selected-tree nil)
  (selected-eqns nil)
  (instruction-view nil)
  ;; history
  (eqn-history-before nil)
  (eqn-history-after nil)
  (eqn-history-backup nil)
  ;; submit
  (submit-selections nil :type list)
  (submit-tree nil)
  (submit-enabled nil)
  (submit-answer nil :type list)
  (submit-post-fn nil)
  (green-box-hints nil)
  ;; problem state
  (hint-modifiers nil)
  (current-problem nil)
  (total-points 0 :type fixnum)
  (score 0 :type fixnum)
  (score-view nil)
  (problems-view nil)
  (instructions nil)
  ;; called when the problem is finished
  (callback nil)
  (num-problems nil)
  (problem-score 0 :type fixnum)
  (auto-whats-next-mode nil)
  (global-whats-next nil)
  (click-regions nil :type list)
  (incorrect-submit nil :type boolean))

(defvar *tutor* nil
  "Contains a TUTOR structure with all the tutor state in it.")

;; (defvar *total-points* 0)
;; (defvar *score-view*)
;; (defvar *score* 0
;;   "how much they get paid is dependant on their score.")

;; ;;; Problem globals

;; (defvar *problem-set* nil
;;   "the problems and instructions for the problems.")

;; (defvar *current-problem* nil)

;; (defvar *problem-score* 0)

;; (defvar *auto-whats-next-mode* nil
;;   "When set to T hints are automatically shown.")

;; (defvar *global-whats-next* nil
;;   "This is an override to turn on whats next.")

;; (defvar *click-regions* nil
;;   "keep track of clickable areas.")

(defclass tutor-window (window)
  ())

(defcustom *simplified-ui* nil
  "Set to T to hide the instructions and some of the buttons when solving a problem on tutor2"
  :bool)

(defun tutor-height ()
  (if (tutor1-p)
      (if *simplified-ui*
          748
        730)
    (if *simplified-ui*
        748
      890)))

(defun tutor-width ()
  (if (tutor1-p)
      (if *simplified-ui*
          ;; 1000
          1024
        1220)
    (if *simplified-ui*
        730
      820)))

(defun make-window ()
  (let ((w (make-instance 'tutor-window
                          :window-title "Tutor"
                          :window-type :document
                          :view-font '("Monaco" 14)
                          :color-p t
                          :view-size (make-point (tutor-width) (tutor-height)))))
    (set-back-color w *background-color*)
    (set-fore-color w *foreground-color*)
    w))
    

(defun repos-window ()
  "Call this for the fmri sessions."
  (when *window*
    (set-view-position *window* 0 20)))

(defun init ()
  "Global init. Returns a TUTOR structure containing a fresh state"
  (setf *window* (make-window)
        *tutor* (make-tutor)
        *input-window* nil)
  (when *simplified-ui*
    (repos-window)))

;; Backup the load-path. This is where we'll write the history file.
(defvar *history-path* nil)
(setf *history-path* *load-pathname*)

(defun true-p (x)
     (if x t nil))

(defun listnn (&rest values)
  "Return a list of VALUES excluding nil."
  (delete nil values))

;; (defvar *instruction-view* nil)

;; (defvar *incorrect-submit* nil
;;   "When this is true the subject made a mistake. This is set to
;;   one and the state is tagged. If you click the 'Last Mistake'
;;   button it goes back to the state before the one marked as a
;;   mistake.")

;; A common macro writing helper
(defmacro with-gensyms ((&rest syms) &body body)
  `(let ,(mapcar (lambda (s)
                   `(,s (gensym ,(symbol-name s))))
                 syms)
     ,@body))

;; (defmacro curry (fn (&rest got) (&rest needed))
;;   `(lambda ,needed (,fn ,@got ,@needed)))

(defmacro button-fn (&body body)
  "Return a function that takse one argument and ignores in, then
  execs body. We never care about the widget passed in when a
  button is clicked."
  (with-gensyms (b)
  `(lambda (,b) 
     (declare (ignore ,b))
     ,@body)))

(defparameter +incorrect-text+
  "Your answer is incorrect.
Use the <- and -> buttons (or the left and right arrow keys)
and the First Mistake button
to review your work and correct the mistake.")

(defparameter +not-done-text+
  "You are not finished. You need to do more to the problem.")

(defun prefixify (tree)
  "Given an infix eqn, (1 + 3), turn it into (+ 1 3)."
  (cond ((numberp tree)
         tree)
        ((symbolp tree)
         tree)
        ((= (length tree) 1)
         (prefixify (first tree)))
        (t (list (second tree)
                 (prefixify (first tree))
                 (prefixify (third tree))))))


(defvar *refresh-eqn-fn* nil)
(defvar *draw-layout-fn* nil)
(defvar *redraw-tree-fn* nil)

(defcustom *which-tutor* 'tutor1
  "Choose between tutor1 - the box and arrow tutor - or tutor2 - the regular algebra tutor."
  :enum tutor1 tutor2)

(defvar *recenter-tree-fn* nil)

(defun nconc1 (&rest lists)
  "destructiveyl append the lists together. If elements are not
  in a list, then wrap it in a list and append it."
  (apply 'nconc (mapcar (lambda (l)
                          (if (and (atom l)
                                   (not (null l)))
                              (list l)
                            l))
                        lists)))

(defun append1 (&rest lists)
  "append the lists together. If elements are not
  in a list, then wrap it in a list and append it."
  (apply 'append (mapcar (lambda (l)
                          (if (and (atom l)
                                   (not (null l)))
                              (list l)
                            l))
                        lists)))

(defcustom *allow-pre-evaled-box-input* nil
  "When set to T, allow the user to skip an evaluation step by typing the answer in the box."
  :bool)

(defcustom *0back-after-submit* nil
  "Set this to T if you'd like to play the 0back game after each submit."
  :bool)

(defcustom *0back-iterations* 8
  "How many letters should the 0back game display?"
  :number)

(defcustom *0back-on-delay* 1000
  "How many miliseconds should the letter be displayed for."
  :number)

(defcustom *0back-off-delay* 250
  "How many miliseconds should there be nohting displayed."
  :number)

(defcustom *instruction-delay* nil
  "Whether the tutor should pause after displaying the
  instructions before continuing to display the rest."
  :bool)

(defcustom *instruction-delay-amount* 30000
  "The amount of time only the instructions are shown before the problem."
  :number)

(defcustom *input-method* :kbd
  "Set to :mouse to pop up a mouse controlled calculator to input data, :kbd to pop up a similar input box, or :dialog to use a dialog box."
  :enum :mouse :kbd :dialog)

(defcustom *force-correct-fill-in* nil
  "Set to T to force the user to submit the correct answer."
  :bool)

(defcustom *hint-types* (list :normal)
  "What kind of hints to give. one
of :FINE-GRAINED, :WRITTEN, :NORMAL, :USER. :USER will prompt the
user to fill in the hints."
  :list :fine-grained :written :normal :user)

(defvar *fill-in-object* nil
  "The box or arg that is currently being filled in.")

(defcustom *odd-problems* t
  "Do the even or odd problems?"
  :bool)

(defcustom *force-read-instructions* nil
  "When set to T, the user must click a button below the instructions before the problem 'activates'."
  :bool)

(defcustom *sticky-instructions* nil
  "When set to T, the instructions from the sample problems stay when doing the actual problems."
  :bool)

(defcustom *ask-experimenter-button* nil
  "Set to T to get an ask experimenter button."
  :bool)

(defvar *hints-stage* nil
  "The current stage that a hint was asked for. Fine grained
hints are given until the subject enters a new stage. There are 3
stages:

:select selecting boxes
:operation choosing the operation
:fillin filling in the blanks")

(defvar *cycle-hints-p* nil
  "When this is T then give hints for the current cycle. This is
automatically set to NIL when the cycle ends.")

(defvar *nohint-override* nil
  "When this is true, the hint button and auto hints are
  suppressed. Used in angela's experiment.")

(defcustom *offer-hints-after-first-mistake* nil
  "When this is T then hints are offered when they hit first mistake."
  :bool)

(defvar *instruction-callback* nil
  "When this is a callback, clicking on the window calls this callback. Used for clicking through the instructions in fmri mode.")

(defcustom *crosshair-before-problem* nil
  "Should a crosshair appear before each problem?"
  :bool)

(defcustom *crosshair-before-problem-time* 4000
  "How long should the crosshair appear? In miliseconds"
  :number)

(defcustom *ask-experimenter-wait-length* 30
  "How long should the tutor wait when they hit the ask experimenter button?"
  :number)

(defcustom *no-solve-instructions* nil
  "Set to T to have no instructions instead of the usual 'solve
the following problems' text."
  :bool)

(defvar *current-user-supplied-hints* nil
  "Store the hint for each stage in this variable. when the cycle
  completes these are written to the appropriate slots in XXX.")

(defvar *valid-operations* nil
  "Lists All operations the subject is allowed to do.")

(defun auto-whats-next-mode ()
  (or (tutor-global-whats-next *tutor*)
      (tutor-auto-whats-next-mode *tutor*)))

(defun fmri-enable ()
  "call this function to turn on the features used when scanning subjects."
  (setf *0back-after-submit* t
        *instruction-delay* t
        *input-method* :mouse
        *simplified-ui* t
        *force-correct-fill-in* t))

(defun fmri-disable ()
   "call this function to turn off the features used when scanning subjects."
  (setf *0back-after-submit* nil
        *instruction-delay* nil
        *input-method* :mouse
        *simplified-ui* nil
        *force-correct-fill-in* nil))

(defun angela-experiment (fine-grained-p word-hints-p)
  (setf *sticky-instructions* t
        *force-read-instructions* t
        *offer-hints-after-first-mistake* t
        *ask-experimenter-button* t
        *hint-types* (append1 (and fine-grained-p :fine-grained)
                             (and word-hints-p :written))))

(defun tutor-sleep (miliseconds)
  (assert (= internal-time-units-per-second 1000))
  (let ((end (+ (get-internal-real-time) miliseconds)))
    (loop while (< (get-internal-real-time) end)
          ;;do (ccl::event-poll)
          ;;do (ccl::clear-process-polling)
;;           do (format t "~a~%" *current-process*)
          do (event-dispatch)
          do (window-update-event-handler *window*)
;;           do (process-allow-schedule)
;;           when (ccl::event-available-p)
;;           do (ccl::%event-dispatch)
          ;; do (event-dispatch)
          )))

;; (defun sleep2 (ticks)
;;   (make-

(defcustom *debug* nil
  "Set to T to enable debug output."
  :bool)

(defun dformat (fmt &rest args)
  (when *debug*
    (apply #'format t fmt args)))

(defcustom *special-x-box* nil
  "Set to T to draw (n * X) as nX in tutor2"
  :bool)

(defvar *user-supplied-hints* (make-hash-table)
  "an alist of user supplied hints.")

(defvar *instruction-gen-start-time* nil
  "instruction generation is paid by the hour. This var stores when they started generating instructions.")

(defcustom *instruction-gen-rate* 10
  "How many dollars per hour do they get for generating instructions?"
  :number)

(defun concat (&rest strings)
  "concatenate the strings together"
  (apply 'concatenate 'string strings))

(defun draw-crosshair (window time &optional scroll-p)
  (let* ((maxx (point-h (view-size window)))
         (maxy (point-v (view-size window)))
         (cx (+ maxx (truncate maxx 2)))
         (cy (- (+ maxy (truncate maxy 2)) 15)))
    (when scroll-p
      (set-view-scroll-position window maxx maxy))
    (let ((views (list (draw-line window (- cx 20) cy (+ cx 20) cy)
                       (draw-line window cx (- cy 20) cx (+ cy 20)))))
      (history-add 'cross-hair-start)
      (tutor-sleep time)
      (apply #'remove-subviews window views)
      (history-add 'cross-hair-end))
    (when scroll-p
      (set-view-scroll-position window 0 0))))

(defun split-list (l &rest splits)
  "splits a list at each index in SPLITS"
  (loop for j = 0 then i
        for i in (append1 splits (length l))
        collect (subseq l j i)))

(defun reverse-color ()
  "Toggle between a black and white background."
  (psetf *background-color* *foreground-color*
         *foreground-color* *background-color*))

(defun string-butlast (s)
  (if (> (length s) 0)
      (subseq s 0 (1- (length s)))
    ""))

(defun flatten (tree)
  "Flatten a tree into a list."
  (apply 'nconc
         (mapcar (lambda (a)
            (if (listp a)
                (flatten a)
              (list a)))
          tree)))

(defun current-date ()
  "Return the current date and time as a formatted string."
  (multiple-value-bind (s min h d m y) (decode-universal-time (get-universal-time))
    (format nil "~2,'0d:~2,'0d:~2,'0d ~2,'0d/~2,'0d/~d" h min s d m y)))

(defvar *problem-generation-function* 'gen-problems
  "This function is used to specify an alternative set of
problems. It should take 1 argument, the section symbol and
return hint-exclusions, valid operations, practice problems and
real problems.")
