;;; Messages, Hints, Score

(in-package :tutor)

(defvar *message-widget* nil)
(defvar *hint-text* nil)
(defvar *message-text* nil)

(defun clear-message (&optional txt-too)
  (when *message-widget*
    (remove-subviews (view-container *message-widget*) *message-widget*)
    (setf *message-widget* nil))
  (when txt-too
    (setf *hint-text* nil
          *message-text* nil)))

(defun message-view-container ()
  (if (and (tutor1-p) 
           (not *simplified-ui*))
      *window*
    *message-view*))

(defun message-area-width ()
  (view-width *message-view*))  

(defun wrap-message-text (txt)
  (wrap-text txt (message-area-width) *message-font*))

(defun message-reposition (view &optional (x-offset 0))
  (if (tutor1-p)
      (if *simplified-ui*
          (set-view-position view x-offset 0)
        (set-view-position view
                           (+ 550 x-offset)
                           (- (point-v (add-points (view-position *message-view*)
                                                   (view-size *message-view*)))
                              (point-v (view-size view)))))
    (if *simplified-ui*
        (set-view-position view x-offset 0)
      (set-view-position view x-offset 5))))

(defun message-resize (view)
  (if *simplified-ui*
      (let ((c (view-container view)))
        (set-view-size c (max (view-width c) (view-width view))
                       (max (view-height c) (view-height view))))))

(defun redraw-message (&optional input-next)
  ;; if there's a user-instruction widget we should keep it around
  (let ((user (first (all-subviews (message-view-container) 'user-hint-instruction))))
    (clear-message)
    (let* ((msg (or *hint-text* *message-text*))
           (err (if *hint-text*
                    *message-text*
                    nil)))
      (when msg
        (setf *message-widget* (cond ((and err msg
                                           (stringp msg))
                                      (draw-text-plus-header (message-view-container)
                                                             5 0 err msg *message-font*))
                                     ((and err msg)
                                      (with-container ((message-view-container) view)
                                        (let ((txt (draw-text-with-font view 5 0 err *message-font* *red-color*)))
                                          (draw-user-instruction view 0 (view-height txt) (message-area-width) 30 "Click here to enter your hint" (or input-next (and user (user-text user))) t t))))
                                     ((eq msg :input)
                                      (draw-user-instruction (message-view-container) 0 0 (message-area-width) 30 "Click here to enter your hint" (or input-next (and user (user-text user))) t t))
                                     (t (draw-text-with-font (message-view-container) 5 0 msg *message-font*))))
        (message-reposition *message-widget* (view-x *message-widget*))
        (message-resize *message-widget*)))))

(defun set-message-text-with-button (txt button-txt action)
  "Set the message to some text and a button. When the button is
clicked ACTION is performed."
  (clear-message)
  (let* ((view (make-instance 'view 
                              :view-container (message-view-container)))
         (tv (draw-text view 0 0 txt))
         (bv (make-instance 'button-dialog-item
                            :view-container view
                            :dialog-item-text button-txt
                            :dialog-item-action (lambda (b) (declare (ignore b)) (funcall action)))))
    (order-views (list tv bv) 0 0 0)
    (resize-view-for view (list tv bv))
    (message-reposition view)
    (setf *message-widget* view)))

(defun init-message-widget ()
  (setf *message-widget* nil))

(defun set-hint-text (txt &optional input-text)
  "Set hint portion of text-above-the-problem. INPUT-TEXT is only used when txt is :INPUT"
  (when (not (equal txt *hint-text*))
    (setf *hint-text* txt)
    (redraw-message input-text)))

(defun set-message-text (txt)
  "Set message portion of text-above-the-problem"
  (when (string/= txt *message-text*)
    (setf *message-text* txt)
    (redraw-message)))

(defun calculate-money (score)
  (format nil "$~,2F" ;;(* (/ score *total-points*) (* 32 (/ *total-points* +total-points+)))))
          (/ score 100)))

(defun draw-score ()
  (when (tutor-score-view *tutor*)
    (let ((v (draw-text-with-font (tutor-score-view *tutor*) 0 0 (calculate-money (tutor-score *tutor*)) '("Geneva" 14))))
      (set-view-position v (- (view-width (tutor-score-view *tutor*)) (view-width v)) 0))))

(defun draw-problems (n max)
  (let ((txt (format nil "Problem ~a of ~a" (- max n) max)))
    (if (tutor-problems-view *tutor*)
        (let ((v (draw-text-with-font (tutor-problems-view *tutor*) 0 0 txt '("Geneva" 14))))
          (set-view-position v (- (view-width (tutor-problems-view *tutor*)) (view-width v)) 0))
        ;; just print a message if there's no problem view. Leave it out
        ;; if there's a hint. No need to tell them on a practice problem.
        (unless *hint-text*
          ;; let'em know what problem they're on
          (set-message-text txt)))))

(defun draw-stats (n max)
  (draw-score)
  (draw-problems n max))

(defun clear-stats ()
  (when (tutor-problems-view *tutor*)
    (delete-subviews (tutor-problems-view *tutor*)))
  (when (tutor-score-view *tutor*)
    (delete-subviews (tutor-score-view *tutor*))))
