;;; Otto the algebra robot.

(in-package cl-user)

;; (:EQUATION (BOX1256 BOX1259 NIL) (BOX1259 (BOX1262 * 8) NIL) (BOX1262 (5 + 7) (:ARROW)))
(defstruct kind
  eqn opn selected)

(defvar *otto-kinds* nil)
(defvar *otto-kinds-section* nil)
(defvar *otto-kinds-all* nil)

(defun type-seq (seq)
  ;;(dformat "type: ~s" seq)
  (map 'list (lambda (c)
                 (tutor:execute-command :type c))
       seq))

(defun type-em (&rest things)
  (type-seq (apply #'concatenate 'string (mapcar 'prin1-to-string things))))

(defvar *already-done* nil)

(defun otto-operations ()
  "order the operations such that evaluation is a last resort."
  (tutor::hintable-operations)
;;   (prog1
;;       (if *already-done*
;;           (if (tutor::tutor1-p)
;;               '(tutor::reorder tutor::invert tutor::evaluate tutor::combine tutor::subtract)
;;             '(tutor::distribute tutor::unwind tutor::evaluate tutor::collect tutor::subtract))
;;         (if (tutor::tutor1-p)
;;             '(tutor::subtract tutor::reorder tutor::invert tutor::evaluate tutor::combine)
;;           '(tutor::subtract tutor::distribute tutor::unwind tutor::evaluate tutor::collect)))
;;     (setf *already-done* t))
  )

(defun otto-select-boxes ()
  ;; we need a hint.
  (multiple-value-bind (opn boxes) (tutor::find-operation-on-boxes (otto-operations))
    (declare (ignore opn))
    (tutor::dformat "hehre1~%")
    (mapc (lambda (b)
            (tutor:execute-command :click b))
          (mapcar 'tutor::box-id boxes))))

(defun otto-choose-operation ()
  (multiple-value-bind (opn boxes) (tutor::find-operation-on-boxes (otto-operations))
    (let* ((ch (tutor::operation-quick-key (tutor::find-operation opn))))
      (push (make-kind :eqn (tutor::box-to-tree (tutor::tutor-eqn-root tutor::*tutor*) t) :opn ch :selected (length boxes)) *otto-kinds*)
      (tutor::dformat "press: '~a'~%" ch)
      (tutor:execute-command :type ch))))

(defun otto-fill-in ()
  (let* ((eqn (cdr (assoc :equation (tutor:dump-screen))))
         (things (loop for e in eqn
                       when (find (first (third e)) '(:box-editable :arg1-editable :arg2-editable :op-editable :arg-op-editable :op-arg-editable :answer-editable))
                       collect (cons (first e) (first (third e))))))
    (tutor::dformat "things: ~a" things)
    (dolist (i things)
      (let ((box (tutor::find-tree (tutor::tutor-eqn-root tutor::*tutor*) (lambda (box)
                                                                            (eq (tutor::box-id box) (car i))))))
        (ecase (type-of box)
          ((tutor::equal-box tutor::arg-box)
           (let ((a1 (tutor::box-arg1 box))
                 (a2 (tutor::box-arg2 box))
                 (op (tutor::box-op box)))
             (ecase (cdr i)
               (:box-editable
                (tutor:execute-command :click (tutor::box-id box))
                (type-seq (format nil "~a~a~a"
                                  (tutor::arg-answer a1)
                                  (tutor::arg-answer op)
                                  (tutor::arg-answer a2))))
               (:arg1-editable
                (tutor:execute-command :click (tutor::arg-id (tutor::box-arg1 box)))
                (type-em (tutor::arg-answer (tutor::box-arg1 box))))
               (:arg2-editable
                (tutor:execute-command :click (tutor::arg-id (tutor::box-arg2 box)))
                (type-em (tutor::arg-answer (tutor::box-arg2 box))))
               (:op-editable
                (tutor:execute-command :click (tutor::arg-id (tutor::box-op box)))
                (type-em (tutor::arg-answer (tutor::box-op box))))
               (:arg-op-editable
                (tutor:execute-command :click (tutor::box-id box))
                (type-em (tutor::arg-answer (tutor::box-arg1 box)) (tutor::arg-answer (tutor::box-op box))))
               (:op-arg-editable
                (tutor:execute-command :click (tutor::box-id box))
                (type-em (tutor::arg-answer (tutor::box-op box)) (tutor::arg-answer (tutor::box-arg2 box)))))))
          (tutor::answer-box
           (tutor:execute-command :click (tutor::arg-id (tutor::answer box)))
           (type-seq (prin1-to-string (tutor::arg-answer (tutor::answer box))))))
        (tutor:execute-command :submit nil)))))

(defun otto-forward-phase ()
  (ecase (tutor::problem-stage nil)
    (:select (otto-select-boxes))
    (:operation (otto-choose-operation))
    (:fill (otto-fill-in))
    (:done (tutor:execute-command :click :next-problem))))

(defvar *otto-operations* 0
  "The number of operations otto performed")

(defvar *otto-opns-list* nil)

(defvar *otto-check-opns* t)

(defun otto-do-next ()
  (labels ((s= (s1 s2)
             (let ((min (min (length s1)
                             (length s2))))
               (string-equal s1 s2 :end1 min :end2 min))))
    (let ((txt (second (assoc :text (tutor:dump-screen))))
          (take "sections to go!")
          (end "End of Experiment."))
      (cond 
        ((and txt (s= take (subseq txt 2)))
         (push *otto-operations* *otto-opns-list*)
         (push (nreverse *otto-kinds-section*) *otto-kinds-all*)
         (setf *otto-operations* 0
               *otto-kinds-section* nil)
         (tutor:execute-command :click :continue))
        ((and txt (s= end txt))
         (throw 'done t))
        ((eq (tutor::problem-stage) :done)
         (when *otto-check-opns*
           (assert (=  *otto-operations* (/ (tutor::tutor-problem-score tutor::*tutor*) tutor::+operation-points+))))
         ;; push the answer onto the list
         (push (make-kind :eqn (tutor::box-to-tree (tutor::tutor-eqn-root tutor::*tutor*) t) :opn nil :selected 0) *otto-kinds*)
         (push (nreverse *otto-kinds*) *otto-kinds-section*)
         (setf *otto-operations* 0
               *otto-kinds* nil
               *already-done* nil)
         (otto-forward-phase))
        (t
         (incf *otto-operations*)
         (otto-forward-phase)
         (otto-forward-phase)
         ;; canonicalize doesn't have a fillin stage
         (when (eq (tutor::problem-stage nil) :fill)
           (otto-forward-phase)))))))

(defun otto (&optional (wait 1))
  (setf *otto-operations* 0
        *otto-opns-list* nil
        *otto-kinds* nil
        *otto-kinds-section* nil
        *otto-kinds-all* nil)
  (catch 'done
    (loop
     (otto-do-next)
     (sleep wait))))

;; 

(defun infix-it (eqn &optional no-parens)
  (if (atom eqn)
      (let ((s (prin1-to-string eqn)))
        (list s  (make-string (length s) :initial-element #\Space)))
    (if (= (length eqn) 1)
        (infix-it (first eqn) no-parens)
      (let ((a (infix-it (second eqn) (eq (first eqn) '=)))
            (b (infix-it (third eqn) (eq (first eqn) '=))))
        (list (concatenate 'string (unless no-parens "(")
                      (first a)
                      " "
                      (symbol-name (first eqn))
                      " "
                      (first b)
                      (unless no-parens ")"))
              (concatenate 'string (unless no-parens " ")
                      (second a)
                      " "
                      (if (eq (fourth eqn) :selected)
                          "^" " ")
                      " "
                      (second b)
                      (unless no-parens " ")))))))

(defun dump-kind-data (data)
  "Print text summarizing the operations and number of operators
selected for each problem of each section in DATA."
  (let ((section '(ch1-s1 ch1-s2 ch1-s7 ch2-s1 ch2-s3 ch2-s4 ch2-s5 ch2-s6 ch2-s7 ch3-s1 ch3-s2 ch3-s4 ch4-s1 ch4-s2 ch4-s3))
        (out1 (make-array 0 :adjustable t :fill-pointer t :element-type 'character))
        (out2 (make-array 0 :adjustable t :fill-pointer t :element-type 'character)))
    (dolist (s (reverse data))
      (dolist (p s)
        (dolist (o p)
          (let* ((*package* (find-package :tutor))
                 (eqn (infix-it (kind-eqn o) t))
                 (opn (when (kind-opn o)
                        (tutor::operation-t2-name (tutor::find-operation (kind-opn o))))))
            (format out1 "~a ~a=> " (first eqn) opn)
            (format out2 "~a ~a   " (second eqn) (make-string (length (symbol-name opn)) :initial-element #\Space))))
        (tutor::dformat "~a~%" out1)
        (tutor::dformat "~a~%" out2)
        (setf (fill-pointer out1) 0)
        (setf (fill-pointer out2) 0))
      (pop section)
      (tutor::dformat "~%"))))
