;;; A window that makes it easier to explore the tutor
(in-package :tutor)

(defmacro toggle (value)
  `(setf ,value (not ,value)))

(defun panel ()
  (let ((panel (make-instance 'window
                              :window-title "Panel"
                              :window-type :document
                              :view-font '("Monaco" 14)
                              :view-size (make-point 150 700))))
    (create-layout `(:vbox :contents ((:button :text "1-1" :action ,(button-fn (do-section 1 1)))
                                      (:button :text "1-2" :action ,(button-fn (do-section 1 2)))
                                      (:button :text "1-7" :action ,(button-fn (do-section 1 7)))
                                      (:button :text "2-1" :action ,(button-fn (do-section 2 1)))
                                      (:button :text "2-3" :action ,(button-fn (do-section 2 3)))
                                      (:button :text "2-4" :action ,(button-fn (do-section 2 4)))
                                      (:button :text "2-5" :action ,(button-fn (do-section 2 5)))
                                      (:button :text "2-6" :action ,(button-fn (do-section 2 6)))
                                      (:button :text "2-7" :action ,(button-fn (do-section 2 7)))
                                      (:button :text "3-1" :action ,(button-fn (do-section 3 1)))
                                      (:button :text "3-2" :action ,(button-fn (do-section 3 2)))
                                      (:button :text "3-4" :action ,(button-fn (do-section 3 4)))
                                      (:button :text "3-5" :action ,(button-fn (do-section 3 5)))
                                      (:button :text "4-1" :action ,(button-fn (do-section 4 1)))
                                      (:button :text "4-2" :action ,(button-fn (do-section 4 2)))
                                      (:button :text "4-3" :action ,(button-fn (do-section 4 3)))
                                      (:space)
                                      (:button :text "Tutor 1" :action ,(button-fn (tutor::setup-tutor1)))
                                      (:button :text "Tutor 2" :action ,(button-fn (tutor::setup-tutor2)))
                                      (:space)
                                      (:button :text "Session 1" :action ,(button-fn (session1 1)))
                                      (:button :text "Session 2" :action ,(button-fn (session2 1)))
                                      (:button :text "Session 3" :action ,(button-fn (session3 1)))
                                      (:space)
                                      (:button :text "Next Problem" :action ,(button-fn (setf (tutor-submit-enabled *tutor*) nil
                                                                                              *nohint-override* nil)
                                                                                        (funcall (tutor-callback *tutor*))))
                                      (:button :text "Step" :action ,(button-fn (common-lisp-user::otto-forward-phase)))
                                      (:button :text "Toggle Hints" :action ,(button-fn (toggle (tutor-global-whats-next *tutor*))))
                                      (:button :text "Calculator" :action ,(button-fn (calc:calc)))
                                      (:button :text "Reload" :action ,(button-fn (format t "~s~%" '(Load #P"Macintosh HD:Users:sabetts:src:tutor2:loader.lisp"))
                                                                                  (Load #P"Macintosh HD:Users:sabetts:src:tutor2:loader.lisp")))))
                   panel)))

(defclass fsm-next-window (window)
  ())

(defun fsm-panel (fsm)
  "Generate buttons to jump to each section in the fsm."
  (declare (ignore fsm))
  (let* ((p (make-instance 'fsm-next-window
                           :window-title "FMRI"
                           :window-type :document
                           :view-font '("Monaco" 14)
                           :view-size (make-point 300 30 ;; (* (1+ (truncate (length fsm) 2)) 30)
                                                  ))))
;;     (loop for i = fsm then (cddr i)
;;           while i
;;           for j from 1
;;           do (format t "AG: ~a~%" i)
;;           do (let ((lexical-hack i))
;;                (make-instance 'button-dialog-item
;;                               :view-container p
;;                               :view-position (make-point 0 (* (1- j) 30))
;;                               :view-size #@(150 22)
;;                               :dialog-item-text (format nil "Part ~a" j)
;;                               :dialog-item-action (button-fn
;;                                                    (delete-subviews *window*)
;;                                                    (window-select *window*)
;;                                                    (fsm-start lexical-hack)))))
    
    (draw-text p 30 0 "Press SPACE to continue.")
    p))

(defmethod view-key-event-handler :before ((w fsm-next-window) ch)
  ""
  (when (char= ch #\Space)
    (window-select *window*)
    (fsm-next)))

(defun fmri-panel ()
  (princ "Subject number: ")
  (let ((subject (read))
        (p (make-instance 'window
                          :window-title "FMRI Sessions"
                          :window-type :document
                          :view-font '("Monaco" 14)
                          :view-size (make-point 150 (* 8 30)))))

    (assert (typep subject 'fixnum))
    (labels ((but (txt fn y)
                  (make-instance 'button-dialog-item
                                 :view-container p
                                 :view-position (make-point 0 y)
                                 :view-size #@(150 22)
                                 :dialog-item-text txt
                                 :dialog-item-action (button-fn
                                                      (funcall fn subject)))))
      (loop for i in '(("Session 1" fmri-session1)
                       ("Session 2 Struct." fmri-session2-part1)
                       ("Session 2 Magnet" fmri-session2-part2)
                       ("Session 3" fmri-session3)
                       ("Session 4" fmri-session4)
                       ("Session 4 (Alt)" fmri-session4-type2)
                       ("Session 5 Struct." fmri-session5-part1)
                       ("Session 5 Magnet" fmri-session5-part2))
            for j from 0 by 30
            do (but (first i) (second i) j)))))

(defun customize ()
  (let ((panel (make-instance 'window
                              :window-title "Customize"
                              :window-type :document
                              :view-font '("Monaco" 14)
                              :view-size #@(250 100))))
    (labels ((toggle-var (var)
               (lambda () (setf (symbol-value var) (first (select-item-from-list '(t nil) :window-title "Select Item")))))
             ;;(lambda () (setf (symbol-value var) (not (symbol-value var)))))
             (cycle-var (var opts)
               ;; (lambda () (setf (symbol-value var) (or (second (member (symbol-value var) opts))
;;                                                        (first opts))))
               (lambda () (setf (symbol-value var) (first (select-item-from-list opts :window-title "Select Item")))))
             (set-var (var)
               (lambda () (setf (symbol-value var) (read-from-string (get-string-from-user "Enter Value" :initial-string (prin1-to-string (symbol-value var)))))))
             (set-list (var opts)
               (lambda () (setf (symbol-value var) (select-item-from-list opts :window-title "Select Items" :selection-type :disjoint))))
             (but (txt fn y)
               (make-instance 'button-dialog-item
                              :view-container panel
                              :view-position (make-point 5 y)
                              :view-size (make-point (- (view-width panel) 10) 22)
                              :dialog-item-text txt
                              :dialog-item-action (button-fn (funcall fn)))))
      (loop for i being each symbol in :tutor
         with j = 5
         do (when (get i :customize)
              (ecase (get i :type)
                (:bool (but (symbol-name i) (toggle-var i) j))
                (:enum (but (symbol-name i) (cycle-var i (get i :values)) j))
                (:number (but (symbol-name i) (set-var i) j))
                (:list (but (symbol-name i) (set-list i (get i :values)) j)))
              (incf j 30))
         finally (set-view-size panel (view-width panel) j)))))

(defun dump-configuration ()
  "Return a defun that sets all customize variables."
  `(defun my-config ()
     ,(loop for i being each symbol in :tutor
         when (get i :customize)
         collect `(setf ,i ',(symbol-value i)))))     
          
(defun slide-show-problem-selection (problems)
  (let* ((window (make-instance 'window
                                :window-title "Select Problem"
                                :window-type :document
                                :view-size #@(300 400)))
         (view (make-instance 'view
                              :view-container window))
         (before-problems nil)
         (current-problem nil)
         (after-problems problems))
    (labels ((draw-current-problem ()
               (let ((tree (make-and-space-tree (problem-eqn current-problem))))
                 (tutor1-recenter-tree tree view t 300 400)
                 (set-view-position view #@(0 40))
                 ;;(set-view-size w (view-size view))
                 (delete-subviews view)
                 (tutor1-redraw-tree tree view)))
             (next (but)
               (declare (ignore but))
               (unless after-problems
                 (setf after-problems (nreverse before-problems)
                       before-problems nil))
               (when after-problems
                 (push current-problem before-problems)
                 (setf current-problem (pop after-problems))
                 (draw-current-problem)))
             (previous (but)
               (declare (ignore but))
               (unless before-problems
                 (setf before-problems (nreverse after-problems)
                       after-problems nil))
               (when before-problems
                 (push current-problem after-problems)
                 (setf current-problem (pop before-problems))
                 (draw-current-problem)))
             (solve (but)
               (declare (ignore but))
               (do-problem (problem-eqn current-problem) (first (problem-answer current-problem)))))
      (make-instance 'button-dialog-item
                     :view-container window
                     :view-position #@(0 10)
                     :dialog-item-text "<-"
                     :dialog-item-action #'previous)
      (make-instance 'button-dialog-item
                     :view-container window
                     :view-position #@(80 10)
                     :dialog-item-text "->"
                     :dialog-item-action #'next)
      (make-instance 'button-dialog-item
                     :view-container window
                     :view-position #@(160 10)
                     :dialog-item-text "Solve"
                     :dialog-item-action #'solve)
      (setf current-problem
            (pop after-problems))
      (draw-current-problem))))

(defun short-tutor-problems ()
  (slide-show-problem-selection 
   (extract-problems-from-sets
    (apply 'append
           (build-session-sections 
            '(ch1-s1 ch1-s2 ch1-s7 ch2-s6 ch2-s7 ch3-s1 ch3-s2 ch3-s4 ch3-s5 ch4-s1 ch4-s2 ch4-s3)
            'get-short-section)))))

(defun select-problem (&rest sections)
  (slide-show-problem-selection 
   (extract-problems-from-sets
    (apply 'append
           (build-session-sections 
            (or sections
                '(ch1-s1 ch1-s2 ch1-s7 ch2-s6 ch2-s7 ch3-s1 ch3-s2 ch3-s4 ch3-s5 ch4-s1 ch4-s2 ch4-s3)))))))
  
