;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; GENERAL FUNCTIONS AND PARAMETERS
;;; An input module for retrieving hint information.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *state*)                        ;; A variable holding current interaction state
(defvar *ignore-error*)
(defparameter *last-message* nil)
(defparameter *problem-size* 1000)
(defparameter *highlight-state* nil)
(defparameter *old-state* nil)
(defparameter *episode-history* nil)
(defparameter *operators* nil)
(defparameter *nothing* '("Here is the diagram before you made your first mistake."))
(defparameter *transitions* nil)
(defparameter *actions* nil)

(defun success ()
  (cond ((or (not (= (length (assoc :equation (dump-screen))) *problem-size*) )
            (and *highlight-state* (not (highlight-state)))
            (and (member *old-state* '(finished cfinished bfinished))
                 (member (perceive-state) '(ebox input-number bbox combine-a))))
        (setf *episode-history* (list (perceive-state))) (setf *highlight-state* (highlight-state)) t)
       (t (setf *highlight-state* (highlight-state)) nil)))


            

(defun update-success () (setf *problem-size* (length (assoc :equation (dump-screen)))))

(defun make-episode () (update-success)  (gentemp "STATE"))



(defun flail-test (state) 
  (setf *episode-history* (cons (perceive-state) *episode-history* ))
  (do ((temp (cdr *episode-history*) (cdr temp)))
      ((null temp) nil)
    (if (member (car temp) (cdr temp)) (return t))))

(defun highlight-state ()
  (let* ((old  (cdr (assoc ':old-equation (dump-screen)))))
    (member '(:SELECTED) old :test 'equal :key 'third)))



;; Gets the value of the slot ELEM in the
;; slot definition LIS.
;;
(defun get-value (lis elem) 
  (second (member elem lis)))


;; CREATE-CHUNK
;; Creates a new chunk in the specified buffer, according to the specified
;; description. It also set the BUSY parameter to nil
;;
(defun create-chunk (buffer-name chunk-description busy)
  ;  (format t "~%Called CREATE-CHUNK in buffer ~A and with description = ~A~%" buffer-name chunk-description)
  (set busy nil)
  (create-new-buffer-chunk buffer-name chunk-description))


(defun get-time ()
 (* 1000 (mp-time)))

(defun record-it (action)
  (push action *actions*)
  (if (no-output (car (sgp :v))) (print action)))

(defun setup (lis)
  (let ((offset (/ (car lis) 1000))
        (newprods '((operator1c isa operator pre ebox action just-numbers)
                    (operator3c isa operator pre ebox1 action evaluate )
                    (operator5c isa operator pre green-little-box action get-green-box)
                    (operator7c isa operator pre arg-editable action type-value)
                    (operator9c isa operator pre finished action next)
                    (operator11c isa operator pre input-number action just-numbers)
               (operator1c1 isa operator pre EBBOX action just-numbers)
                (operator3c1 isa operator pre ebbox1 action evaluate)
                (operator9c1 isa operator pre bfinished action next)
                (operator13c isa operator pre bbox action find-bottom-box)
                (operator13c1 isa operator pre bbox2 action find-above-box)
                (operator12c isa operator pre BBOX1 action find-bottom-box)
                (operator14c isa operator pre bbox12 action invert)
                (operator5c1 isa operator pre GREEN-BIG-BOX action get-green-box)
                (operator17c isa operator pre box-editable action express)
                    (blank isa operator))))
    (do ((temp (cdr lis) (cdr temp)))
        ((null temp) nil)
      (cond ( (caar temp) (no-output (cond ((not (eval `(dm ,(caar temp)))) (eval `(add-dm ,(assoc (caar temp) newprods))))))
          (setf (chunk-creation-time (caar temp)) (- (second (car temp)) offset))
          (setf (chunk-reference-count (caar temp)) (third( car temp))))))))








;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; THE VERBAL MODULE
;;; An input module for retrieving hint information.  The module is used the 
;;; read task information from special areas of the screen.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar *comprehension*)
(defparameter *verbal-busy* nil)        ;; Whether the verbal module is busy or not
(defvar *verbal-error*)                 ;; If an error has occurred in the verbal module

;; The QUERY function for the VERBAL module
;;
(defun verbal-query (instance buffer-name slot value)
  (declare (ignore buffer-name instance))
  (case slot
    (state
     (case value
       (busy *verbal-busy*)
       (free (not *verbal-busy*))
       (error-free (not *verbal-error*))
       (error *verbal-error*)
       (t (print-warning "Unknown state query ~S to goal module" value)
          nil)))))


;; The REQUEST function
;;
(defun request-verbal-chunk (instance buffer-name chunk-spec)
  (declare (ignore instance))
  (let* (sign (specs  (chunk-spec-to-chunk-def chunk-spec))
         (description
          (case (get-value  specs 'command)
          (read-hint  (setf sign (tentative-add (follow-help)))  (cdr (follow-help)))))
         (delay  (if sign  6 .1)))
    (setf *verbal-busy* t)
    (setf *verbal-error* nil)
    (cond (description (schedule-event-relative delay 'create-verbal-chunk 
                                                :module buffer-name
                                                :params (list buffer-name description)))
          (t (schedule-event-relative 1 'create-verbal-error 
                                      :module buffer-name
                                      :priority -100)))))

(defun select-help (state options)
  (do ((temp (permut options) (cdr temp)))
      ((null temp) nil)
    (if (eq (fifth (car temp)) state) (return (car temp)))))

(defun permut (lis)
  (do* ((temp lis (remove item temp :test 'equal))
        (item (nth (random (length lis)) lis) (nth (random (length temp)) temp))
        (result (list item) (cons item result)))
       ((null (cdr temp)) result)))
               
(defun follow-help ()  
  (or (describe-arrow (cdr (assoc :red-arrow (dump-screen))))
      (select-help (perceive-state) (cdr (assoc (clean (second (assoc ':text (dump-screen)))) *mapping* :test 'equal))) 
      '(blank isa operator)))


(defun clean (seq)
  (if (listp seq) (setf seq (car (last seq))))
  (do ((temp '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\+ #\- #\* #\/ #\newline) (cdr temp))
      (result seq (remove (car temp) result)))
      ((null temp) result)))

(defun tentative-add (chunk) 
  (or (if (not (no-output (dm-fct (list (car chunk))))) (add-dm-fct (list chunk)))
      (equal chunk '(BLANK ISA OPERATOR))))

;; Sets the appropriate parameter in case o a verbal error
;;
(defun create-verbal-error ()
  (setf *verbal-busy* nil)
  (setf *verbal-error* t))

;; Resets the parameter the their normal values after
;; a verbal error has occurred.
;;
(defun free-verbal-error ()
  (setf *verbal-error* nil))

;; Creates a new chunk to be put in the BUFFER-NAME buffer.
;; The chunk is created according to the CHUNK-DESCRIPTION
;;
(defun create-verbal-chunk (buffer-name chunk-description)
  (setf *verbal-busy* nil)
  (create-new-buffer-chunk buffer-name chunk-description))

;; The RESET function
;;
(defun reset-verbal-2 (instance)
  (declare (ignore instance))
  (specify-compilation-buffer-type verbal perceptual))

;; VERBAL MODULE DEFINITION
;; The formal definition of the VERBAL module
;;
(define-module-fct 'verbal
  '(verbal) 
  nil
  :version "0.2a"
  :documentation "An input model for reading text"
  :query 'verbal-query
;  :reset '(nil reset-verbal-2)
  :request 'request-verbal-chunk)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; THE DIGITAL MODULE
;;; The digital module is an output module: Its requests are interpreted as
;;; finger movements to be executed on the tutor panels.
;;; 
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *digital-busy* nil)   ;; Whether the module is busy (e.g., executing a command) or not.
(defvar       *digital-error*)      ;; THe error state of the module  

;; The QUERY function
;; The module is FREE when *DIGITAL-BUSY* is NIL, BUSY otherwise.
;; It is in Error state iff *DIGITAL-ERROR* is T.
;;
(defun digital-query (instance buffer-name slot value)
  (declare (ignore buffer-name instance))
  (case slot
    (state
     (case value
       (busy *digital-busy*)
       (free (not *digital-busy*))
       (error *digital-error*)
       (t (print-warning "Unknown state query ~S to goal module" value)
          nil)))))


;; The REQUEST function
;;
;; This functions handles the request specification as a command to execute 
;; an output motor function on the screen.  the motor functions is usually a
;; a mouse or a key press---hence the name 'digital'.
;;
;; Before the execution, the module is set to BUSY and its error state is NIL.
;; The course of execution wiull reset this value properly.
;;
;; As usual, execution is achieved by scheduling an event (with a specified delay)
;; in the ACT-R scheduler
;;
(defun request-digital-chunk (instance buffer-name chunk-spec)
   (declare (ignore instance))
   (let* ((specs  (chunk-spec-to-chunk-def chunk-spec))
          (arg (get-value specs 'item))
          (state (get-value specs 'state))
          (description
           (case (get-value specs 'command) 
             (click    (encode-click arg))
             (enter     (setf *input-length* 0)
                        (type-string (format nil  "~a" arg ) t))))
          (delay (* (length description) 0.4)))
     ;;
     ;; Sets the module state as BUSY and without ERROR,
     ;; Until the scheduled events are executed
     ;;
     (setf *digital-busy* t)
     (setf *digital-error* nil)
     ;;
     ;; Executions: Events are scheduled according to the value of the DESCRIPTION
     ;; local var. When not-nil, it is interpreted as a command. If NIl, an error state is created
     ;;
     (cond (description (schedule-event-relative delay 'perform-digital
                                                 :module buffer-name
                                                 :priority -100
                                                 :params (list description)))
           (t (schedule-event-relative .3 'create-digital-error 
                                       :module buffer-name
                                       :priority -100)))))


(defun buffer-slot (buffer index) 
  (no-output (chunk-slot-value-fct (car (buffer-chunk-fct (list buffer))) index)))


(defun chunk-slot-value-fct (chunk-name slot-name)
  "Return the value of a slot for the named chunk"
  (let ((c (get-chunk chunk-name)))
    (when c
      (if (valid-slot-name slot-name (act-r-chunk-chunk-type c))
          (chk-slot-value c slot-name) nil))))


;; TYPE-STRING
;; Types a string of test by creating a list
;; of EXECUTE-COMMAND functions whose arguments are the
;; characters of the string STR.
;; 
(defun type-string (str flag)
  (if (equal str "NIL") (setf str ""))
  (if str
  (do ((i 0 (1+ i))
       (result nil (cons (list 'execute-command ':type (aref str i)) result)))
      ((= i (length str)) 
       (if flag (reverse (cons (list 'execute-command ':submit nil) result))
             (reverse result))))))

;; The ENCODE-CLICK function
;; This functions creates a list that is an unavaluated EXECUTE-COMMAND
;; functions.  The command is a  click, and the target is the button whose
;; name is ITEM (if item is a string) or the associated value of a list,
;; if ITEM is nota  string.
;;
(defun encode-click (item)
  ;(format t "~A" item)       ;; Prints the value of item on the screen --- for debugging only
  (list (list 'execute-command 
              ':click (or (map-click item) item))))




;; CREATE-DIGITAL-ERROR
;; Setes the parameters for a digital error
;; When queried, the module will result FREE and ERROR.
;;
(defun create-digital-error ()
  (setf *digital-busy* nil)
  (setf *digital-error* t))


;; PERFORM-DIGITAL
;; The function that (finally) updates the system state
;; after the execution of a digital command.
;;
(defun perform-digital (description)
  (setf *digital-busy* nil) 
  (if (equal (second (car description)) :click)
    (setf *ignore-error* nil))
  (ignore-errors (mapcar 'eval description))
  (clear-buffer 'screen)
  (schedule-event-relative .3 
                           'update-state  
                           :module 'screen
                           :priority -100))


;; The DIGITAL MODULE definition
;;
(define-module-fct 'digital
  '(digital) 
  nil
  :version       "0.2a"
  :documentation "A module for executing finger movements (cliks) on a screen" 
  :query         'digital-query
  :request       'request-digital-chunk)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; THE GRAPHICAL MODULE
;;; The graphical module is both and input and an output module for the system.
;;; It is mainly used to output operators by requesting the execution of an 
;;; operator and collecting the subsequent result.  Operator commands are
;;; passed thorough a BOX chunk type, and they are collected by creating
;;; a similar chunk type in a subsequent phase.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *graphical-busy* nil)
(defvar *graphical-error*)

;; The REQUEST Function
;; Gets an operator specification, and interprets it as command.
;; The execution of the command is stored in the description variable
;; 
(defun request-graphical-chunk (instance buffer-name chunk-spec)
   (declare (ignore instance))
   (let* ((specs       (chunk-spec-to-chunk-def chunk-spec))
          (pattern   (get-value specs 'command) )
          (description (append '(isa box val) (or (extract-pattern pattern) (list nil))))
          (delay       .2))         ;; End of let*
     ;;
     ;; Set the buffer busy and error-free before starting processing
     ;;
     (setf *graphical-busy*  t)
     (setf *graphical-error* nil)
     ;;
     ;; The schedules thre creation of a new chunk in the buffer, or the
     ;; setting up of an error condition, depending on the value of DESCRIPTION
     ;;
     ; (format t "~%Executing a request from a graphical module~%")
     (cond (description (schedule-event-relative delay 
                                                 'create-chunk 
                                                 :params (list buffer-name description '*graphical-busy*)))
           (t (schedule-event-relative delay 
                                       'create-graphical-error 
                                       :module buffer-name
                                       :priority -100)))))


(defparameter *new-state* nil)

(defun update-state () 
  (setf *old-state* *new-state*) (setf *new-state* (perceive-state))
  (create-new-buffer-chunk 'screen (list 'isa 'screen 'state (perceive-state) 'hint (perceive-hint) 
                                         'flail (flail-test (perceive-state) ) 'success (success)))
  (setf *state* nil))

(defun perceive-hint ()
     (let ((message  (if  (assoc :equation (dump-screen)) 
                       (or (clean (second (assoc ':text (dump-screen))))
                           (describe-arrow (cdr (assoc :red-arrow (dump-screen))))))))
       (if (and (equal (perceive-state) (car *episode-history*))(equal message *last-message*)) nil
           (progn (setf *last-message* message)
                  (if (and message (not (member message *nothing* :test 'equal))) t nil)))))

(defun string-to-type (string)
  (let* ((bound (- (length string) 5))
        (result (make-string bound)))
    (do ((i 0 (1+ i)))
        ((= i bound) result)
      (setf (aref result i) (aref string (+ i 5))))))

(defun describe-arrow (lis)
  (let ((item (first lis)) (content (second lis)))
    (if (and item (not (equal (second (assoc ':text (dump-screen))) "Your answer is incorrect.")))
      (cond ((equal item :INPUT-BOX) (add-cname (list  'isa 'operator 'pre (perceive-state) 'action  (extract-transform (perceive-state)))))
           (t  (do ((temp *basic-actions* (cdr temp)))
                      ((null temp) nil)
                    (let ((cand (eval (car temp))))
                      (cond ((atom cand) (if (eq cand item) (return (add-cname (list 'isa 'operator 'pre (perceive-state) 'action (car temp))))))
                            (t (if (member item (extract-pattern (car temp)))
                     (return (add-cname (list  'isa 'operator 'pre (perceive-state) 'action (car temp))))))))))))))

(defun extract-transform (state)
  (case state
    (arg-editable 'type-value)
    (GREEN-BIG-BOX 'express)
    (COMBINE-BOX-EDIT 'combine-box-type)
    (COMBINE-OP-EDIT (if (zerop (random 2)) 'COMBINE-OP-TYPE 'COMBINE-OP-bad-TYPE))))

(defun add-cname (lis)
  (let ((string (concatenate 'string (string (fourth lis)) (string (sixth lis)))))
    (make-symbol string) (cons (intern string) lis)))


;; CREATE-GRAPHICAL-ERROR
;; When an error occurs processing a request, the parameters are updated
;; so that module results Free, but responds positively to Error queries
;; 
(defun create-graphical-error ()
  (setf *graphical-busy* nil)
  (setf *graphical-error* t))


;; The QUERY Function
;; The module supports four (4) states: Busy if *graphical-busy* is non-nil,
;; Free otherwise, Error-Free if *graphical-error* is not-nil, and Error
;; otherwise.
;;
(defun graphical-query (instance buffer-name slot value)
  (declare (ignore buffer-name instance))
  (case slot
    (state
     (case value
       (busy       *graphical-busy*)
       (free       (not *graphical-busy*))
       (error-free (not  *graphical-error*))
       (error      *graphical-error*)
       (t (print-warning "Unknown state query ~S to goal module" value)
          nil)))))

;; THE RESET function
;;
(defun reset-graphical-2 (instance)
  (declare (ignore instance))
  (specify-compilation-buffer-type graphical perceptual))


;;; GRAPHICAL
;;; The graphical module is an input module
;;;
(define-module-fct 'graphical
  '(graphical) 
  nil
  :version "0.2a"
  :documentation "A module for retrieving visual information about the screen"
  :query 'graphical-query
  :reset '(nil reset-graphical-2)
  :request 'request-graphical-chunk)


(defun perceive-state ()
  (if (wrong-answer) 'wrong-answer
  (car (recognize-the-pattern (read-from-string (pad-string (dump-eqn)))))))

(defun wrong-answer ()
  (equal (clean (second (assoc ':text (dump-screen)))) 
         "Your answer is incorrect.Use the < and > buttons (or the left and right arrow keys)and the First Mistake buttonto review your work and correct the mistake."))




(defun pad-string (string)
  (let* ((j 0)
        (bound (length string))
        (new (make-string (* 2 bound))))
    (cond ((member (aref string 0) '(#\{  #\[ )) (setf (aref new 0) #\()(setf j 1)))
    (do ((i 0 (1+ i)))
        ((= i bound) (if (member (aref string 0) '(#\{  #\[ )) (setf (aref new j) #\)))
           new)
      (cond ((member (aref string i) '(#\{ #\} #\[ #\]))
             (setf (aref new j) #\Space)
             (setf (aref new (1+ j)) (aref string i))
             (setf (aref new (+ j 2)) #\Space)
             (setf j (+ j 3)))
            (t (setf (aref new j) (aref string i)) (setf j (1+ j)))))))
    



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; THE MATHEMATICAL MODULE
;;; An output module---the more questionable one. May be removed as
;;; further improvements are made. This module substitutes the calculator
;;; on the tutor. Basically, its requests are intepreted as operations
;;; to be performed over specified numbers.    
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *mathematical-busy* nil)           ;; Whether the module is busy (e.g., computing) or not
(defvar       *mathematical-error*)              ;; Whether an error has occurred in the model or not

;; MATHEMATICAL-QUERY
;; This is the request function of the mathematical buffer.
;; When a request is performed, if the command is EVALUATE, 
;; The arguments in the ARG1 slot are evaluated by the EVALUATE-BOX
;; function, which stores the value in the VAL slot of the requested
;; chunk
;;
(defparameter *calculator* nil)
(defun request-mathematical-chunk (instance buffer-name chunk-spec)
  (declare (ignore instance))
  (setf *calculator* nil)
  (let* ((specs  (chunk-spec-to-chunk-def chunk-spec))
         (description
          (case (get-value specs 'command) 
            (Evaluate-box (evaluate-box  (get-value specs 'arg1)(get-value specs 'arg2)))))
         (delay (if *calculator* 3 .5)))
    (setf *mathematical-busy* t)  (setf *mathematical-error* nil)
    (cond (description (schedule-event-relative delay 'create-chunk 
                                                :params (list buffer-name description '*mathematical-busy*)))
          (t (schedule-event-relative delay 'create-mathematical-error 
                                      :module buffer-name
                                      :priority -100)))))

(defun evaluate-box (box action)
  (let* ((expr (second (or (assoc box (cdr (assoc ':equation (dump-screen))))
                           (assoc box (cdr (assoc ':old-equation (dump-screen)))))))
         (val (cond ((eq action 'combine-box-type) (combine-args expr))
                    ((eq action 'combine-op-type) (combine-op expr))
                    ((eq action 'combine-op-bad-type) (combine-op-bad expr))
                    ((eq action 'get-number) (if (numberp (car expr)) (car expr) (third expr)))
                    ((atom expr)
                     (cond ((numberp expr) expr)
                           ((equal expr :x) (evaluate-it (solve-formula (construct-formula))))))                  
                    ((and (eq (second expr) '*) (or (eq (first expr) 0) (eq (third expr) 0))) 0)
                    ((equal (third (assoc box (cdr (assoc ':equation (dump-screen))))) '(:BOX-EDITABLE))
                     (solve-formula (construct-formula)))
                    ((and (numberp (first expr)) (numberp (third expr)))
                     (if (and (eq (second expr) '*) 
                              (or (> (first expr) 11) (> (third expr) 11)))
                       (setf *calculator* t))
                     (eval (list (second expr) (first expr) (third expr))))
                    ((numberp (first expr)) (first expr))
                    ((numberp (third expr)) (third expr))                    
                   ((eq (second expr) '=) (reconstruct (construct-formula))))))
    (list 'isa 'verbage 'arg1 box 'result val)))

(defun construct-formula () 
  (do ((temp  (cdr (assoc ':old-equation (dump-screen))) (cdr temp))
       (result nil (if (equal (third (car temp)) '(:SELECTED))
                     (if (eq (second (second (car temp))) '=) (append result (list (second (car temp)))) (cons (second (car temp)) result))
                     result)))
      ((null temp) (if (numberp (first (second result))) (append (list (first (second result)) '=) (first result))
                       (append (first result) (cdr (second result)))))))

(defun solve-formula (formula)
  (let ((number1 (if (numberp (first formula)) (first formula) (third formula)))
        (number2 (if (fifth formula) (fifth formula) (third formula))))
  (case (second formula) 
         (+ (list number2 '- number1))
         (* (list number2 '/ number1))
         (- (if (numberp (third formula)) (list number2 '+ number1) (list number1 '- number2)))
         (/ (if (numberp (third formula)) (list number2 '* number1) (list number1 '/ number2))))))

(defun evaluate-it (expr) (if expr (eval (append (list (second expr) (first expr)) (cddr expr)))))

(defun reconstruct (formula) 
  (let ((val (evaluate-it (solve-formula formula))))
    (if (numberp (first formula)) (list (first formula) (second formula) val) 
        (list val (second formula) (third formula)))))

(defun construct-combination (expr)
  (cond ((numberp (car expr))
         (list (first expr) (second expr) (construct-expr (third expr))))
        (t (list (construct-expr (first expr)) (second expr) (third expr)))))

(defun construct-expr (node)
  (let ((expr (second (assoc node (cdr (assoc :old-equation (dump-screen)))))))
  (cond ((numberp (car expr))
         (list (first expr) (second expr) 'x))
        (t (list 'x (second expr) (third expr))))))

(defun apply-combine (expr)
  (let ((algebra (construct-combination expr)))
    (if (numberp (first algebra))
      (if (member (second algebra) '(+ *))
        (if (numberp (first (third algebra)))
          (cons (list (first algebra) (second algebra) (first (third algebra)))
                (cdr (third algebra)))
          (cons (cons (first algebra) (cdr (third algebra))) (list (second algebra) 'x)))
        (if (numberp (first (third algebra)))
          (list (list (first algebra) (second algebra) (first (third algebra)))
                (case (second (third algebra)) (+ '-)(- '+)(* '/)(/ '*)) 'x)
          (cons (list (first algebra) 
                      (case (second (third algebra)) (+ '-)(- '+)(* '/)(/ '*))
                      (third (third algebra))) (list (second algebra) 'x))))
      (if (numberp (first (first algebra)))
        (cons (cons (first (first algebra)) (cdr algebra)) (cdr (first algebra)))
        (if (member (second (first algebra)) '(+ *))
          (cons (cons (third (first algebra)) (cdr algebra)) (list (second (first algebra)) 'x))
          (cons (cons (* -1 (third (first algebra))) (cdr algebra)) '(+ x)))))))

(defun combine-args (expr) (first (apply-combine expr)))

(defparameter *combine* nil)

(defun combine-op-bad (expr) 
  (if (eq (second expr) (second (apply-combine expr))) (push 1 *combine*) (push 2 *combine*))(second expr))

(defun combine-op (expr) (if (not (eq (second expr) '/)) (mod-focus watch nil)) (push 0 *combine*) (second (apply-combine expr)))

       

(defun create-mathematical-error ()
  (setf *mathematical-busy* nil)
  (setf *mathematical-error* t))

;; MATHEMATICAL-QUERY
;; This is the query of the matematical buffer.
;; The buffer is busy whenever the variable *mathematical-busy* is T, and
;; Free otherwise.  The query returns Error whenever the *mathematical-error* 
;; parameter is T. Both of the paramters are handled by the mathematical request
;; function.
;; 
(defun mathematical-query (instance buffer-name slot value)
  (declare (ignore buffer-name instance))
  (case slot
    (state
     (case value
       (busy *mathematical-busy*)
       (free (not *mathematical-busy*))
       (error *mathematical-error*)
       (t (print-warning "Unknown state query ~S to goal module" value)
          nil)))))

;; The RESET function
;; Simply specifies the compilation buffer type.
;;
(defun reset-mathematical-2 (instance)
  (declare (ignore instance))
  (specify-compilation-buffer-type mathematical perceptual))

;; the MATHEMATICAL MODULE definition
;; As usual
;;
(define-module-fct 'mathematical
  '(mathematical) 
  nil
  :version "0.2a"
  :documentation "A module for performing fast mathematical computations"
  :query 'mathematical-query
  :reset '(nil reset-mathematical-2)
  :request 'request-mathematical-chunk)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; THE SCREEN MODULE
;;; The most important visual input module. Contains info about the current screen.
;;; Input is gathers through the <DUMP-SCREEN> function, which is defined in the
;;; AL-model file.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun request-screen-chunk (instance buffer-name chunk-spec)  
  (declare (ignore instance buffer-name slot chunk-spec))
  nil)

(defun screen-query (instance buffer-name slot value)  
  (declare (ignore instance buffer-name slot value))
  nil)

(defun reset-screen-2 (instance)
	(declare (ignore instance))
	(specify-compilation-buffer-type screen goal))

;; SCREEN
;; 
(define-module-fct 'screen
  '(screen) 
  nil
  :version       "0.2"
  :documentation "The most important visual input module. Contains info about the current screen"
  :query #'goal-style-query
  :reset '(nil reset-screen-2)
  :request #'goal-style-request)

(defun test-problems (ch sect time )
  (let ()
    (make-operators)
    (setf *verbal-busy* nil
          *transitions* nil
          *graphical-busy* nil 
          *verbal-error* nil 
          *graphical-error* nil 
          *new-state* nil
          *combine* nil
          *digital-busy* nil 
          *digital-error* nil       
          *mathematical-busy* nil 
          *mathematical-error* nil
          *ignore-error* nil
          *last-message* nil
          *actions* nil
          *state* nil)
    (tutor::setup-tutor1)
    (do-section ch sect 1)
    (let ((goal (new-name-fct 'goal)))
      (eval `(add-dm (,goal isa task state start step nil)))
      (eval `(goal-focus ,goal))
      (run .05)
      (update-state) )
    (setf *problem-size*  (length (assoc :equation (dump-screen))))
    (setf *episode-history* nil)
    (run time)
    *transitions*))




(defun instruction26 (n)
  (let ()     (set-condition 'long 'vd)          
  (reset) (setf *operator-history* nil) (setup1-7)
  (do ((count 1 (1+ count))
       (result  (progn (test-problems 2 6 2000) (analyze-combine *combine*))
                (progn (test-problems 2 6 2000) (sum-in (analyze-combine *combine*) result))))
      ((= count n)(window-close tutor::*window*) result)
    (window-close tutor::*window*)(reset) (setup1-7))))

(defun sum-in (lis1 lis2)
  (if (listp (car lis1)) (mapcar 'sum-in lis1 lis2)
      (mapcar '+ lis1 lis2)))


(defun analyze-combine (lis)
  (do ((temp (reverse lis) (cdr temp))
       (result (list 0) (if (= (car temp) 2) (cons (1+ (car result)) (cdr result)) (cons 0 result))))
      ((null temp) (reverse (cdr result)))))


(defun set-condition (exp cond)
  (case exp (long 
             (setf tutor::*problem-generation-function* 'TUTOR::GEN-PROBLEMS)
               (case cond 
                    (wa (angela-experiment nil nil))
                    (dd (angela-experiment t nil))
                    (vd (angela-experiment nil t))
                    (bo (angela-experiment t t))))
            (short 
             (setf tutor::*problem-generation-function* 'tutor::gen-short-problems)
               (case cond 
                 (wa-g (angela-experiment nil nil)
                  (setf tutor::*no-solve-instructions* nil))
                 (dd-g (angela-experiment t nil)
                        (setf tutor::*no-solve-instructions* nil))
                 (wa-ng (angela-experiment nil nil)
                       (setf tutor::*no-solve-instructions* t))
                 (dd-ng (angela-experiment t nil)
                       (setf tutor::*no-solve-instructions* t))))))

(defun test-section (ch sect) (let ((result (test-problems ch sect 2000))) (window-close tutor::*window*) result))

(defun run-condition (exp cond)
  (set-condition exp cond) (reset)
  (let ((result (problem-times (append (test-section 1 1) (test-section 1 2) (test-section 1 7) (test-section 2 6)))))
    (print (if (eq exp 'long) (filter-times result) result))))

(defun problem-times (lis)
  (do ((temp  (cdr (member 'done lis))(next-problem temp))
       (time 0 (third temp))
       (result nil (cons (- (first temp) time) result)))
      ((null temp) (reverse result))))

(defun next-problem (lis) 
  (let ((cand (cdr (member 'done lis))))
    (cond ( (eq (second cand) 'WRONG-ANSWER) (next-problem cand))
          (t cand))))

(defun filter-times (lis)
  (do ((indices (mapcar '1- '(1 2 14 15 16 23 24 25 32 36 39 46 49 50 57 62 67)) (cdr indices))
       (result nil (cons (nth (car indices) lis) result)))
      ((null indices) (reverse result))))

(defun average-run (exp cond n)
  (do ((i 1 (1+ i))
       (result (run-condition exp cond) (mapcar '+ (run-condition exp cond) result)))
      ((= i n) (mapcar #'(lambda (x) (/ x n)) result))))

(defun instruction (n)
  (let ((short  (progn (setf *similarity-list* *similarity-list-short*) (average-run 'short 'dd-g n))) 
        (long  (progn (setf *similarity-list* *similarity-list-long*) (average-run 'long 'dd n))))
  (Print "short:")
  (print short)
  (Print "long:")
  (print long) nil))

(defun discovery (n)
  (let ((short  (progn (setf *similarity-list* *similarity-list-short*) (average-run 'short 'wa-g n))) 
        (long  (progn (setf *similarity-list* *similarity-list-long*) (average-run 'long 'wa n))))
  (Print "short:")
  (print short)
  (Print "long:")
  (print long) nil))

(defun find-mistake ()
  (if (member 'COMBINE-OP-EDIT *transitions*) 'COMBINE-OP-bad-TYPE 'decompose))



(defun make-operators () 
   (setf *operators* (do ((temp *mapping* (cdr temp))
                               (result *basic-actions* (append result (mapcar 'car (cdar temp)))))
                              ((null temp) (remove-duplicates result)))))

(defun make-declarative-memory ()
  (let ((lis (append (do ((temp1 (cons 'finished *states*) (cdr temp1))
                          (result1 nil (cons (list (car temp1) 'isa 'chunk) result1)))
                         ((null temp1) result1))
                     (do ((temp2 *basic-actions* (cdr temp2))
                          (result2 nil (cons (list (car temp2) 'isa 'operator 'pre (car temp2) 'action (car temp2)) result2)))
                         ((null temp2) result2)))))
    (eval (cons 'add-dm lis))))


(defun start-up ()

(make-declarative-memory)

(eval (cons 'set-similarities *similarity-list*))
)

(defun recognize-the-pattern (target)
  (if (not (assoc :equation (dump-screen)) ) (list nil target)
  (do ((temp *states* (cdr temp)))
      ((null temp)  (list 'finished target) )
    (let ((result (match-a-pattern (eval (car temp)) target)))
      (if result (return (cons (car temp) result)))))))

(defun retrieve-patterns (lis)
  (if (eq (car lis) 'mode) (retrieve-mode (cdr lis))
  (let*  ((index (car lis))
         (set (cdr lis))
         (head (car (second (assoc index (dump-screen))))))
    (do ((temp set (cdr temp)))
        ((null temp) nil)
      (let ((result (match-pattern (car temp) head index)))
        (if result (return result)))))))

(defun match-pattern (pat node index)
  (let* ( (nodes (cdr (assoc index (dump-screen))))
          (struct (second (assoc node nodes)))
          (result (match-node pat node index)))
      (if result result
       (if (listp struct)
         (if (and (assoc (car struct) nodes) (setf result (match-pattern pat (car struct) index))) result
             (if (and (assoc (third struct) nodes) (setf result (match-pattern pat (third struct) index))) result))
         (if (and (assoc struct nodes) (setf result (match-pattern pat struct index))) result)))))

(defun retrieve-mode (lis)
  (let ((set (cdr (assoc :equation (dump-screen)))))
    (do ((temp lis (cdr temp)))
        ((null temp) nil)
    (let ((result (car (member (car temp) set :key 'caaddr))))
      (if result (return (car result)))))))

(defun extract-pattern (pat)
  (let* ((box (retrieve-patterns (eval pat)))
         (struct  (assoc box (cdr (assoc :equation (dump-screen)))))
         (mode (car (third struct))))
    (list (case mode 
            (:ARG1-EDITABLE (first (first (second struct))) )
            (:ARG2-EDITABLE (first (third (second struct))))
            (:op-EDITABLE (first (second (second struct))))
            (:answer-EDITABLE (second struct))
            (t box)))))

(defun match-a-pattern (pattern string)
  (case (car pattern)
    (or (do ((temp (cdr pattern) (cdr temp)))
            ((null temp) nil)
          (let ((result (match-basic (car temp) string)))
            (if result (return (list (car temp) result))))))))

(defun match-node (pattern node index)
  (let* ( (result node)
          (nodes (cdr (assoc index (dump-screen))))
          (struct (second (assoc node nodes))))
    (cond ((null struct) nil)
          ((and (eq pattern 'n) (numberp struct)) node)
          ((atom struct) (and (equal pattern '(n)) (numberp struct) result))
          (t (do ((temp (match-leaves pattern struct index)  (cdr temp)))
                 ((null temp) result)
               (cond ((not (car temp)) (return nil)) 
                     ((not (eq (car temp) t)) (setf result (car temp)))))))))

(defun match-leaves (pattern struct index)
  (if (and (listp pattern) (list struct) (= (length pattern) (length struct)))
      (mapcar #'(lambda (x y) (match-leaf x y index)) pattern struct)
      (list nil)))



(defun match-leaf (pat elem index)
  (cond ((listp pat)
         (cond ((eq (car pat) 'or) (if (member elem (cdr pat)) t))
               ((eq (car pat) 'embed) (if (second pat) 
                                        (if (match-node (second pat) elem index) elem) elem))
               (t (if (match-node pat elem index) t))))
        ((eq pat 'n) (numberp elem))
        ((eq pat elem) t)
        ((eq pat 'any) t)))


(defun match-basic (pattern string)
  (or (match-direct pattern string)
      (do ((temp string (cdr temp)))
          ((null temp) nil)
      (let ((new (extract pattern temp)))
          (if new (return new))))))

(defun extract (pattern lis)
  (cond ((listp (car lis)) (match-basic pattern (car lis)))
        ((member (car lis) '({ [))
    (let ((squiggle 0) (square 0) result)
      (loop (case (car lis)
              ({ (incf squiggle))
              ([ (incf square))
              (} (decf squiggle))
              (] (decf square)))
            (setf result (cons (car lis) result))
            (if (and (zerop squiggle) (zerop square)) (return (match-direct pattern (reverse result))))
            (setf lis (cdr lis)))))))
      


(defun match-direct (pattern string)
  (and (= (length pattern) (length string))
       (do ((temp1 pattern (cdr temp1))
            (temp2 string (cdr temp2)))
           ((null temp1) string)
;         (print (car temp1)) (print (car temp2))
         (cond ((listp (car temp1))
                (cond ((eq (caar temp1) 'or) 
                       (if (not (member (car temp2) (car temp1))) (return nil)))
                      ((not (listp (car temp2))) (return nil))
                      ((not (match-direct (car temp1) (car temp2))) (return nil))))
               ((eq (car temp1) 'n) (if (not (numberp (car temp2))) (return nil)))
               ((eq (car temp1) 'any))
               ((not (eq (car temp1) (car temp2))) (return nil))))))

;instructional mappings

(defparameter *mapping* 
             '(
              ("Find a box with two numbers and an operator and click it. " 
               (operator1c isa operator pre ebox action just-numbers)
               (operator1c1 isa operator pre EBBOX action just-numbers)
               )
               ("Click the button labeled \"Evaluate\" "
                (operator3c isa operator pre ebox1 action evaluate)
                (operator3c1 isa operator pre ebbox1 action evaluate))
               ("Click the little green box. "
                (operator5c isa operator pre green-little-box action get-green-box)
                (operator22c isa operator pre COMBINE-OP action GET-GREEN-OP)
                (operator22c isa operator pre COMBINE-BOTH action GET-GREEN-OP))
               ("Find the answer by evaluating the box above and enter it "
                (operator7c isa operator pre arg-editable action type-value)) 
               ("Enter the number from the top box. "
                (operator7c isa operator pre arg-editable action type-value))
               ("Your Answer is correct.Click the Next Problem button."
                (operator9c isa operator pre finished action next)
                (operator9c1 isa operator pre bfinished action next)
                (operator9c1 isa operator pre cfinished action next))
               ("Find a box with just a number and click it. "
               (operator11c isa operator pre input-number action just-numbers))
               ("Your answer is incorrect.Use the < and > buttons (or the left and right arrow keys)and the First Mistake buttonto review your work and correct the mistake."
               (operator-go-back isa operator pre wrong-answer action first-mistake))
               ("Find a bottom box with just a number and a box feeding into that box and click them. "
;                 (operator13c isa operator pre bbox action find-above-box)
                (operator13c isa operator pre bbox action find-bottom-box)
                (operator13c1 isa operator pre bbox2 action find-above-box)
                (operator12c isa operator pre BBOX1 action find-bottom-box)
                )
               ("Find two boxes with addition or subtraction that are connected and click them. "
;                 (operator20c isa operator pre combine-a action FIND-TOP-COMBINE)
                (operator20c isa operator pre combine-a action FIND-BOTTOM-COMBINE)
                (operator20c1 isa operator pre combine-a2 action FIND-TOP-COMBINE)
                (operator20c isa operator pre combine-a1 action FIND-BOTTOM-COMBINE)
                )
               ("Click the button labeled \"Invert\" "
                (operator14c isa operator pre bbox12 action invert))
               ("Click the button labeled \"Combine\" "
                (operator21c isa operator pre COMBINE-A12 action combine))
               ("Click the green box. "
                (operator5c1 isa operator pre GREEN-BIG-BOX action get-green-box))
               ("Enter the number from the bottom box, then a new operator, which is the inverse of the operator from the box above, and then the number from the box above. "
                (operator17c isa operator pre box-editable action express)
                )
               ("Enter the sign from the little box in the box above. "
                (operator23c isa operator pre COMBINE-OP-EDIT action COMBINE-OP-bad-TYPE)
                (operator23c1 isa operator pre COMBINE-OP-EDIT action COMBINE-OP-TYPE)
                )
               ("Enter the operator from the box below. "
                (operator23c isa operator pre COMBINE-OP-EDIT action COMBINE-OP-bad-TYPE)
                (operator23c1 isa operator pre COMBINE-OP-EDIT action COMBINE-OP-TYPE)
                )
               ("Click the big green box. "
                (operator24c isa operator pre COMBINE-BOX action get-green-box)
                )
               ("Enter the signed number from the box above, then the operator from the box below, and then the number from the box below. "
                (operator25c isa operator pre COMBINE-BOX-EDIT action combine-box-type)
                )
               ))


;special state information
(defun good-state (pre post) 
  (let ((sequence '(ebox input-number ebbox bbox combine-a ebox1 ebbox1 bbox1 bbox2 combine-a1 combine-a2 bbox12 combine-a12 combine-both 
                    green-little-box green-big-box
                    type-in-box finished combine-op-edit combine-box combine-box-edit BOX-EDITABLE ARG-EDITABLE bfinished)))
    (member post (cdr (member pre sequence)))))

(defun terminus (pre) (member pre '(ebox input-number ebbox bbox finished bfinished cfinished)))

(defun finished (state) (member state '(finished bfinished cfinished)))



;prior history
(defun setup1-1 ()
  (setup '(358100.62 (JUST-NUMBERS -1.0E+8 2007.06 291.6651599999999) (POINTED-NUMBERS -1.0E+8 2008.46 295.50764) (EVALUATE -1.0E+8 2006.32 298.5847) 
           (TYPE-VALUE -1.0E+8 2007.16 301.78613999999993) (GET-GREEN-BOX -1.0E+8 2002.42 -6789.359680000001) (BURP -1.0E+8 2000.18 -42970.36046) 
           (NEXT -1.0E+8 2000.48 -32924.51243999999) (OPERATOR1C 0.7999999999999997 48.06 351.58920000000006)  (OPERATOR3C 7.396880000000003 47.28 351.9526400000001) 
           (OPERATOR5C 14.059539999999986 47.8 353.2573200000001) (OPERATOR7C 20.45156000000002 48.06 354.6642000000001) (OPERATOR9C 41.70382000000001 17.76 357.4006199999999) )))

(defun setup1-2 ()
  (setup '(197201.9 (JUST-NUMBERS -1.0000035810062002E+8 2009.6599999999999 149.2076) (POINTED-NUMBERS -1.0000035810062002E+8 2012.6599999999994 135.33190000000002) 
           (EVALUATE -1.0000035810062002E+8 2009.02 -4877.338) (TYPE-VALUE -1.0000035810062002E+8 2009.4600000000003 -4882.445099999999) 
           (GET-GREEN-BOX -1.0000035810062002E+8 2003.8199999999997 -9877.1394) (BURP -1.0000035810062002E+8 2000.1799999999998 -50000.0) 
           (NEXT -1.0000035810062002E+8 2000.78 -34966.8385) (FIND-BOTTOM-BOX -1.0E+8 2000.0 -50000.0) (FIND-ABOVE-BOX -1.0E+8 2000.0 -50000.0) 
           (INVERT -1.0E+8 2000.0 -50000.0) (DECOMPOSE -1.0E+8 2000.0 -50000.0) (EXPRESS -1.0E+8 2000.0 -50000.0) (SOLVE -1.0E+8 2000.0 -50000.0) 
           (OPERATOR1C -357.30062 67.25999999999999 188.5012) (OPERATOR3C -350.70374 74.37999999999998 190.9776) (OPERATOR5C -344.04108 75.19999999999999 190.7493) 
           (OPERATOR7C -337.6490599999999 76.55999999999999 192.51479999999998) (OPERATOR9C -316.3968 27.959999999999997 196.50189999999998) 
           (OPERATOR11C 0.7999999999999999 8.7 155.6353) )))

(defun setup1-7 ()
  (setup '(604670.6 (JUST-NUMBERS -1.0000055530252002E+8 2019.4599999999998) (POINTED-NUMBERS -1.0000055530252002E+8 2014.6599999999994) 
           (EVALUATE -1.0000055530252002E+8 2016.02) (GET-GREEN-BOX -1.0000055530252002E+8 2006.2199999999998) (BURP -1.0000055530252002E+8 2002.78) 
           (NEXT -1.0000055530252002E+8 2001.3799999999999) (FIND-BOTTOM-BOX -1.000001972019E+8 2005.2) (FIND-ABOVE-BOX -1.000001972019E+8 2001.2) 
           (INVERT -1.000001972019E+8 2001.8) (DECOMPOSE -1.000001972019E+8 2000.8) (EXPRESS -1.000001972019E+8 2000.6) (SOLVE -1.000001972019E+8 2001.8) 
           (TYPE-VALUE -1.0000055530252002E+8 2012.6600000000003) (OPERATOR1C -554.50252 67.25999999999999) (OPERATOR1C1 37.85940000000001 31.0) 
           (OPERATOR3C -547.90564 74.37999999999998) (OPERATOR3C1 44.1594 30.4) (OPERATOR5C -541.24298 106.6) (OPERATOR7C -534.8509599999999 108.35999999999999) 
           (OPERATOR9C -513.5987 30.759999999999998) (OPERATOR9C1 55.809400000000004 25.2) (OPERATOR11C -196.40189999999998 8.7) (NIL) (OPERATOR13C 0.8 38.4) 
           (OPERATOR13C1 29.7246 36.6) (NIL) (OPERATOR14C 15.740000000000004 36.0) (OPERATOR5C1 21.79 39.0) (OPERATOR17C 28.089999999999996 36.2))))







;state patterns
(setf combine-a '(or (n (or + -) (any (or + -) n))
                     ((any (or + -) n) (or + -) n)
                     (n (or + -) (n (or + -) any))
                     ((n (or + -) any) (or + -) n)
                     (n (or * /) (any (or * /) n))
                     ((any (or * /) n) (or * /) n)
                     (n (or * /) (n (or * /) any))
                     ((n (or * /) any) (or * /) n)))
(setf combine-a1 '(or (n (or + -) { any (or + -) n })
                     ({ any (or + -) n } (or + -) n)
                     (n (or + -) { n (or + -) any })
                     ({ n (or + -) any } (or + -) n)
                     (n (or * /) { any (or * /) n })
                     ({ any (or * /) n } (or * /) n)
                     (n (or * /) { n (or * /) any })
                     ({ n (or * /) any } (or * /) n)))
(setf combine-a2 '(or ({ n (or + -) (any (or + -) n) })
                     ({(any (or + -) n) (or + -) n })
                     ({ n (or + -) (n (or + -) any) })
                     ({(n (or + -) any) (or + -) n })
                     ({ n (or * /) (any (or * /) n) })
                     ({(any (or * /) n) (or * /) n })
                     ({ n (or * /) (n (or * /) any) })
                     ({(n (or * /) any) (or * /) n })))
(setf combine-a12 '(or ({ n (or + -) { any (or + -) n } })
                     ({ { any (or + -) n } (or + -) n })
                     ({ n (or + -) { n (or + -) any } })
                     ({ { n (or + -) any } (or + -) n })
                     ({ n (or * /) { any (or * /) n } })
                     ({ { any (or * /) n } (or * /) n })
                     ({ n (or * /) { n (or * /) any } })
                     ({ { n (or * /) any } (or * /) n })))
(setf combine-b '(or (n (or + -) ({ n * x } (or + -) n))
                     (({ n * x } (or + -) n) (or + -) n)
                     (n (or + -) (n (or + -) { n * x }))
                     ((n (or + -) { n * x }) (or + -) n)
                     (n (or * /) ({ n * x } (or * /) n))
                     (({ n * x } (or * /) n) (or * /) n)
                     (n (or * /) (n (or * /) { n * x }))
                     ((n (or * /) { n * x }) (or * /) n)))
(setf combine-b1 '(or (n (or + -) { { n * x } (or + -) n })
                     ({ { n * x } (or + -) n } (or + -) n)
                     (n (or + -) { n (or + -) { n * x } })
                     ({ n (or + -) { n * x } } (or + -) n)
                     (n (or * /) { { n * x } (or * /) n })
                     ({ { n * x } (or * /) n } (or * /) n)
                     (n (or * /) { n (or * /) { n * x } })
                     ({ n (or * /) { n * x } } (or * /) n)))
(setf combine-b2 '(or ({ n (or + -) ({ n * x } (or + -) n) })
                     ({({ n * x } (or + -) n) (or + -) n })
                     ({ n (or + -) (n (or + -) { n * x }) })
                     ({(n (or + -) { n * x }) (or + -) n })
                     ({ n (or * /) ({ n * x } (or * /) n) })
                     ({({ n * x } (or * /) n) (or * /) n })
                     ({ n (or * /) (n (or * /) { n * x }) })
                     ({(n (or * /) { n * x }) (or * /) n })))
(setf combine-b12 '(or ({ n (or + -) { { n * x } (or + -) n } })
                     ({ { { n * x } (or + -) n } (or + -) n })
                     ({ n (or + -) { n (or + -) { n * x } } })
                     ({ { n (or + -) { n * x } } (or + -) n })
                     ({ n (or * /) { { n * x } (or * /) n } })
                     ({ { { n * x } (or * /) n } (or * /) n })
                     ({ n (or * /) { n (or * /) { n * x } } })
                     ({ { n (or * /) { n * x } } (or * /) n })))
(setf combine-box '(or (any [ (or + - * /) ] ( [  ]   [  ]   [  ] ) )
                     (( [  ]   [  ]   [  ] )  [ (or + - * /) ] any)
                     (any [ (or + - * /) ] ( [  ]   [  ]   [  ] ) )
                     (( [  ]   [  ]   [  ] )  [ (or + - * /) ] any)))
(setf combine-both '(or (any [  ] ( [  ]   [  ]   [  ] ) )
                     (( [  ]   [  ]   [  ] )  [  ] any)
                     (any [  ] ( [  ]   [  ]   [  ] ) )
                     (( [  ]   [  ]   [  ] )  [  ] any)))
(setf combine-op '(or (any [] ( [ N ]   [ (or + - * /) ]   [ N ] ) )
                     (( [ N ]   [ (or + - * /) ]   [ N ] )  [  ] any)
                     (any [  ] ( [ N ]   [ (or + - * /) ]   [ N ] ) )
                     (( [ N ]   [ (or + - * /) ]   [ N ] )  [  ] any)))
(setf combine-box-edit '(or (any [  ] ( [ F ]   [ F ]   [ F ] ) )
                     (( [ F ]   [ F ]   [ F ] )  [  ] any)
                     (any [  ] ( [ F ]   [ F ]   [ F ] ) )
                     (( [ F ]   [ F ]   [ F ] )  [  ] any)
                     (any [ (or + - * /) ] ( [ F ]   [ F ]   [ F ] ) )
                     (( [ F ]   [ F ]   [ F ] )  [ (or + - * /) ] any)
                     (any [ (or + - * /) ] ( [ F ]   [ F ]   [ F ] ) )
                     (( [ F ]   [ F ]   [ F ] )  [ (or + - * /) ] any)
                     (any [  ] ( [ F N ]   [ F (or + - * /) ]   [ F N ] ) )
                     (( [ F N ]   [ F (or + - * /) ]   [ F N ] )  [  ] any)
                     (any [  ] ( [ F N ]   [ F (or + - * /) ]   [ F N ] ) )
                     (( [ F N ]   [ F (or + - * /) ]   [ F N ] )  [  ] any)))
(setf combine-op-edit '(or (any [ F ] ( [  ]   [  ]   [  ] ) )
                     (( [  ]   [  ]   [  ] )  [ F ] any)
                     (any [ F ] ( [  ]   [  ]   [  ] ) )
                     (( [  ]   [  ]   [  ] )  [ F ] any)
                     (any [ F ] ( [ N ]   [ (or + - * /) ]   [ N ] ) )
                     (( [ N ]   [ (or + - * /) ]   [ N ] )  [ F ] any)
                     (any [ F ] ( [ N ]   [ (or + - * /) ]   [ N ] ) )
                     (( [ N ]   [ (or + - * /) ]   [ N ] )  [ F ] any)
                     (any [ F (or + - * /)] ( [  ]   [  ]   [  ] ) )
                     (( [  ]   [  ]   [  ] )  [ F (or + - * /)] any)
                     (any [ F (or + - * /)] ( [  ]   [  ]   [  ] ) )
                     (( [  ]   [  ]   [  ] )  [ F (or + - * /)] any)))
(setf finished '(or (n)))

(setf ebox '(or (n any n)))
(setf input-number '(or ((n) any n)( n any (n ))))
(setf eboxa '(or (([ n ] [ any ] [ n ]) [ any ] any)
                (any [ any ]([ n ] [ any ] [ n ]))))
(setf ebox1 '(or ({ n any n } any any)
                (any any { n any n })
                ({ n any n })
                   ({ n } any n)( n any { n })))
(setf ebox2 '(or ({ (n any n) any any })
                ({ any any (n any n) })
                ({ (n) any  n })({ n any (n)})))
(setf ebox12 '(or ({ { n any n } any any })
                ({ any any { n any n })
                ({ { n } any n })({ n any { n } })))
(setf green-little-box '(or ([ ] (or + - * /) any)
                            (any (or + - * /) [ ])
                            ([ ]) 
                            ([ N ] (or + - * /) any)
                            (any (or + - * /) [ N ])
                            ([ N ])))
(setf arg-editable '(or ([ n ] (or + - * /) any)
                        (any (or + - * /) [ F ])
                        ([ F ]) ([ F N ])
                        ([ F N ] (or + - * /) any)
                        (any (or + - * /) [ F N ])))
(setf bbox '(or ((any (or + - / *) any) =  n)))
(setf bbox1 '(or ({ any (or + - / *) any } = n)))
(setf bbox12 '(or ({ { any (or + - / *) any } =  n })))
(setf bbox2 '(or ({ (any (or + - / *) any) =  n })))
(setf ebbox '(or (any = (n any n))))
(setf ebbox1 '(or (any = { n any n })))
(setf ebbox12 '(or ({ any = { n any n } })))
(setf ebbox2 '(or ({ any = (n any n) } )))
(setf green-big-box '(or ([ ] [ ] [ ])
                      ([ N ] [ any ] [ N ])))
(setf box-editable '(or ([ F ] [ F ] [ F ])
                     ([ F N ] [ F any ] [ F N ])))
(setf bfinished '(or (x =  n)))
(setf cfinished '(or (any (or + - * /)  any)({ any (or + - * /)  any })))
(defparameter *states* '( combine-box-edit combine-op-edit box-editable arg-editable  combine-both combine-box  combine-op   green-big-box green-little-box
                      ebbox12  ebbox1  ebbox2  ebbox ebox12 ebox1 ebox2 ebox  combine-a combine-a1 combine-a2 combine-a12  
                      combine-b combine-b1 combine-b2 combine-b12 
                   eboxa   bbox12 bbox1 bbox2 bbox input-number bfinished cfinished))


;actions
(setf just-numbers '(:equation (n any n) n (x any)))
(setf get-green-box '(mode :answer-editable :ARG1-EDITABLE :ARG2-EDITABLE :BOX-EDITABLE))
(setf pointed-numbers '(:equation ((n any n) any any)
                      (any any (n any n))
                       (n any (n))
                           ((n) any n)))
(setf find-above-box '(:equation ((embed (any any any)) =  n)))
(setf find-bottom-box '(:equation ((any any any) =  n)))
(setf find-top-combine '(:EQUATION (N (OR + -) (embed (ANY (OR + -) N))) ((embed (ANY (OR + -) N)) (OR + -) N)
                                  (N (OR + -) (embed (N (OR + -) ANY))) ((embed (N (OR + -) ANY)) (OR + -) N) 
                                  (N (OR * /) (embed (ANY (OR * /) N))) ((embed (ANY (OR * /) N)) (OR * /) N) 
                                  (N (OR * /) (embed (N (OR * /) ANY))) ((embed (N (OR * /) ANY)) (OR * /) N)))
(setf find-bottom-combine (cons :equation (cdr combine-a)))
(setf get-green-op '(mode :op-editable))
(setf operate-on1 '(:equation (any (or + - * /)  any)))
(setf operate-on2 '(:equation ((embed (any (or + - * /)  any)) (or + - * /)  any)
                              (any (or + - * /)  (embed (any (or + - * /)  any)))))

(setf decompose '(:equation (any = any)))
(setf find-spurious-combine '(:equation ((embed (n * any)) (or + -) any)))
(setf express '(:equation (any = (embed))))
(setf solve '(:equation ((embed) = any)))
(setf type-value '(:old-equation (n any n) n (any = n)))
(setf combine-op-type (cons :old-equation (cdr combine-a)))
(setf combine-op-bad-type combine-op-type)
(setf combine-box-type (cons :old-equation (cdr combine-a)))
(setf get-number '(:old-equation (any (or + - * /)  any)))

(setf evaluate :evaluate next :next-problem burp nil invert :invert combine :combine)
(defparameter *basic-actions* '(just-numbers pointed-numbers evaluate get-green-box burp next find-bottom-box find-above-box find-spurious-combine invert
                                decompose express solve type-value find-top-combine combine find-bottom-combine combine-op-type combine-op-bad-type combine-box-type get-green-op
                                operate-on1 operate-on2 get-number))

;action-types
(defun box-action (act) (member act '(just-numbers get-green-box pointed-numbers find-above-box find-bottom-box find-top-combine find-bottom-combine  find-spurious-combine
                                      get-green-op operate-on1 operate-on2)))
(defun button-action (act) (member act '(evaluate invert combine)))
(defun number-action (act) (member act '(decompose express solve type-value combine-op-type combine-op-bad-type combine-box-type get-number)))
(defun map-click (item)
  (case item (read :READ-INSTRUCTIONS)
        (evaluate  :evaluate)
             (invert :invert)
             (back :back)
             (combine :combine)
             (first-mistake :last-correct-state)
             (next :next-problem)))

(defparameter *similarity-list-long* 
                '((ebox just-numbers -.2)
                  (ebox pointed-numbers -.2)
                  (ebox evaluate -.2)
                  (ebox type-value -.2)

                  (ebox1 just-numbers -.2)
                  (ebox1 pointed-numbers -.2)
                  (ebox1 evaluate -.2)
                  (ebox1 type-value -.2)

                  (ebox2 just-numbers -.2)
                  (ebox2 pointed-numbers -.2)
                  (ebox2 evaluate -.2)
                  (ebox2 type-value -.2)

                  (ebox12 just-numbers -.2)
                  (ebox12 pointed-numbers -.2)
                  (ebox12 evaluate -.2)
                  (ebox12 type-value -.2)

                  (green-little-box get-green-box -.2)
                  (green-little-box evaluate -.2)
                  (green-little-box type-value -.2)

                  (arg-editable type-value -.2)
                  (arg-editable burp -.2)

                  (finished next -.2)
                  (finished evaluate -.2)
                  (finished type-value -.2)

                  (input-number just-numbers -.2)
                  (input-number pointed-numbers -.35)
                  (input-number evaluate -.35)
                  (input-number type-value -.35)

                  (bbox FIND-ABOVE-BOX -.15)
                  (bbox FIND-BOTTOM-BOX -.15)
                  (bbox invert -.2)
                  (bbox type-value -.2)

                  (bbox1 FIND-ABOVE-BOX -.2)
                  ;(bbox1 FIND-BOTTOM-BOX -.2)
                  (bbox1 FIND-BOTTOM-BOX -.15)
                  (bbox1 invert -.2)
                  (bbox1 type-value -.2)

                  (bbox2 FIND-ABOVE-BOX -.2)
                  ;(bbox2 FIND-ABOVE-BOX -.15)
                  (bbox2 FIND-BOTTOM-BOX -.2)
                  (bbox2 invert -.2)
                  (bbox2 type-value -.2)

                  (bbox12 FIND-ABOVE-BOX -.2)
                  (bbox12 FIND-BOTTOM-BOX -.2)
                  (bbox12 invert -.15)
                  (bbox12 type-value -.2)

                  (green-big-box get-green-box -.15)
                  (green-big-box invert -.2)
                  (green-big-box type-value -.2)

                  (box-editable solve -.2)
                  (box-editable burp -.2)
                  (box-editable decompose -.2)
                  (box-editable express -.2)
                  
                  (ebbox just-numbers -.2)
                  (ebbox pointed-numbers -.2)
                  (ebbox evaluate -.2)
                  (ebbox type-value -.2)

                  (ebbox1 just-numbers -.2)
                  (ebbox1 pointed-numbers -.2)
                  (ebbox1 evaluate -.15)
                  (ebbox1 type-value -.2)

                  (ebbox2 just-numbers -.2)
                  (ebbox2 pointed-numbers -.2)
                  (ebbox2 evaluate -.2)
                  (ebbox2 type-value -.2)

                  (ebbox12 just-numbers -.2)
                  (ebbox12 pointed-numbers -.2)
                  (ebbox12 evaluate -.2)
                  (ebbox12 type-value -.2)

                  (bfinished next -.15)
                  (bfinished evaluate -.2)
                  (bfinished type-value -.2)

                  (combine-a find-bottom-combine -.2)
                  (combine-a find-top-combine -.2)
                  (combine-a find-spurious-combine -.35)
                  (combine-a combine -.2)
                  (combine-a burp -.2)

                  (combine-a1 find-bottom-combine -.2)
                  (combine-a1 find-top-combine -.2)
                  ;(combine-a1 find-spurious-combine -.25)
                  (combine-a find-spurious-combine -.35)
                  (combine-a1 combine -.2)
                  (combine-a1 burp -.2)

                  (combine-a2 find-bottom-combine -.2)
                  (combine-a2 find-top-combine -.2)
                  (combine-a find-spurious-combine -.35)
                  (combine-a2 combine -.2)
                  (combine-a2 burp -.2)

                  (combine-a12 find-bottom-combine -.2)
                  (combine-a12 find-top-combine -.2)
                  (combine-a find-spurious-combine -.35)
                  (combine-a12 combine -.2)
                  (combine-a12 burp -.2)

                  (combine-b find-bottom-combine -.2)
                  (combine-b find-top-combine -.2)
                  (combine-b find-spurious-combine -.15)
                  (combine-b combine -.2)
                  (combine-b burp -.2)

                  (combine-b1 find-bottom-combine -.2)
                  (combine-b1 find-top-combine -.2)
                  (combine-b1 find-spurious-combine -.15)
                  (combine-b1 combine -.2)
                  (combine-b1 burp -.2)

                  (combine-b2 find-bottom-combine -.2)
                  (combine-b2 find-top-combine -.2)
                  (combine-b2 find-spurious-combine -.15)
                  (combine-b2 combine -.2)
                  (combine-b2 burp -.2)

                  (combine-b12 find-bottom-combine -.2)
                  (combine-b12 find-top-combine -.2)
                  (combine-b12 find-spurious-combine -.15)
                  (combine-b12 combine -.2)
                  (combine-b12 burp -.2)

                  (combine-both GET-GREEN-OP -.2)
                  (combine-both GET-GREEN-OP -.15)
                  (combine-both GET-GREEN-box -.5)
                  (combine-both combine -.2)
                  (combine-both type-value -.2)

                  (COMBINE-OP-EDIT COMBINE-OP-TYPE -.15)
                  (COMBINE-OP-EDIT GET-GREEN-box -.5)
                  (COMBINE-OP-EDIT type-value -.2)
                  (COMBINE-OP-EDIT combine -.2)

                  (COMBINE-box-EDIT COMBINE-box-TYPE -.15)
                  (COMBINE-box-EDIT GET-GREEN-op -.5)
                  (COMBINE-box-EDIT type-value -.2)
                  (COMBINE-box-EDIT combine -.2)

                  (combine-op GET-GREEN-OP -.15)
                  (combine-op GET-GREEN-box -.5)
                  (combine-op combine -.2)
                  (combine-op type-value -.2)

                  (combine-box GET-GREEN-OP -.5)
                  (combine-box GET-GREEN-box -.15)
                  (combine-box combine -.2)
                  (combine-box type-value -.2)

                  (cfinished next -.15)
                  (cfinished evaluate -.2)
                  (cfinished type-value -.2)))


(defparameter *similarity-list-short* 
                '((ebox just-numbers -.2)
                  (ebox pointed-numbers -.2)
                  (ebox evaluate -.2)
                  (ebox type-value -.2)

                  (ebox1 just-numbers -.2)
                  (ebox1 pointed-numbers -.2)
                  (ebox1 evaluate -.2)
                  (ebox1 type-value -.2)

                  (ebox2 just-numbers -.2)
                  (ebox2 pointed-numbers -.2)
                  (ebox2 evaluate -.2)
                  (ebox2 type-value -.2)

                  (ebox12 just-numbers -.2)
                  (ebox12 pointed-numbers -.2)
                  (ebox12 evaluate -.2)
                  (ebox12 type-value -.2)

                  (green-little-box get-green-box -.2)
                  (green-little-box evaluate -.2)
                  (green-little-box type-value -.2)

                  (arg-editable type-value -.2)
                  (arg-editable burp -.2)

                  (finished next -.2)
                  (finished evaluate -.2)
                  (finished type-value -.2)

                  (input-number just-numbers -.2)
                  (input-number pointed-numbers -.35)
                  (input-number evaluate -.35)
                  (input-number type-value -.35)

                  (bbox FIND-ABOVE-BOX -.2)
                  (bbox FIND-BOTTOM-BOX -.2)
                  (bbox invert -.2)
                  (bbox type-value -.2)

                  (bbox1 FIND-ABOVE-BOX -.2)
                  (bbox1 FIND-BOTTOM-BOX -.2)
                  (bbox1 invert -.2)
                  (bbox1 type-value -.2)

                  (bbox2 FIND-ABOVE-BOX -.2)
                  ;(bbox2 FIND-ABOVE-BOX -.15)
                  (bbox2 FIND-BOTTOM-BOX -.2)
                  (bbox2 invert -.2)
                  (bbox2 type-value -.2)

                  (bbox12 FIND-ABOVE-BOX -.2)
                  (bbox12 FIND-BOTTOM-BOX -.2)
                  (bbox12 invert -.2)
                  (bbox12 type-value -.2)

                  (green-big-box get-green-box -.2)
                  (green-big-box invert -.2)
                  (green-big-box type-value -.2)

                  (box-editable solve -.2)
                  (box-editable burp -.2)
                  (box-editable decompose -.2)
                  (box-editable express -.2)
                  
                  (ebbox just-numbers -.2)
                  (ebbox pointed-numbers -.2)
                  (ebbox evaluate -.2)
                  (ebbox type-value -.2)

                  (ebbox1 just-numbers -.2)
                  (ebbox1 pointed-numbers -.2)
                  (ebbox1 evaluate -.2)
                  (ebbox1 type-value -.2)

                  (ebbox2 just-numbers -.2)
                  (ebbox2 pointed-numbers -.2)
                  (ebbox2 evaluate -.2)
                  (ebbox2 type-value -.2)

                  (ebbox12 just-numbers -.2)
                  (ebbox12 pointed-numbers -.2)
                  (ebbox12 evaluate -.2)
                  (ebbox12 type-value -.2)

                  (bfinished next -.2)
                  ;(bfinished next -.15)
                  (bfinished evaluate -.2)
                  (bfinished type-value -.2)

                  (combine-a find-bottom-combine -.2)
                  (combine-a find-top-combine -.2)
                  (combine-a find-spurious-combine -.25)
                  (combine-a combine -.2)
                  (combine-a burp -.2)

                  (combine-a1 find-bottom-combine -.2)
                  (combine-a1 find-top-combine -.2)
                  (combine-a1 find-spurious-combine -.25)
                  (combine-a1 combine -.2)
                  (combine-a1 burp -.2)

                  (combine-a2 find-bottom-combine -.2)
                  (combine-a2 find-top-combine -.2)
                  (combine-a2 find-spurious-combine -.25)
                  (combine-a2 combine -.2)
                  (combine-a2 burp -.2)

                  (combine-a12 find-bottom-combine -.2)
                  (combine-a12 find-top-combine -.2)
                  (combine-a12 find-spurious-combine -.25)
                  (combine-a12 combine -.2)
                  (combine-a12 burp -.2)

                  (combine-b find-bottom-combine -.2)
                  (combine-b find-top-combine -.2)
                  (combine-b find-spurious-combine -.15)
                  (combine-b combine -.2)
                  (combine-b burp -.2)

                  (combine-b1 find-bottom-combine -.2)
                  (combine-b1 find-top-combine -.2)
                  (combine-b1 find-spurious-combine -.15)
                  (combine-b1 combine -.2)
                  (combine-b1 burp -.2)

                  (combine-b2 find-bottom-combine -.2)
                  (combine-b2 find-top-combine -.2)
                  (combine-b2 find-spurious-combine -.15)
                  (combine-b2 combine -.2)
                  (combine-b2 burp -.2)

                  (combine-b12 find-bottom-combine -.2)
                  (combine-b12 find-top-combine -.2)
                  (combine-b12 find-spurious-combine -.15)
                  (combine-b12 combine -.2)
                  (combine-b12 burp -.2)

                  (combine-both GET-GREEN-OP -.2)
                  (combine-both GET-GREEN-OP -.15)
                  (combine-both GET-GREEN-box -.5)
                  (combine-both combine -.2)
                  (combine-both type-value -.2)

                  (COMBINE-OP-EDIT COMBINE-OP-TYPE -.2)
                  (COMBINE-OP-EDIT GET-GREEN-box -.5)
                  (COMBINE-OP-EDIT type-value -.2)
                  (COMBINE-OP-EDIT combine -.2)

                  (COMBINE-box-EDIT COMBINE-box-TYPE -.2)
                  (COMBINE-box-EDIT GET-GREEN-op -.5)
                  (COMBINE-box-EDIT type-value -.2)
                  (COMBINE-box-EDIT combine -.2)

                  (combine-op GET-GREEN-OP -.2)
                  (combine-op GET-GREEN-box -.5)
                  (combine-op combine -.2)
                  (combine-op type-value -.2)

                  (combine-box GET-GREEN-OP -.5)
                  (combine-box GET-GREEN-box -.2)
                  (combine-box combine -.2)
                  (combine-box type-value -.2)

                  (cfinished next -.2)
                  (cfinished evaluate -.2)
                  (cfinished type-value -.2)))

     
(setf *similarity-list* *similarity-list-long*)    











        
        


    
    




