;;;; Display text in the window with next/back buttons

(in-package :tutor)

(defparameter *text-block-point* #@(10 10))

(defvar *text-block-list* nil
  "A list of text blocks to display in order")

(defvar *current-text-block* nil)

(defvar *next-button* nil)
(defvar *prev-button* nil)

(defvar *text-block-widgets* nil)

(defvar *text-view* nil)
;; (defvar *scroll-view* nil)

(defclass fsm-text (fsm-block)
  ())

(defclass scroll-bar-display (view) ())

;; (defmethod scroll-bar-changed ((scrollee scroll-bar-display)
;;                                scroll-bar)
;;   "Used to scroll the *text-view*."
;;   (set-view-scroll-position scrollee (make-point 0 (scroll-bar-setting scroll-bar)))
;;   (view-focus-and-draw-contents scrollee))

(defmethod fsm-init ((blk (eql 'text)) state)
  (setf *text-view* (make-instance 'scroll-bar-display
				   :view-position (make-point 0 0)
				   :view-size (make-point (truncate +window-width+ 2)
							  (- +window-height+ 50))
				   :view-container *window*))
;; 	*scroll-view* (make-instance 'scroll-bar-dialog-item
;; 				     :view-position (make-point (+ 10 (truncate +window-width+ 2)) 0)
;; 				     :view-container *window*
;; 				     :length (- +window-height+ 50)
;; 				     :scroll-size 20
;; 				     :page-size (- +window-height+ 70)
;; 				     :track-thumb-p t
;; 				     :scrollee *text-view*))
  (setf *text-block-widgets* nil)
  (apply 'display-text state))

;; (defun text-block-end ()
;;   (delete-subviews *window*))

;; (defgeneric text-draw-block (blk x y)
;;   (:documentation "Return the height and the list of widgets of the blk drawn."))

(defun split-by-one-space (string)
  "Returns a list of substrings of string
divided by ONE space each.
Note: Two consecutive spaces will be seen as
if there were an empty string between them."
  (loop for i = 0 then (1+ j)
        as j = (position #\Space string :start i)
        collect (subseq string i j)
        while j))

(defun wrap-text (text width font)
  "Wrap TEXT to fit the width of VIEW using FONT."
  (let ((paragraph (split-by-one-space (substitute #\Space #\Newline text)))
        lines
        line
        (x 0))
    (dolist (word paragraph)
      ;; XXX: hack alert.
      (setf word (concatenate 'string word " "))
      (when (>= (+ x (string-width word font)) width)
        (push (nreverse line) lines)
        (setf x 0
              line nil))
      (push word line)
      (incf x (string-width word font)))
    ;; don't forget the last line
    (push (nreverse line) lines)
    (format nil "~{~{~a~}~^~%~}" (nreverse lines))))

(defun draw-text-block (blk x y text-view)
  ""
  (let* ((p (make-point x y))
	 (view (draw-text-with-font text-view (point-h p) (point-v p) (wrap-text blk (view-width text-view) *text-font*) *text-font*)))
    (values (+ (point-v (view-size view)) +view-pad+)
	    (list view))))

;; (defun draw-eqn (eqn x y text-view)
;;   ;; Create and setup the tree.
;;   (let* ((tree (make-and-space-tree eqn))
;; 	 (view (make-instance 'view
;; 			      :view-position (make-point x y)
;; 			      :view-size (eqn-size tree)
;; 			      :view-container text-view)))
;;     ;; draw the tree
;;     (recenter-tree tree view nil)
;;     (redraw-tree tree view)
;; ;;     (draw-rect *text-view* (point-h (view-position view)) (point-v (view-position view))
;; ;; 	       (point-h (add-points (view-position view) (view-size view))) (point-v (add-points (view-position view) (view-size view))))
;;     (values (point-v (view-size view))
;; 	    view)))

;; (defun draw-eqn-block (eqns x y text-view)
;;   "EQNS is a list of equations that will be drawn from left to right."
;;   (let ((cur-x x)
;; 	(max-h 0)
;; 	(views '()))
;;     (dolist (i eqns)
;;       (multiple-value-bind (h view) (draw-eqn i cur-x y text-view)
;; ;;  	(format t "tdb: ~a ~a ~a ~a ~a~%" h max-h view (point-h (view-size view)) (point-v (view-size view)))
;; 	(incf cur-x (point-h (view-size view)))
;; 	(when (> h max-h) (setf max-h h))
;; 	(push view views)))
;; ;;       ;; Add 'becomes'
;; ;;       (let ((v (draw-arrow *window* cur-x (+ y 20) 20 20)))
;; ;; 	(incf cur-x (point-h (view-size v)))
;; ;; 	(push v views)))
;;     ;; bottom Align them.
;;     (dolist (i views)
;;       (set-view-position i
;; 			 (add-points (view-position i)
;; 				     (make-point 0 (- max-h (point-v (view-size i)))))))
;;     (values (+ max-h +view-pad+) views)))

(defun draw-special-text-block (txt type x y text-view)
  (let ((view (ecase type
		(:title (draw-text-with-font text-view x y txt *title-font*))
		(:section (draw-text-with-font text-view x y txt *section-font*)))))
    (values (+ (point-v (view-size view)) +view-pad+)
	    (list view))))

(defun draw-title-block (title x y view)
  (draw-special-text-block title :title x y view))  

(defun draw-section-block (title x y view)
  (draw-special-text-block title :section x y view))

(defun draw-input-block (initial-text x y view)
  (values (- (view-height view) y)
          (list (draw-user-instruction view (- x 10) y (view-width view) (- (view-height view) y) initial-text))))


(defun draw-block (txt &optional (view *text-view*))
  (if txt
      (let ((x 10)
            (y 10))
        (dolist (i txt)
          (multiple-value-bind (h views) (ecase (first i)
                                           (:text (draw-text-block (second i) x y view))
                                           (:title (draw-title-block (second i) x y view))
                                           (:section (draw-section-block (second i) x y view))
                                           (:input (draw-input-block (second i) x y view))
                                           ;; (:eqn (draw-eqn-block (rest i) x y view))
                                           )
            (incf y h)
            (setf *text-block-widgets* (nconc *text-block-widgets* views)))))
      ;; 	  ;; Update the scrollbar
      ;; 	  (set-scroll-bar-max *scroll-view* (- y (- +window-height+ 50))))
      (fsm-next)))

(defun display-current-block ()
  (let ((txt (car *current-text-block*)))
    (if txt
        (draw-block txt *text-view*)
      (fsm-next))))

(defun display-next-text-block ()
  (history-add-next)
  (pop *current-text-block*)
  (when *text-block-widgets*
    (apply #'remove-subviews *text-view* *text-block-widgets*)
    (setf *text-block-widgets* nil))
  (when (null (cdr *current-text-block*))
    (set-dialog-item-text *next-button* "Begin"))
  (display-current-block))
  

(defun display-prev-text-block ()
  (history-add-prev)
  (when (null (cdr *current-text-block*))
    (set-dialog-item-text *next-button* "Next"))
  ;; Move backwards in the list
  (let ((n (position (car *current-text-block*) *text-block-list* :test #'equal)))
    (setf *current-text-block* (if (and n (> n 0))
				   (nthcdr (1- n) *text-block-list*)
				 *text-block-list*)))
  (when *text-block-widgets*
    (apply #'remove-subviews *text-view* *text-block-widgets*)
    (setf *text-block-widgets* nil))
  (display-current-block))

(defun make-text-button (name point action)
  (make-instance 'button-dialog-item
		 :view-container *window*
		 :view-position point 
		 :dialog-item-text name
		 :dialog-item-action (lambda (button)
				       (declare (ignore button))
				       (funcall action))))
(defun draw-text-buttons ()
  (setf *prev-button* (make-text-button "Back" (make-point 10 (- +window-height+ 50)) 'display-prev-text-block)
	*next-button* (make-text-button "Next" (make-point 80 (- +window-height+ 50)) 'display-next-text-block)))

(defun display-text (&rest text)
  (setf *text-block-list* text
	*current-text-block* *text-block-list*)
  (draw-text-buttons)
  (display-current-block))

(defun slurp-stream (stream)
  "Return the contents of file as a string."
  (with-output-to-string (out)
    (do ((line (read-line stream nil stream)
	       (read-line stream nil stream)))
	((eq line stream))
      (write-line line out))))

(defun slurp-file (file)
  "Return the contents of the file as a string."
  (with-open-file (in file)
    (slurp-stream in)))

(defun slurp-file-2 (file)
  "Read the list of strings in file and return it."
  (with-open-file (in file)
    (read in)))
	
(defun display-file (file)
  (apply 'display-text (slurp-file-2 file)))

(defmethod fsm-dump-screen ((blk (eql 'text)))
  (list (cons :text
	      (car *current-text-block*))
	(list :buttons :prev :next)))

(defmethod fsm-execute-command ((blk (eql 'text)) (cmd (eql :click)) button)
  (cond ((eql button :next)
	 (display-next-text-block))
        ((eql button :back)
	 (display-prev-text-block))
	(t (error "Unknown button ~a" button))))
