;;; What's next functionality

;; find an operation that can be performed and tell the user about it.

(in-package :tutor)

(defvar *whatsnext-arrows* nil)

(defvar *whatsnext-hint* nil
  "")

(defun hintable-operations ()
  "Return the list of operations that can be used."
  ;; of the valid operations some apply to tutor2 some to tutor1, so
  ;; remove any that are not found.
  (let ((operations (remove-if-not 'find-operation *valid-operations*)))
    (if (find :no-canonicalize (current-hint-modifiers))
        (remove 'canonicalize operations)
        operations)))

(defun init-whatsnext-arrows ()
  (setf *whatsnext-arrows* nil
        *fine-grained-hint-view* nil
        *fine-grained-hint-pointing-to* nil))

(defun box-left-side (b)
  (values (- (+ (point-h (view-position (tutor-eqn-view *tutor*))) (box-x b))
	     (point-h (view-scroll-position (tutor-eqn-view *tutor*)))
	     ;; Make a bit of a space between the arrow tip and the box
	     25
	     +pad+)
	  (+ (- (+ (point-v (view-position (tutor-eqn-view *tutor*))) (box-y b))
		(point-v (view-scroll-position (tutor-eqn-view *tutor*))))
	     +pad+)))

(defun box-clear-arrows (root)
  (setf (box-pointed-to root) nil)
  (mapc #'box-clear-arrows (only-boxes (children root))))

(defun clear-fine-grained-hint-arrow ()
  (when *fine-grained-hint-view* 
    (remove-subviews *window* *fine-grained-hint-view*)
    (setf *fine-grained-hint-view* nil
          *fine-grained-hint-pointing-to* nil)))

(defun clear-arrows ()
  (clear-fine-grained-hint-arrow)
  ;; clear the box's arrow cache
  (box-clear-arrows (tutor-eqn-root *tutor*))
  (when *whatsnext-arrows*
    (apply #'remove-subviews (tutor-eqn-area-view *tutor*) *whatsnext-arrows*)
    (setf *whatsnext-arrows* nil)))

(defun point-to-boxes (boxes)
;;   (clear-arrows)
;;   (setf *whatsnext-arrows*
	(loop for b in boxes
	      do (setf (box-pointed-to b) t))
        (refresh-eqn))
;; 	      collect (multiple-value-bind (x y) (box-left-side b)
;; 			(draw-arrow (tutor-eqn-area-view *tutor*) x y 20 20 *red-color*)))))

;; (defun valid-operation-patterns (patterns root)
;;   (labels ((valid-operation (pattern root)
;;              (multiple-value-bind (matched alist) (matcher pattern root)
;;                ;; (format t "alist: ~a rest: ~a~%" alist rest)
;;                (when matched
;;                  (values (position pattern patterns :test #'equal) alist)))))
;;     (dolist (p patterns)
;;       (multiple-value-bind (i v) (valid-operation p root)
;;         (if (and i v)
;;             (return-from valid-operation-patterns (values i v)))))))

;; (defun minus-case-is-allowed (i)
;;   "Return T if this problem set allows minus hints of the Ith kind."
;;   (find (case i (0 :number) (1 :x) (2 :subexpr)) (hint-types (tutor-problem-set *tutor*))))

(defun check-operation (opn root)
  "Return a list of boxes the user must select and the specific
transformation, if this succeeds on ROOT."
  (multiple-value-bind (match trans) (find-working-transformation opn root)
    (when match
      (list (first match) trans))))

(defun walk-boxes (opn root)
  (or (check-operation opn root)
      (loop for i in (only-boxes (children root))
            thereis (walk-boxes opn i))))

(defun find-operation-on-boxes (operations)
  (dolist (i operations)
    (iflet w (walk-boxes (find-operation i) (tutor-eqn-root *tutor*))
      (return-from find-operation-on-boxes
        (values-list (cons i w))))))

(defun can-do-more ()
  "Return T if something more can be done to the problem. NIL
  otherwise."
  (when (find-operation-on-boxes (hintable-operations))
    t))

(defun box-text (singular &optional number-box)
  (if (tutor1-p)
      (if singular
          "box"
        "boxes")
    (if number-box
        "variable"
      (if singular
          "expression"
        "expressions"))))

;; (defun next-operation ()
;;   "Return the operation that can be performed next."
;;   (multiple-value-bind (opn boxes trans) (find-operation-on-boxes (hintable-operations))
;;     (declare (ignore boxes trans))
;;     opn))
             
(defun whats-next (root &optional (operations (hintable-operations)))
  (cond
   ;; if we are in fill-in-the-boxes state, then tell them to click on a green box
   ;; if all the green boxes are filled in, then tell them to click the submit box
   ((tutor-submit-enabled *tutor*)
     (if (input-filled-in-p root)
	 (unless (correct-answer-p)
           (values (click-instruction-text)
                   "Your answer is incorrect."))
       (click-instruction-text)))
   ((correct-problem-answer-p)
    (format nil "Your Answer is correct.~%Click the ~a button~a." 
            (if *simplified-ui* "Next Prob." "Next Problem")
            (if *simplified-ui* "" " (or press the 'n' key)")))
   ;; if not in fill-in-the-boxes state:
   ;; if there are selected boxes, see if an operation can be done on
   ;; them. if it can, tell them to perform it. If not tell them to
   ;; unselect the boxes as no operation can be performed.
   (t
    (multiple-value-bind (opn boxes) (find-operation-on-boxes operations)
      (if opn
	  (progn
	    (point-to-boxes boxes)
	    (cond ((null (set-exclusive-or boxes (selected-boxes)))
		   (format nil "Click the ~a button~a." (op-name (find-operation opn))
                           (if *simplified-ui*
                               ""
                             (format nil "~%(or press the '~C' key)" (operation-quick-key (find-operation opn))))))
		  ;;(< (length boxes) (length (selected-boxes)))
		  ((set-difference (selected-boxes) boxes)
		   (format nil "Click the selected ~a not pointed to by the arrow~p." 
                           (box-text (= 1 (length (set-difference (selected-boxes) boxes))))
                           (length (set-difference (selected-boxes) boxes))))
		  ((set-difference boxes (selected-boxes))
		   (format nil "Click the ~a pointed to by the arrow~p."
                           (box-text (= 1 (length boxes)) 
                                     (typep (first boxes) 'number-box)) 
                           (length boxes)))))
            (format nil "Your answer is incorrect.~%Use the <- and -> buttons (or the left and right arrow keys)~%to review your work and correct the mistake."))))))


(defun problem-stage (&optional (filling-p t))
  "Return the stage in the problem we're at. if FILLING-P is true, return :filling, not just :fill."
  (cond
   ((tutor-submit-enabled *tutor*)
    (if (and *input-window* filling-p)
        ;; actually typing input
        :filling
      :fill))
   ((correct-problem-answer-p)
    :done)
   (t
    (multiple-value-bind (opn boxes) (find-operation-on-boxes (hintable-operations))
      (if (and opn 
               (set-exclusive-or boxes (selected-boxes)))
          :select
        :operation)))))

(defun get-angela-hint (op kind hint)
  (let ((num (if (numberp hint) hint (position hint '(:select :operation :fill :filling)))))
    (assert (numberp num))
    (wrap-message-text (nth num (nth kind (cdr (assoc op *angela-hints*)))))))

(defun get-next-word-hint (operation hint-index)
  "Return the next written hint."
  (let* ((stage (problem-stage))
         (hints (nth hint-index (cdr (assoc operation *angela-hints*))))
         (editables (editable-boxes (tutor-eqn-root *tutor*)))
         (box (find-empty-editable (tutor-eqn-root *tutor*))))
    (wrap-message-text (ecase stage
                         (:select (nth 0 hints))
                         (:operation (nth 1 hints))
                         (:fill (nth (+ 2 (* (position box editables) 2)) hints))
                         (:filling (nth (+ 3 (* (position box editables) 2)) hints))))))

(defun word-hints-whats-next (root &optional (operations (hintable-operations)))
  (cond ((correct-problem-answer-p)
         (format nil "Your Answer is correct.~%Click the Next Problem button."))
        ((and (tutor-submit-enabled *tutor*)
              (input-filled-in-p root))
         (if ;;*fill-in-object*
             (find-empty-editable root)
             ;; give them the hint when they're refilling in the box
             (values (get-next-word-hint (submit-operation (tutor-submit-enabled *tutor*))
                                         (submit-hint-index (tutor-submit-enabled *tutor*)))
                     "Your answer is incorrect.")
           "Your answer is incorrect."))
        ((tutor-submit-enabled *tutor*)
         (get-next-word-hint (submit-operation (tutor-submit-enabled *tutor*))
                             (submit-hint-index (tutor-submit-enabled *tutor*))))
        (t (multiple-value-bind (opn boxes trans) (find-operation-on-boxes operations)
             (if opn
                 (values (get-next-word-hint opn (transformation-hint-index trans))
                         ;; sometimes we need to notify them that they've selected too many boxes
                         (let* ((diff (set-difference (selected-boxes) boxes))
                                (len (length diff))
                                (plural (> len 1)))
                           (if diff
                               (format nil "You've selected ~d box~a that shouldn't be selected.~%Unselect ~a by clicking ~a."
                                       len (if plural "es" "") (if plural "them" "it") (if plural "them" "it"))
                               "")))
                 "Click the button labeled \"First Mistake\"")))))

(defun no-hints-whats-next (root &optional (operations (hintable-operations)))
  (declare (ignore operations))
  (cond ((and (tutor-submit-enabled *tutor*)
              (input-filled-in-p root))
         "Your answer is incorrect.")
        (t nil)))

(defun handle-hint (clicked-p)
  "User clicked the hint button."
  (history-add 'whats-next clicked-p)
  ;; incentives. Don't change them if they're already getting hints.
  (unless (or (tutor-auto-whats-next-mode *tutor*)
              *cycle-hints-p*
              *hints-stage*)
    (decf (tutor-problem-score *tutor*) +hint-points+))
  (when (and (or (find :fine-grained *hint-types*)
                 (find :written *hint-types*))
             ;; only do the hint stage stuff if we are'nt giving hints
             ;; the whole way through
             (not (or (tutor-global-whats-next *tutor*)
                      (tutor-auto-whats-next-mode *tutor*))))
    (setf *hints-stage* (problem-stage nil)))
  (do-whats-next *hint-types*))

(defun update-hints ()
  "Call this function any time the hints may need to be updated."
  (when (or (auto-whats-next-mode)
            *cycle-hints-p*
            *hints-stage*)
    (do-whats-next *hint-types*)))

(defun do-whats-next (hint-types)
  (if (and (or (find :fine-grained hint-types)
               (find :written hint-types))
           *hints-stage*
           (not (eq *hints-stage* (problem-stage nil))))
      ;; we got to the next stage, so don't give anymore hints.
      (progn
        (clear-fine-grained-hint-arrow)
        (setf *hints-stage* nil))
      (progn
        (clear-arrows)
        (unless *nohint-override*
          ;; :user overrides all other hint types
          (if (find :user hint-types)
              (user-supplied-hint (tutor-eqn-root *tutor*) (hintable-operations))
              (progn
                (when (find :fine-grained hint-types)
                  (fine-grained-hint))
                (multiple-value-bind (hint header) (funcall (cond ((find :written hint-types)
                                                                   'word-hints-whats-next)
                                                                  ((find :normal hint-types)
                                                                   'whats-next)
                                                                  (t 'no-hints-whats-next))
                                                            (tutor-eqn-root *tutor*)
                                                            (hintable-operations))
                  ;; Only overwrite the error text if the hint has
                  ;; changed. Yes, its a hack.
                  (when (or header 
                            ;; There needs to be hint text before we'll erase
                            ;; the message-text.
                            (and *hint-text*
                                 (not (string-equal hint *hint-text*))))
                    (set-message-text header))
                  (set-hint-text hint)))))))
  ;; make sure the new message is visible
  (when *simplified-ui*
    (simplified-offset-for-message)))

(defun store-stage-hint (hint)
  (setf (getf *current-user-supplied-hints* (problem-stage t)) hint))

(defun set-user-hint-text (operation stage text)
  (let ((v (gethash operation *user-supplied-hints*)))
    (setf (getf v stage) text
          (gethash operation *user-supplied-hints*) v)))

(let (last-stage)
  (defun user-supplied-hint (root &optional (operations (hintable-operations)))
    "Give the user a chance to fill in the hint text. They must
fill it in, then do the stage then click the done button."
    (declare (ignore operations))
    (let ((stage (problem-stage t)))
      (unless (eq stage last-stage)
        (set-message-text nil)
        (set-hint-text nil))
      ;; always set the hint text to nil when the stage is fill and
      ;; there are >1 green boxes. This way they always need to fill
      ;; it in. yes its a hack and yes it works.
      (unless (eq stage :done)
        (set-hint-text :input (if (or (find stage '(:select :operation))
                                      (= (length (editable-boxes root)) 1))
                                  (getf *current-user-supplied-hints* stage)
                                  nil)))
      (when (and (tutor-submit-enabled *tutor*)
                 (input-filled-in-p root))
        (set-message-text "Your answer is incorrect."))
      (setf last-stage stage))))

;;; Fine grained hints

;; (defvar *hint-dest-point* nil)
;; (defvar *hint-source-point* nil)

;; (defun move-hint (percent)
;;   "move the hint text to PERCENT close to the destination."
;;   (let ((sx (point-h *hint-source-point*))
;;         (sy (point-v *hint-source-point*))
;;         (dx (point-h *hint-dest-point*))
;;         (dy (point-v *hint-dest-point*)))
;;     (set-view-position *fine-grained-hint-view* 
;;                        (+ sx (truncate (* (- dx sx) percent)))
;;                        (+ sy (truncate (* (- dy sy) percent))))))

;; (defun move-proc ()
;;   (loop for i from 0 to 100 by 20
;;         do (move-hint (/ i 100))
;;         do (sleep 0.01)))

(defun overlapping-rects-p (x1 y1 w1 h1 x2 y2 w2 h2)
  (not (or (> x1 (+ x2 w2))
           (> y1 (+ y2 h2))
           (> x2 (+ x1 w1))
           (> y2 (+ y1 h1)))))

(defun overlaps-boxes-p (x y text)
  "Does the text at x,y overlap any boxes?"
  (let* ((len (+ (string-width text '("Monaco" 12)) 24))
         (vert 24)
         (x (- x len)))
    (labels ((doboxes (root)
                      (or (overlapping-rects-p x y len vert (abs-box-x root (tutor-eqn-view *tutor*)) (abs-box-y root (tutor-eqn-view *tutor*)) +box-width+ +box-height+)
                          (loop for i in (only-boxes (children root))
                                thereis (doboxes i)))))
      (doboxes (tutor-eqn-root *tutor*)))))

(defun set-fine-grained-hint (x y text length thing &optional dont-flip)
  (when *fine-grained-hint-view* 
    (remove-subviews *window* *fine-grained-hint-view*))
  (setf *fine-grained-hint-view*
        (let* ((view (make-instance 'view
                                    :view-position (make-point x y)
                                    :view-container *window*))
               (overlaps (and (not dont-flip) (overlaps-boxes-p x y text)))
               (h 28)
               (sv (if overlaps
                       (list (draw-arrow view 0 0 h h :left *red-color*)
                             (draw-text-with-font view 0 0 text *fine-grained-hint-font* `(:text ,*white-color* :body ,*red-color*) t))
                       (list (draw-text-with-font view 0 0 (concat " " text) *fine-grained-hint-font* `(:text ,*white-color* :body ,*red-color*) t)
                             (draw-arrow view 0 0 h h :right *red-color*)))))
          (order-views sv)
          (resize-view-for view sv)
          (if overlaps
              (set-view-position view (+ x length) y)
              (set-view-position view (- x 2 (view-width view)) y))
          view)
        *fine-grained-hint-pointing-to* (list thing text)))

(defun point-to-button (button)
  (let* ((op (typecase button
               (symbol (let ((op (find-operation button)))
                         (if op 
                             (op-name op)
                           (symbol-to-button button))))
               (t (op-name button))))
         (thing (typecase button
                  (symbol button)
                  (t (operation-name button t))))
         (view (find-if (lambda (b) 
                          (string= (dialog-item-text b) op))
                        (all-subviews *window* 'button-dialog-item))))
    (set-fine-grained-hint (- (abs-x view) +pad+) (abs-y view) "Click this" 0 thing)))

(defun point-to-box (box)
  (set-fine-grained-hint (- (abs-box-x box (tutor-eqn-view *tutor*)) (* +pad+ 2))
                         (+ (abs-box-y box (tutor-eqn-view *tutor*)) 3)
                         "Click this"
                         (+ 13 +box-width+)
                         box))


(defun point-to-box-editable (box &optional hint)
  (when (box-answer-view box)
    (if (and hint
             *input-window* 
             (typep *input-window* 'input-window))
        (set-fine-grained-hint (view-x *input-window*)
                               (view-y *input-window*)
                               (format nil "Type ~{~a ~^~}" (mapcar 'arg-answer (editables box)))
                               (+ 7 (view-width (box-answer-view box))) 
                               :input-box t)
        (set-fine-grained-hint (- (abs-x (box-answer-view box)) (* +pad+ 2))
                               (abs-y (box-answer-view box))
                               (if hint
                                   (format nil "Type ~{~a ~^~}" (mapcar 'arg-answer (editables box)))
                                   "Click this")
                               (+ 7 (view-width (box-answer-view box)))
                               (if hint :input-box box)))))

(defun point-to-box-editable-arg (box &optional hint)
  ;; editables better only 1
  (let ((arg (first (editables box))))
    (if (and hint
             *input-window* 
             (typep *input-window* 'input-window))
        (set-fine-grained-hint (view-x *input-window*)
                               (view-y *input-window*)
                               (format nil "Type ~s" (arg-answer arg))
                               (view-width (arg-widget arg)) 
                               :input-box t) 
        (set-fine-grained-hint (abs-x (arg-widget arg))
                               (abs-y (arg-widget arg))
                               (if hint
                                   (format nil "Type ~s" (arg-answer arg))
                                   "Click this")
                               (view-width (arg-widget arg))
                               (if hint :input-box arg) t))))

(defun find-empty-editable (tree)
  "return a box that contains an empty editable arg/box."
  (or *fill-in-object*
      (let ((e (editables tree)))
        (if (or
             ;; either its empty
             (and e
                  (loop for i in e never (arg-user-answer i))
                  (null (box-user-input tree)))
             ;; or filled in incorrectly
             (and e
                  (loop for i in e thereis (not (eql (arg-user-answer i) (arg-answer i))))))
             tree
          (loop for i in (only-boxes (children tree))
                do (let ((ed (find-empty-editable i)))
                     (when ed
                       (return ed))))))))

(defun fine-grained-hint ()
  (let ((empty-editable (find-empty-editable (tutor-eqn-root *tutor*)))
        (operations (hintable-operations)))
    (multiple-value-bind (opn boxes) (find-operation-on-boxes operations)
      (cond
       ((correct-problem-answer-p)
        (point-to-button :next-problem))
       ((and (not (tutor-submit-enabled *tutor*))
             opn)
        (cond ((null (set-exclusive-or boxes (selected-boxes)))
               (point-to-button (find-operation opn)))
;;               ((set-difference (selected-boxes) boxes)
;;                (let ((unclick (set-difference (selected-boxes) boxes)))
;;                  (point-to-box (first unclick))))
              ((set-exclusive-or boxes (selected-boxes))
               (let ((click (set-exclusive-or boxes (selected-boxes))))
                 (point-to-box (first click))))))
       (empty-editable
        (or (point-to-box-editable empty-editable *input-window*)
            (point-to-box-editable-arg empty-editable *input-window*)))
       ;; FIXME: should the arrow disappear?
       ((and (tutor-submit-enabled *tutor*)
             (input-filled-in-p (tutor-eqn-root *tutor*)))
        (clear-fine-grained-hint-arrow))
       (t (point-to-button :last-correct-state))))))
